@@ -140,8 +140,8 @@ typedef jit_int32_t		jit_fpr_t;
 #  include <lightning/jit_hppa.h>
 #elif defined(__aarch64__)
 #  include <lightning/jit_aarch64.h>
-#elif defined(__s390x__)
-#  include <lightning/jit_s390x.h>
+#elif defined(__s390__) || defined(__s390x__)
+#  include <lightning/jit_s390.h>
 #elif defined(__alpha__)
 #  include <lightning/jit_alpha.h>
 #endif
@@ -174,10 +174,10 @@ typedef struct jit_node		jit_node_t;
 typedef struct jit_state	jit_state_t;
 
 typedef enum {
-#define jit_data(u,v,w)		_jit_data(_jit,u,v,w)
     jit_code_data,
 #define jit_live(u)		jit_new_node_w(jit_code_live, u)
-    jit_code_live,
+#define jit_align(u)		jit_new_node_w(jit_code_align, u)
+    jit_code_live,		jit_code_align,
     jit_code_save,		jit_code_load,
 #define jit_name(u)		_jit_name(_jit,u)
     jit_code_name,
@@ -207,6 +207,8 @@ typedef enum {
 #  define jit_getarg_ui(u,v)	_jit_getarg_ui(_jit,u,v)
 #  define jit_getarg_l(u,v)	_jit_getarg_l(_jit,u,v)
 #endif
+#  define jit_putargr(u,v)	_jit_putargr(_jit,u,v)
+#  define jit_putargi(u,v)	_jit_putargi(_jit,u,v)
     jit_code_arg,
 
 #define jit_addr(u,v,w)		jit_new_node_www(jit_code_addr,u,v,w)
@@ -227,7 +229,9 @@ typedef enum {
 #define jit_subxr(u,v,w)	jit_new_node_www(jit_code_subxr,u,v,w)
 #define jit_subxi(u,v,w)	jit_new_node_www(jit_code_subxi,u,v,w)
     jit_code_subxr,		jit_code_subxi,
-
+#define jit_rsbr(u,v,w)		jit_subr(u,w,v)
+#define jit_rsbi(u,v,w)		jit_new_node_www(jit_code_rsbi,u,v,w)
+    jit_code_rsbi,
 #define jit_mulr(u,v,w)		jit_new_node_www(jit_code_mulr,u,v,w)
 #define jit_muli(u,v,w)		jit_new_node_www(jit_code_muli,u,v,w)
     jit_code_mulr,		jit_code_muli,
@@ -325,9 +329,22 @@ typedef enum {
 #  define jit_extr_ui(u,v)	jit_new_node_ww(jit_code_extr_ui,u,v)
 #endif
     jit_code_extr_i,		jit_code_extr_ui,
-#define jit_htonr(u,v)		jit_new_node_ww(jit_code_htonr,u,v)
-#define jit_ntohr(u,v)		jit_new_node_ww(jit_code_htonr,u,v)
-    jit_code_htonr,
+
+#define jit_htonr_us(u,v)	jit_new_node_ww(jit_code_htonr_us,u,v)
+#define jit_ntohr_us(u,v)	jit_new_node_ww(jit_code_htonr_us,u,v)
+    jit_code_htonr_us,
+#define jit_htonr_ui(u,v)	jit_new_node_ww(jit_code_htonr_ui,u,v)
+#define jit_ntohr_ui(u,v)	jit_new_node_ww(jit_code_htonr_ui,u,v)
+#if __WORDSIZE == 32
+#  define jit_htonr(u,v)	jit_new_node_ww(jit_code_htonr_ui,u,v)
+#  define jit_ntohr(u,v)	jit_new_node_ww(jit_code_htonr_ui,u,v)
+#else
+#define jit_htonr_ul(u,v)	jit_new_node_ww(jit_code_htonr_ul,u,v)
+#define jit_ntohr_ul(u,v)	jit_new_node_ww(jit_code_htonr_ul,u,v)
+#  define jit_htonr(u,v)	jit_new_node_ww(jit_code_htonr_ul,u,v)
+#  define jit_ntohr(u,v)	jit_new_node_ww(jit_code_htonr_ul,u,v)
+#endif
+    jit_code_htonr_ui,		jit_code_htonr_ul,
 
 #define jit_ldr_c(u,v)		jit_new_node_ww(jit_code_ldr_c,u,v)
 #define jit_ldi_c(u,v)		jit_new_node_wp(jit_code_ldi_c,u,v)
@@ -524,6 +541,8 @@ typedef enum {
 
 #define jit_arg_f()		_jit_arg_f(_jit)
 #define jit_getarg_f(u,v)	_jit_getarg_f(_jit,u,v)
+#define jit_putargr_f(u,v)	_jit_putargr_f(_jit,u,v)
+#define jit_putargi_f(u,v)	_jit_putargi_f(_jit,u,v)
     jit_code_arg_f,
 
 #define jit_addr_f(u,v,w)	jit_new_node_www(jit_code_addr_f,u,v,w)
@@ -532,6 +551,9 @@ typedef enum {
 #define jit_subr_f(u,v,w)	jit_new_node_www(jit_code_subr_f,u,v,w)
 #define jit_subi_f(u,v,w)	jit_new_node_wwf(jit_code_subi_f,u,v,w)
     jit_code_subr_f,		jit_code_subi_f,
+#define jit_rsbr_f(u,v,w)	jit_subr_f(u,w,v)
+#define jit_rsbi_f(u,v,w)	jit_new_node_wwf(jit_code_rsbi_f,u,v,w)
+    jit_code_rsbi_f,
 #define jit_mulr_f(u,v,w)	jit_new_node_www(jit_code_mulr_f,u,v,w)
 #define jit_muli_f(u,v,w)	jit_new_node_wwf(jit_code_muli_f,u,v,w)
     jit_code_mulr_f,		jit_code_muli_f,
@@ -666,6 +688,8 @@ typedef enum {
 
 #define jit_arg_d()		_jit_arg_d(_jit)
 #define jit_getarg_d(u,v)	_jit_getarg_d(_jit,u,v)
+#define jit_putargr_d(u,v)	_jit_putargr_d(_jit,u,v)
+#define jit_putargi_d(u,v)	_jit_putargi_d(_jit,u,v)
     jit_code_arg_d,
 
 #define jit_addr_d(u,v,w)	jit_new_node_www(jit_code_addr_d,u,v,w)
@@ -674,6 +698,9 @@ typedef enum {
 #define jit_subr_d(u,v,w)	jit_new_node_www(jit_code_subr_d,u,v,w)
 #define jit_subi_d(u,v,w)	jit_new_node_wwd(jit_code_subi_d,u,v,w)
     jit_code_subr_d,		jit_code_subi_d,
+#define jit_rsbr_d(u,v,w)	jit_subr_d(u,w,v)
+#define jit_rsbi_d(u,v,w)	jit_new_node_wwd(jit_code_rsbi_d,u,v,w)
+    jit_code_rsbi_d,
 #define jit_mulr_d(u,v,w)	jit_new_node_www(jit_code_mulr_d,u,v,w)
 #define jit_muli_d(u,v,w)	jit_new_node_wwd(jit_code_muli_d,u,v,w)
     jit_code_mulr_d,		jit_code_muli_d,
@@ -826,6 +853,7 @@ typedef enum {
 #define jit_movi_d_w(u, v)	jit_new_node_wd(jit_code_movi_d_w, u, v)
 
     jit_code_x86_retval_f,	jit_code_x86_retval_d,
+#define jit_code_last_code	jit_code_x86_retval_d
 } jit_code_t;
 
 typedef void* (*jit_alloc_func_ptr)	(size_t);
@@ -835,7 +863,7 @@ typedef void  (*jit_free_func_ptr)	(void*);
 /*
  * Prototypes
  */
-extern void init_jit(char*);
+extern void init_jit(const char*);
 extern void finish_jit(void);
 
 extern jit_state_t *jit_new_state(void);
@@ -846,14 +874,18 @@ extern void _jit_destroy_state(jit_state_t*);
 
 #define jit_address(node)	_jit_address(_jit, node)
 extern jit_pointer_t _jit_address(jit_state_t*, jit_node_t*);
-extern jit_node_t *_jit_data(jit_state_t*, jit_pointer_t,
-			     jit_word_t, jit_int32_t);
-extern jit_node_t *_jit_name(jit_state_t*, char*);
-extern jit_node_t *_jit_note(jit_state_t*, char*, int);
+extern jit_node_t *_jit_name(jit_state_t*, const char*);
+extern jit_node_t *_jit_note(jit_state_t*, const char*, int);
 extern jit_node_t *_jit_label(jit_state_t*);
 extern jit_node_t *_jit_forward(jit_state_t*);
 extern jit_node_t *_jit_indirect(jit_state_t*);
 extern void _jit_link(jit_state_t*, jit_node_t*);
+#define jit_forward_p(u)	_jit_forward_p(_jit,u)
+extern jit_bool_t _jit_forward_p(jit_state_t*,jit_node_t*);
+#define jit_indirect_p(u)	_jit_indirect_p(_jit,u)
+extern jit_bool_t _jit_indirect_p(jit_state_t*,jit_node_t*);
+#define jit_target_p(u)		_jit_target_p(_jit,u)
+extern jit_bool_t _jit_target_p(jit_state_t*,jit_node_t*);
 
 extern void _jit_prolog(jit_state_t*);
 
@@ -870,6 +902,8 @@ extern void _jit_getarg_i(jit_state_t*, jit_gpr_t, jit_node_t*);
 extern void _jit_getarg_ui(jit_state_t*, jit_gpr_t, jit_node_t*);
 extern void _jit_getarg_l(jit_state_t*, jit_gpr_t, jit_node_t*);
 #endif
+extern void _jit_putargr(jit_state_t*, jit_gpr_t, jit_node_t*);
+extern void _jit_putargi(jit_state_t*, jit_word_t, jit_node_t*);
 
 extern void _jit_prepare(jit_state_t*);
 extern void _jit_ellipsis(jit_state_t*);
@@ -907,6 +941,10 @@ extern void _jit_set_code(jit_state_t*, jit_pointer_t, jit_word_t);
 extern jit_pointer_t _jit_get_data(jit_state_t*, jit_word_t*, jit_word_t*);
 #define jit_set_data(u,v,w)	_jit_set_data(_jit,u,v,w)
 extern void _jit_set_data(jit_state_t*, jit_pointer_t, jit_word_t, jit_word_t);
+#define jit_frame(u)		_jit_frame(_jit,u)
+extern void _jit_frame(jit_state_t*, jit_int32_t);
+#define jit_tramp(u)		_jit_tramp(_jit,u)
+extern void _jit_tramp(jit_state_t*, jit_int32_t);
 #define jit_emit()		_jit_emit(_jit)
 extern jit_pointer_t _jit_emit(jit_state_t*);
 
@@ -915,6 +953,8 @@ extern void _jit_print(jit_state_t*);
 
 extern jit_node_t *_jit_arg_f(jit_state_t*);
 extern void _jit_getarg_f(jit_state_t*, jit_fpr_t, jit_node_t*);
+extern void _jit_putargr_f(jit_state_t*, jit_fpr_t, jit_node_t*);
+extern void _jit_putargi_f(jit_state_t*, jit_float32_t, jit_node_t*);
 extern void _jit_pushargr_f(jit_state_t*, jit_fpr_t);
 extern void _jit_pushargi_f(jit_state_t*, jit_float32_t);
 extern void _jit_retr_f(jit_state_t*, jit_fpr_t);
@@ -923,6 +963,8 @@ extern void _jit_retval_f(jit_state_t*, jit_fpr_t);
 
 extern jit_node_t *_jit_arg_d(jit_state_t*);
 extern void _jit_getarg_d(jit_state_t*, jit_fpr_t, jit_node_t*);
+extern void _jit_putargr_d(jit_state_t*, jit_fpr_t, jit_node_t*);
+extern void _jit_putargi_d(jit_state_t*, jit_float64_t, jit_node_t*);
 extern void _jit_pushargr_d(jit_state_t*, jit_fpr_t);
 extern void _jit_pushargi_d(jit_state_t*, jit_float64_t);
 extern void _jit_retr_d(jit_state_t*, jit_fpr_t);
@@ -975,6 +1017,8 @@ extern jit_node_t *_jit_new_node_pwf(jit_state_t*, jit_code_t,
 extern jit_node_t *_jit_new_node_pwd(jit_state_t*, jit_code_t,
 				     jit_pointer_t, jit_word_t, jit_float64_t);
 
+#define jit_arg_register_p(u)		_jit_arg_register_p(_jit,u)
+extern jit_bool_t _jit_arg_register_p(jit_state_t*, jit_node_t*);
 #define jit_callee_save_p(u)		_jit_callee_save_p(_jit,u)
 extern jit_bool_t _jit_callee_save_p(jit_state_t*, jit_int32_t);
 #define jit_pointer_p(u)		_jit_pointer_p(_jit,u)
