@@ -23,11 +23,17 @@
 #define JIT_HASH_CONSTS		1
 #define JIT_NUM_OPERANDS	3
 
+#if __powerpc__
+#  if _CALL_ELF == 2
+/* __BYTE_ORDER == __LITTLE_ENDIAN */
+#    define ABI_ELFv2		1
+#  endif
+#endif
+
 /*
  * Types
  */
 typedef enum {
-#define jit_arg_reg_p(i)	((i) >= 0 && (i) < 8)
 #if __ppc__
 #  define jit_r(i)		(_R11 + (i))
 #else
@@ -41,7 +47,6 @@ typedef enum {
 #  define jit_v(i)		(_R27 - (i))
 #  define jit_v_num()		14
 #endif
-#define jit_arg_f_reg_p(i)	((i) >= 0 && (i) < 13)
 #define jit_f(i)		(_F14 + (i))
 #define jit_f_num()		8
     _R0,
