@@ -50,11 +50,15 @@
 #if defined(__i386__) || defined(__x86_64__)
 #  define JIT_SP		_RSP
 #  define JIT_RET		_RAX
-#  if __WORDSIZE == 32
+#  if __X32
 #    define JIT_FRET		_ST0
 typedef jit_uint32_t		jit_regset_t;
 #  else
-#    define JIT_RA0		_RDI
+#    if __CYGWIN__
+#      define JIT_RA0		_RCX
+#    else
+#      define JIT_RA0		_RDI
+#    endif
 #    define JIT_FA0		_XMM0
 #    define JIT_FRET		_XMM0
 typedef jit_uint64_t		jit_regset_t;
@@ -111,7 +115,7 @@ typedef jit_uint64_t		jit_regset_t;
 #  define JIT_RET		_R0
 #  define JIT_FRET		_V0
 typedef jit_uint64_t		jit_regset_t;
-#elif defined(__s390x__)
+#elif defined(__s390__) || defined(__s390x__)
 #  define JIT_SP		_R15
 #  define JIT_RET		_R2
 #  define JIT_FRET		_F0
@@ -123,6 +127,10 @@ typedef jit_uint32_t		jit_regset_t;
 typedef jit_uint64_t		jit_regset_t;
 #endif
 
+#define jit_data(u,v,w)		_jit_data(_jit,u,v,w)
+extern jit_node_t *_jit_data(jit_state_t*, const void*,
+			     jit_word_t, jit_int32_t);
+
 #define jit_size(vector)	(sizeof(vector) / sizeof((vector)[0]))
 
 #define jit_reg_free_p(regno)						\
@@ -295,6 +303,9 @@ struct jit_node {
     jit_data_t		 v;
     jit_data_t		 w;
     jit_node_t		*link;
+#if DEVEL_DISASSEMBLER
+    jit_uword_t		 offset;
+#endif
 };
 
 struct jit_block {
@@ -345,6 +356,13 @@ struct jit_function {
     jit_int32_t		*regoff;
     jit_regset_t	 regset;
     jit_int32_t		 stack;
+
+    /* Helper for common jit generation pattern, used in GNU Smalltalk
+     * and possibly others, where a static frame layout is required or
+     * assumed. */
+    jit_int32_t		 frame;
+    jit_uint32_t	 define_frame : 1;
+    jit_uint32_t	 assume_frame : 1;
 };
 
 /* data used only during jit generation */
@@ -555,6 +573,9 @@ _jit_save(jit_state_t*, jit_int32_t);
 extern void
 _jit_load(jit_state_t*, jit_int32_t);
 
+#define jit_trampoline(u,v)	_jit_trampoline(_jit, u, v)
+extern void _jit_trampoline(jit_state_t*, jit_int32_t, jit_bool_t);
+
 #define jit_optimize()		_jit_optimize(_jit)
 extern void
 _jit_optimize(jit_state_t*);
@@ -571,6 +592,9 @@ _jit_regarg_p(jit_state_t*, jit_node_t*, jit_int32_t);
 extern jit_pointer_t
 _emit_code(jit_state_t*);
 
+extern void
+jit_flush(void *fptr, void *tptr);
+
 #define emit_ldxi(r0, r1, i0)	_emit_ldxi(_jit, r0, r1, i0)
 extern void
 _emit_ldxi(jit_state_t*, jit_int32_t, jit_int32_t, jit_word_t);
@@ -587,7 +611,7 @@ _emit_ldxi_d(jit_state_t*, jit_int32_t, jit_int32_t, jit_word_t);
 extern void
 _emit_stxi_d(jit_state_t*, jit_word_t, jit_int32_t, jit_int32_t);
 
-extern void jit_init_debug(char*);
+extern void jit_init_debug(const char*);
 extern void jit_finish_debug(void);
 
 extern void jit_init_note(void);
@@ -597,8 +621,11 @@ extern void _jit_set_note(jit_state_t*, jit_note_t*, char*, int, jit_int32_t);
 #define jit_annotate()		_jit_annotate(_jit)
 extern void _jit_annotate(jit_state_t*);
 
-extern jit_pointer_t jit_memcpy(jit_pointer_t,jit_pointer_t,jit_word_t);
-extern jit_pointer_t jit_memmove(jit_pointer_t,jit_pointer_t,jit_word_t);
+#define jit_print_node(u)	_jit_print_node(_jit,u)
+extern void _jit_print_node(jit_state_t*,jit_node_t*);
+
+extern jit_pointer_t jit_memcpy(jit_pointer_t,const void*,jit_word_t);
+extern jit_pointer_t jit_memmove(jit_pointer_t,const void*,jit_word_t);
 extern void jit_alloc(jit_pointer_t*, jit_word_t);
 extern void jit_realloc(jit_pointer_t*, jit_word_t, jit_word_t);
 void jit_free(jit_pointer_t*);
