@@ -467,6 +467,7 @@ typedef union {
 #  define B_C(Cc,Simm19)		oc19(A64_B_C,Cc,Simm19)
 #  define CBZ(Rd,Simm19)		ox19(A64_CBZ|XS,Rd,Simm19)
 #  define CBNZ(Rd,Simm19)		ox19(A64_CBNZ|XS,Rd,Simm19)
+#  define NOP()				ii(0xd503201f)
 static jit_int32_t logical_immediate(jit_word_t);
 #  define oxxx(Op,Rd,Rn,Rm)		_oxxx(_jit,Op,Rd,Rn,Rm)
 static void _oxxx(jit_state_t*,jit_int32_t,jit_int32_t,jit_int32_t,jit_int32_t);
@@ -499,6 +500,8 @@ static void _oxxxc(jit_state_t*,jit_int32_t,jit_int32_t,
 #  define oxxx7(Op,Rt,Rt2,Rn,Simm7)	_oxxx7(_jit,Op,Rt,Rt2,Rn,Simm7)
 static void _oxxx7(jit_state_t*,jit_int32_t,
 		   jit_int32_t,jit_int32_t,jit_int32_t,jit_int32_t);
+#  define nop(i0)			_nop(_jit,i0)
+static void _nop(jit_state_t*,jit_int32_t);
 #  define addr(r0,r1,r2)		ADD(r0,r1,r2)
 #  define addi(r0,r1,i0)		_addi(_jit,r0,r1,i0)
 static void _addi(jit_state_t*,jit_int32_t,jit_int32_t,jit_word_t);
@@ -517,6 +520,8 @@ static void _subci(jit_state_t*,jit_int32_t,jit_int32_t,jit_word_t);
 #  define subxr(r0,r1,r2)		SBCS(r0,r1,r2)
 #  define subxi(r0,r1,i0)		_subxi(_jit,r0,r1,i0)
 static void _subxi(jit_state_t*,jit_int32_t,jit_int32_t,jit_word_t);
+#  define rsbi(r0, r1, i0)		_rsbi(_jit, r0, r1, i0)
+static void _rsbi(jit_state_t*,jit_int32_t,jit_int32_t,jit_word_t);
 #  define mulr(r0,r1,r2)		MUL(r0,r1,r2)
 #  define muli(r0,r1,i0)		_muli(_jit,r0,r1,i0)
 static void _muli(jit_state_t*,jit_int32_t,jit_int32_t,jit_word_t);
@@ -651,9 +656,15 @@ static void _stxi_i(jit_state_t*,jit_word_t,jit_int32_t,jit_int32_t);
 #  define stxi_l(i0,r0,r1)		_stxi_l(_jit,i0,r0,r1)
 static void _stxi_l(jit_state_t*,jit_word_t,jit_int32_t,jit_int32_t);
 #  if __BYTE_ORDER == __LITTLE_ENDIAN
-#    define htonr(r0,r1)		REV(r0,r1)
+#  define htonr_us(r0,r1)		_htonr_us(_jit,r0,r1)
+static void _htonr_us(jit_state_t*,jit_int32_t,jit_int32_t);
+#  define htonr_ui(r0,r1)		_htonr_ui(_jit,r0,r1)
+static void _htonr_ui(jit_state_t*,jit_int32_t,jit_int32_t);
+#    define htonr_ul(r0,r1)		REV(r0,r1)
 #  else
-#    define htonr(r0,r1)		movr(r0,r1)
+#    define htonr_us(r0,r1)		extr_us(r0,r1)
+#    define htonr_ui(r0,r1)		extr_ui(r0,r1)
+#    define htonr_ul(r0,r1)		movr(r0,r1)
 #  endif
 #  define extr_c(r0,r1)			SXTB(r0,r1)
 #  define extr_uc(r0,r1)		UXTB(r0,r1)
@@ -1010,6 +1021,14 @@ _oxxx7(jit_state_t *_jit, jit_int32_t Op,
     ii(i.w);
 }
 
+static void
+_nop(jit_state_t *_jit, jit_int32_t i0)
+{
+    for (; i0 > 0; i0 -= 4)
+	NOP();
+    assert(i0 == 0);
+}
+
 static void
 _addi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
@@ -1110,6 +1129,13 @@ _subxi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
     jit_unget_reg(reg);
 }
 
+static void
+_rsbi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
+{
+    subi(r0, r1, i0);
+    negr(r0, r0);
+}
+
 static void
 _muli(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
@@ -1409,6 +1435,22 @@ _xori(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
     }
 }
 
+#if __BYTE_ORDER == __LITTLE_ENDIAN
+static void
+_htonr_us(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+{
+    htonr_ul(r0, r1);
+    rshi_u(r0, r0, 48);
+}
+
+static void
+_htonr_ui(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+{
+    htonr_ul(r0, r1);
+    rshi_u(r0, r0, 32);
+}
+#endif
+
 static void
 _ldi_c(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
 {
@@ -1717,7 +1759,7 @@ _stxi_c(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0, jit_int32_t r1)
 	STURB(r1, r0, i0 & 0x1ff);
     else {
 	reg = jit_get_reg(jit_class_gpr);
-	addi(rn(reg), r1, i0);
+	addi(rn(reg), r0, i0);
 	str_c(rn(reg), r1);
 	jit_unget_reg(reg);
     }
@@ -1734,7 +1776,7 @@ _stxi_s(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0, jit_int32_t r1)
 	STURH(r1, r0, i0 & 0x1ff);
     else {
 	reg = jit_get_reg(jit_class_gpr);
-	addi(rn(reg), r1, i0);
+	addi(rn(reg), r0, i0);
 	str_s(rn(reg), r1);
 	jit_unget_reg(reg);
     }
@@ -1751,7 +1793,7 @@ _stxi_i(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0, jit_int32_t r1)
 	STURW(r1, r0, i0 & 0x1ff);
     else {
 	reg = jit_get_reg(jit_class_gpr);
-	addi(rn(reg), r1, i0);
+	addi(rn(reg), r0, i0);
 	str_i(rn(reg), r1);
 	jit_unget_reg(reg);
     }
@@ -1768,7 +1810,7 @@ _stxi_l(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0, jit_int32_t r1)
 	STUR(r1, r0, i0 & 0x1ff);
     else {
 	reg = jit_get_reg(jit_class_gpr);
-	addi(rn(reg), r1, i0);
+	addi(rn(reg), r0, i0);
 	str_l(rn(reg), r1);
 	jit_unget_reg(reg);
     }
@@ -2145,6 +2187,13 @@ _calli_p(jit_state_t *_jit, jit_word_t i0)
 static void
 _prolog(jit_state_t *_jit, jit_node_t *node)
 {
+    if (_jitc->function->define_frame || _jitc->function->assume_frame) {
+	jit_int32_t	frame = -_jitc->function->frame;
+	assert(_jitc->function->self.aoff >= frame);
+	if (_jitc->function->assume_frame)
+	    return;
+	_jitc->function->self.aoff = frame;
+    }
     _jitc->function->stack = ((_jitc->function->self.alen -
 			      /* align stack at 16 bytes */
 			      _jitc->function->self.aoff) + 15) & -16;
@@ -2188,6 +2237,8 @@ _prolog(jit_state_t *_jit, jit_node_t *node)
 static void
 _epilog(jit_state_t *_jit, jit_node_t *node)
 {
+    if (_jitc->function->assume_frame)
+	return;
     if (_jitc->function->stack)
 	MOV_XSP(SP_REGNO, FP_REGNO);
 #define LOAD(L, R, O)							\
