@@ -19,6 +19,19 @@
 
 #define rc(value)			jit_class_##value
 #define rn(reg)				(jit_regno(_rvs[jit_regno(reg)].spec))
+#define jit_arg_reg_p(i)		((i) >= 0 && (i) < 6)
+#define jit_arg_f_reg_p(i)		((i) >= 0 && (i) < 6)
+#if __BYTE_ORDER == __LITTLE_ENDIAN
+#  define C_DISP			0
+#  define S_DISP			0
+#  define I_DISP			0
+#  define F_DISP			0
+#else
+#  define C_DISP			8 - sizeof(jit_int8_t)
+#  define S_DISP			8 - sizeof(jit_int16_t)
+#  define I_DISP			8 - sizeof(jit_int32_t)
+#  define F_DISP			8 - sizeof(jit_float32_t)
+#endif
 
 /*
  * Prototypes
@@ -241,12 +254,21 @@ _jit_epilog(jit_state_t *_jit)
     _jitc->function = NULL;
 }
 
+jit_bool_t
+_jit_arg_register_p(jit_state_t *_jit, jit_node_t *u)
+{
+    if (u->code == jit_code_arg)
+	return (jit_arg_reg_p(u->u.w));
+    assert(u->code == jit_code_arg_f || u->code == jit_code_arg_d);
+    return (jit_arg_f_reg_p(u->u.w));
+}
+
 jit_node_t *
 _jit_arg(jit_state_t *_jit)
 {
     jit_int32_t		offset;
     assert(_jitc->function != NULL);
-    if (_jitc->function->self.argi < 6)
+    if (jit_arg_reg_p(_jitc->function->self.argi))
 	offset = _jitc->function->self.argi++;
     else {
 	offset = _jitc->function->self.size;
@@ -255,18 +277,12 @@ _jit_arg(jit_state_t *_jit)
     return (jit_new_node_w(jit_code_arg, offset));
 }
 
-jit_bool_t
-_jit_arg_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (offset >= 0 && offset < 6);
-}
-
 jit_node_t *
 _jit_arg_f(jit_state_t *_jit)
 {
     jit_int32_t		offset;
     assert(_jitc->function != NULL);
-    if (_jitc->function->self.argi < 6)
+    if (jit_arg_f_reg_p(_jitc->function->self.argi))
 	offset = _jitc->function->self.argi++;
     else {
 	offset = _jitc->function->self.size;
@@ -275,18 +291,12 @@ _jit_arg_f(jit_state_t *_jit)
     return (jit_new_node_w(jit_code_arg_f, offset));
 }
 
-jit_bool_t
-_jit_arg_f_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (offset >= 0 && offset < 6);
-}
-
 jit_node_t *
 _jit_arg_d(jit_state_t *_jit)
 {
     jit_int32_t		offset;
     assert(_jitc->function != NULL);
-    if (_jitc->function->self.argi < 6)
+    if (jit_arg_f_reg_p(_jitc->function->self.argi))
 	offset = _jitc->function->self.argi++;
     else {
 	offset = _jitc->function->self.size;
@@ -295,133 +305,176 @@ _jit_arg_d(jit_state_t *_jit)
     return (jit_new_node_w(jit_code_arg_d, offset));
 }
 
-jit_bool_t
-_jit_arg_d_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (jit_arg_f_reg_p(offset));
-}
-
 void
 _jit_getarg_c(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 6)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_c(u, _A0 - v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_c(u, _FP, v->u.w);
-#else
-	jit_ldxi_c(u, JIT_FP, v->u.w + 8 - sizeof(jit_int8_t));
-#endif
-    }
+    else
+	jit_ldxi_c(u, _FP, v->u.w + C_DISP);
 }
 
 void
 _jit_getarg_uc(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 6)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_uc(u, _A0 - v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_uc(u, _FP, v->u.w);
-#else
-	jit_ldxi_uc(u, JIT_FP, v->u.w + 8 - sizeof(jit_uint8_t));
-#endif
-    }
+    else
+	jit_ldxi_uc(u, _FP, v->u.w + C_DISP);
 }
 
 void
 _jit_getarg_s(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 6)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_s(u, _A0 - v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_s(u, _FP, v->u.w);
-#else
-	jit_ldxi_s(u, JIT_FP, v->u.w + 8 - sizeof(jit_int16_t));
-#endif
-    }
+    else
+	jit_ldxi_s(u, _FP, v->u.w + S_DISP);
 }
 
 void
 _jit_getarg_us(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 6)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_us(u, _A0 - v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_us(u, _FP, v->u.w);
-#else
-	jit_ldxi_us(u, JIT_FP, v->u.w + 8 - sizeof(jit_uint16_t));
-#endif
-    }
+    else
+	jit_ldxi_us(u, _FP, v->u.w + S_DISP);
 }
 
 void
 _jit_getarg_i(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 6)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_i(u, _A0 - v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_i(u, _FP, v->u.w);
-#else
-	jit_ldxi_i(u, JIT_FP, v->u.w + 8 - sizeof(jit_int32_t));
-#endif
-    }
+    else
+	jit_ldxi_i(u, _FP, v->u.w + I_DISP);
 }
 
 void
 _jit_getarg_ui(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 6)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_ui(u, _A0 - v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_ui(u, _FP, v->u.w);
-#else
-	jit_ldxi_ui(u, JIT_FP, v->u.w + 8 - sizeof(jit_uint32_t));
-#endif
-    }
+    else
+	jit_ldxi_ui(u, _FP, v->u.w + I_DISP);
 }
 
 void
 _jit_getarg_l(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 6)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_movr(u, _A0 - v->u.w);
     else
 	jit_ldxi_l(u, _FP, v->u.w);
 }
 
+void
+_jit_putargr(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
+	jit_movr(_A0 - v->u.w, u);
+    else
+	jit_stxi(v->u.w, _FP, u);
+}
+
+void
+_jit_putargi(jit_state_t *_jit, jit_word_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
+	jit_movi(_A0 - v->u.w, u);
+    else {
+	regno = jit_get_reg(jit_class_gpr);
+	jit_movi(regno, u);
+	jit_stxi(v->u.w, _FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_getarg_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 6)
+    assert(v->code == jit_code_arg_f);
+    if (jit_arg_f_reg_p(v->u.w))
 	jit_movr_f(u, _F16 - v->u.w);
+    else
+	jit_ldxi_f(u, _FP, v->u.w + F_DISP);
+}
+
+void
+_jit_putargr_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg_f);
+    if (jit_arg_f_reg_p(v->u.w))
+	jit_movr_f(_F16 - v->u.w, u);
+    else
+	jit_stxi_f(v->u.w, _FP, u + F_DISP);
+}
+
+void
+_jit_putargi_f(jit_state_t *_jit, jit_float32_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg_f);
+    if (jit_arg_f_reg_p(v->u.w))
+	jit_movi_f(_F16 - v->u.w, u);
     else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_f(u, _FP, v->u.w);
-#else
-	jit_ldxi_f(u, JIT_FP, v->u.w + 8 - sizeof(jit_float32_t));
-#endif
+	regno = jit_get_reg(jit_class_fpr);
+	jit_movi_f(regno, u);
+	jit_stxi_f(v->u.w, _FP, regno + F_DISP);
+	jit_unget_reg(regno);
     }
 }
 
 void
 _jit_getarg_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 6)
+    assert(v->code == jit_code_arg_d);
+    if (jit_arg_f_reg_p(v->u.w))
 	jit_movr_d(u, _F16 - v->u.w);
     else
 	jit_ldxi_d(u, _FP, v->u.w);
 }
 
+void
+_jit_putargr_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg_d);
+    if (jit_arg_f_reg_p(v->u.w))
+	jit_movr_d(_F16 - v->u.w, u);
+    else
+	jit_stxi_d(v->u.w, _FP, u);
+}
+
+void
+_jit_putargi_d(jit_state_t *_jit, jit_float64_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg_d);
+    if (jit_arg_f_reg_p(v->u.w))
+	jit_movi_d(_F16 - v->u.w, u);
+    else {
+	regno = jit_get_reg(jit_class_fpr);
+	jit_movi_d(regno, u);
+	jit_stxi_d(v->u.w, _FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_pushargr(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function != NULL);
-    if (_jitc->function->call.argi < 6) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movr(_A0 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
@@ -436,7 +489,7 @@ _jit_pushargi(jit_state_t *_jit, jit_int64_t u)
 {
     jit_int32_t		regno;
     assert(_jitc->function != NULL);
-    if (_jitc->function->call.argi < 6) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movi(_A0 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
@@ -453,17 +506,12 @@ void
 _jit_pushargr_f(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function != NULL);
-    if (_jitc->function->call.argi < 6) {
+    if (jit_arg_f_reg_p(_jitc->function->call.argi)) {
 	jit_movr_f(_F16 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
     else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_stxi_f(_jitc->function->call.size, JIT_SP, u);
-#else
-	jit_stxi_f(_jitc->function->call.size + 8 - sizeof(jit_float32_t),
-		   JIT_SP, u);
-#endif
+	jit_stxi_f(_jitc->function->call.size + F_DISP, JIT_SP, u);
 	_jitc->function->call.size += 8;
     }
 }
@@ -473,19 +521,14 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
 {
     jit_int32_t		regno;
     assert(_jitc->function != NULL);
-    if (_jitc->function->call.argi < 6) {
+    if (jit_arg_f_reg_p(_jitc->function->call.argi)) {
 	jit_movi_f(_F16 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
     else {
 	regno = jit_get_reg(jit_class_fpr);
 	jit_movi_f(regno, u);
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_stxi_f(_jitc->function->call.size, JIT_SP, regno);
-#else
-	jit_stxi_f(_jitc->function->call.size + 8 - sizeof(jit_float32_t),
-		   JIT_SP, regno);
-#endif
+	jit_stxi_f(_jitc->function->call.size + F_DISP, JIT_SP, regno);
 	_jitc->function->call.size += 8;
 	jit_unget_reg(regno);
     }
@@ -495,7 +538,7 @@ void
 _jit_pushargr_d(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function != NULL);
-    if (_jitc->function->call.argi < 6) {
+    if (jit_arg_f_reg_p(_jitc->function->call.argi)) {
 	jit_movr_d(_F16 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
@@ -510,7 +553,7 @@ _jit_pushargi_d(jit_state_t *_jit, jit_float64_t u)
 {
     jit_int32_t		regno;
     assert(_jitc->function != NULL);
-    if (_jitc->function->call.argi < 6) {
+    if (jit_arg_f_reg_p(_jitc->function->call.argi)) {
 	jit_movi_d(_F16 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
@@ -741,9 +784,19 @@ _emit_code(jit_state_t *_jit)
 	if (_jit->pc.uc >= _jitc->code.end)
 	    return (NULL);
 
+#if DEVEL_DISASSEMBLER
+	node->offset = _jit->pc.w;
+#endif
 	value = jit_classify(node->code);
 	jit_regarg_set(node, value);
 	switch (node->code) {
+	    case jit_code_align:
+		assert(!(node->u.w & (node->u.w - 1)) &&
+		       node->u.w <= sizeof(jit_word_t));
+		if (node->u.w == sizeof(jit_word_t) &&
+		    (word = _jit->pc.w & (sizeof(jit_word_t) - 1)))
+		    nop(sizeof(jit_word_t) - word);
+		break;
 	    case jit_code_note:		case jit_code_name:
 		node->u.w = _jit->pc.w;
 		break;
@@ -764,6 +817,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrw(subc,);
 		case_rrr(subx,);
 		case_rrw(subx,);
+		case_rrw(rsb,);
 		case_rrr(mul,);
 		case_rrw(mul,);
 		case_rrrr(qmul,);
@@ -842,7 +896,9 @@ _emit_code(jit_state_t *_jit)
 		case_wrr(stx, _i);
 		case_rrr(stx, _l);
 		case_wrr(stx, _l);
-		case_rr(hton,);
+		case_rr(hton, _us);
+		case_rr(hton, _ui);
+		case_rr(hton, _ul);
 		case_rr(ext, _c);
 		case_rr(ext, _uc);
 		case_rr(ext, _s);
@@ -933,6 +989,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrf(add, _f, 32);
 		case_rrr(sub, _f);
 		case_rrf(sub, _f, 32);
+		case_rrf(rsb, _f, 32);
 		case_rrr(mul, _f);
 		case_rrf(mul, _f, 32);
 		case_rrr(div, _f);
@@ -1015,6 +1072,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrf(add, _d, 64);
 		case_rrr(sub, _d);
 		case_rrf(sub, _d, 64);
+		case_rrf(rsb, _d, 64);
 		case_rrr(mul, _d);
 		case_rrf(mul, _d, 64);
 		case_rrr(div, _d);
@@ -1097,15 +1155,19 @@ _emit_code(jit_state_t *_jit)
 		jmpr(rn(node->u.w));
 		break;
 	    case jit_code_jmpi:
-		temp = node->u.n;
-		assert(temp->code == jit_code_label ||
-		       temp->code == jit_code_epilog);
-		if (temp->flag & jit_flag_patch)
-		    jmpi(temp->u.w);
-		else {
-		    word = jmpi_p(_jit->pc.w);
-		    patch(word, node);
+		if (node->flag & jit_flag_node) {
+		    temp = node->u.n;
+		    assert(temp->code == jit_code_label ||
+			   temp->code == jit_code_epilog);
+		    if (temp->flag & jit_flag_patch)
+			jmpi(temp->u.w);
+		    else {
+			word = jmpi_p(_jit->pc.w);
+			patch(word, node);
+		    }
 		}
+		else
+		    jmpi(node->u.w);
 		break;
 	    case jit_code_callr:
 		callr(rn(node->u.w));
@@ -1177,6 +1239,7 @@ _emit_code(jit_state_t *_jit)
 	    }
 	}
 	jit_regarg_clr(node, value);
+	assert(_jitc->regarg == jit_carry == _NOREG ? 0 : (1 << jit_carry));
 	/* update register live state */
 	jit_reglive(node);
     }
@@ -1198,6 +1261,8 @@ _emit_code(jit_state_t *_jit)
 	patch_at(_jitc->patches.ptr[offset].inst, word);
     }
 
+    jit_flush(_jit->code.ptr, _jit->pc.uc);
+
     return (_jit->code.ptr);
 }
 
@@ -1206,6 +1271,11 @@ _emit_code(jit_state_t *_jit)
 #  include "jit_alpha-fpu.c"
 #undef CODE
 
+void
+jit_flush(void *fptr, void *tptr)
+{
+}
+
 void
 _emit_ldxi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
