@@ -152,6 +152,12 @@ static void _swf_negr_d(jit_state_t*,jit_int32_t,jit_int32_t);
 #  define swf_subi_f(r0,r1,i0)		swf_fff_(__aeabi_fsub,r0,r1,i0)
 #  define swf_subr_d(r0,r1,r2)		swf_ddd(__aeabi_dsub,r0,r1,r2)
 #  define swf_subi_d(r0,r1,i0)		swf_ddd_(__aeabi_dsub,r0,r1,i0)
+#  define swf_rsbr_f(r0, r1, r2)	swf_subr_f(r0, r2, r1)
+#  define swf_rsbi_f(r0, r1, i0)	_swf_rsbi_f(_jit, r0, r1, i0)
+static void _swf_rsbi_f(jit_state_t*,jit_int32_t,jit_int32_t,jit_float32_t);
+#  define swf_rsbr_d(r0, r1, r2)	swf_subr_d(r0, r2, r1)
+#  define swf_rsbi_d(r0, r1, i0)	_swf_rsbi_d(_jit, r0, r1, i0)
+static void _swf_rsbi_d(jit_state_t*,jit_int32_t,jit_int32_t,jit_float64_t);
 #  define swf_mulr_f(r0,r1,r2)		swf_fff(__aeabi_fmul,r0,r1,r2)
 #  define swf_muli_f(r0,r1,i0)		swf_fff_(__aeabi_fmul,r0,r1,i0)
 #  define swf_mulr_d(r0,r1,r2)		swf_ddd(__aeabi_dmul,r0,r1,r2)
@@ -690,6 +696,28 @@ _swf_fff_(jit_state_t *_jit, float (*i0)(float, float),
     jit_unget_reg_args();
 }
 
+static void
+_swf_rsbi_f(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_float32_t i0)
+{
+    union {
+	jit_int32_t	i;
+	jit_float32_t	f;
+    } data;
+    jit_get_reg_args();
+    data.f = i0;
+    movi(_R0_REGNO, data.i);
+    if (jit_fpr_p(r1))
+	swf_ldrin(_R1_REGNO, _FP_REGNO, swf_off(r1) + 8);
+    else
+	movr(_R1_REGNO, r1);
+    swf_call(__aeabi_fsub, fallback, _R3_REGNO);
+    if (jit_fpr_p(r0))
+	swf_strin(_R0_REGNO, _FP_REGNO, swf_off(r0) + 8);
+    else
+	movr(r0, _R0_REGNO);
+    jit_unget_reg_args();
+}
+
 static void
 _swf_ddd_(jit_state_t *_jit, double (*i0)(double, double),
 	  jit_int32_t r0, jit_int32_t r1, jit_float64_t i1)
@@ -699,6 +727,7 @@ _swf_ddd_(jit_state_t *_jit, double (*i0)(double, double),
 	jit_float64_t	d;
     } data;
     jit_get_reg_args();
+
     data.d = i1;
     if (jit_fpr_p(r1)) {
 	if (!jit_thumb_p() && jit_armv5e_p())
@@ -730,6 +759,45 @@ _swf_ddd_(jit_state_t *_jit, double (*i0)(double, double),
     jit_unget_reg_args();
 }
 
+static void
+_swf_rsbi_d(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_float64_t i0)
+{
+    union {
+	jit_int32_t	i[2];
+	jit_float64_t	d;
+    } data;
+    jit_get_reg_args();
+    data.d = i0;
+    movi(_R0_REGNO, data.i[0]);
+    movi(_R1_REGNO, data.i[1]);
+    if (jit_fpr_p(r1)) {
+	if (!jit_thumb_p() && jit_armv5e_p())
+	    LDRDIN(_R2_REGNO, _FP_REGNO, swf_off(r1) + 8);
+	else {
+	    swf_ldrin(_R2_REGNO, _FP_REGNO, swf_off(r1) + 8);
+	    swf_ldrin(_R3_REGNO, _FP_REGNO, swf_off(r1) + 4);
+	}
+    }
+    else {
+	movr(_R2_REGNO, r1);
+	movr(_R3_REGNO, r1 + 1);
+    }
+    swf_call_with_get_reg(__aeabi_dsub, fallback);
+    if (jit_fpr_p(r0)) {
+	if (!jit_thumb_p() && jit_armv5e_p())
+	    STRDIN(_R0_REGNO, _FP_REGNO, swf_off(r0) + 8);
+	else {
+	    swf_strin(_R0_REGNO, _FP_REGNO, swf_off(r0) + 8);
+	    swf_strin(_R1_REGNO, _FP_REGNO, swf_off(r0) + 4);
+	}
+    }
+    else {
+	movr(r0, _R0_REGNO);
+	movr(r0 + 1, _R1_REGNO);
+    }
+    jit_unget_reg_args();
+}
+
 static void
 _swf_iff(jit_state_t *_jit, int (*i0)(float, float),
 	 jit_int32_t r0, jit_int32_t r1, jit_int32_t r2)
@@ -1802,7 +1870,7 @@ _swf_movr_d(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 	    }
 	}
 	else if (jit_fpr_p(r0)) {
-	    if (!jit_thumb_p() && jit_armv5e_p())
+	    if (!jit_thumb_p() && jit_armv5e_p() && !(r1 & 1))
 		STRDIN(r1, _FP_REGNO, swf_off(r0) + 8);
 	    else {
 		swf_strin(r1, _FP_REGNO, swf_off(r0) + 8);
@@ -1899,9 +1967,9 @@ _swf_absr_d(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
     if (jit_fpr_p(r1)) {
 	if (jit_fpr_p(r0) && !jit_thumb_p() && jit_armv5e_p() &&
 	    r0 != r1 && (reg = jit_get_reg_pair()) != JIT_NOREG) {
-	    LDRDIN(rn(reg), _FP_REGNO, swf_off(r1) + 4);
-	    swf_bici(rn(reg), rn(reg), 0x80000000);
-	    STRDIN(rn(reg), _FP_REGNO, swf_off(r0) + 4);
+	    LDRDIN(rn(reg), _FP_REGNO, swf_off(r1) + 8);
+	    swf_bici(rn(reg) + 1, rn(reg) + 1, 0x80000000);
+	    STRDIN(rn(reg), _FP_REGNO, swf_off(r0) + 8);
 	    jit_unget_reg_pair(reg);
 	}
 	else {
@@ -1971,9 +2039,9 @@ _swf_negr_d(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
     if (jit_fpr_p(r1)) {
 	if (jit_fpr_p(r0) && !jit_thumb_p() && jit_armv5e_p() &&
 	    r0 != r1 && (reg = jit_get_reg_pair()) != JIT_NOREG) {
-	    LDRDIN(rn(reg), _FP_REGNO, swf_off(r1) + 4);
-	    EORI(rn(reg), rn(reg), encode_arm_immediate(0x80000000));
-	    STRDIN(rn(reg), _FP_REGNO, swf_off(r0) + 4);
+	    LDRDIN(rn(reg), _FP_REGNO, swf_off(r1) + 8);
+	    EORI(rn(reg) + 1, rn(reg) + 1, encode_arm_immediate(0x80000000));
+	    STRDIN(rn(reg), _FP_REGNO, swf_off(r0) + 8);
 	    jit_unget_reg_pair(reg);
 	}
 	else {
@@ -2129,7 +2197,7 @@ _swf_ldr_d(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 	    jit_unget_reg(reg);
 	}
     }
-    else if (!jit_thumb_p() && jit_armv5e_p())
+    else if (!jit_thumb_p() && jit_armv5e_p() && !(r0 & 1))
 	LDRDI(r0, r1, 0);
     else {
 	ldxi_i(r0, r1, 0);
@@ -2173,7 +2241,7 @@ _swf_ldi_d(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
 	    swf_strin(rn(rg0), _FP_REGNO, swf_off(r0) + 4);
 	    jit_unget_reg(rg0);
 	}
-	else if (!jit_thumb_p() && jit_armv5e_p())
+	else if (!jit_thumb_p() && jit_armv5e_p() && !(r0 & 1))
 	    LDRDI(r0, rn(rg1), 0);
 	else {
 	    ldxi_i(r0, rn(rg1), 0);
@@ -2221,7 +2289,7 @@ _swf_ldxr_d(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_int32_t r2)
 	}
     }
     else {
-	if (!jit_thumb_p() && jit_armv5e_p())
+	if (!jit_thumb_p() && jit_armv5e_p() && !(r0 & 1))
 	    LDRD(r0, r1, r2);
 	else {
 	    rg1 = jit_get_reg(jit_class_gpr);
@@ -2292,9 +2360,11 @@ _swf_ldxi_d(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 	}
     }
     else {
-	if (!jit_thumb_p() && jit_armv5e_p() && i0 >= 0 && i0 <= 255)
+	if (!jit_thumb_p() && jit_armv5e_p() &&
+	    i0 >= 0 && i0 <= 255 && !(r0 & 1))
 	    LDRDI(r0, r1, i0);
-	else if (!jit_thumb_p() && jit_armv5e_p() && i0 < 0 && i0 >= -255)
+	else if (!jit_thumb_p() && jit_armv5e_p() &&
+		 i0 < 0 && i0 >= -255 && !(r0 & 1))
 	    LDRDIN(r0, r1, -i0);
 	else if (i0 >= 0 && i0 + 4 <= 4095) {
 	    ldxi_i(r0, r1, i0);
@@ -2349,7 +2419,7 @@ _swf_str_d(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 	}
     }
     else {
-	if (!jit_thumb_p() && jit_armv5e_p())
+	if (!jit_thumb_p() && jit_armv5e_p() && !(r1 & 1))
 	    STRDI(r1, r0, 0);
 	else {
 	    stxi_i(0, r0, r1);
@@ -2401,7 +2471,7 @@ _swf_sti_d(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0)
     else {
 	rg1 = jit_get_reg(jit_class_gpr);
 	movi(rn(rg1), i0);
-	if (!jit_thumb_p() && jit_armv5e_p())
+	if (!jit_thumb_p() && jit_armv5e_p() && !(r0 & 1))
 	    STRDI(r0, rn(rg1), 0);
 	else {
 	    stxi_i(0, rn(rg1), r0);
@@ -2449,7 +2519,7 @@ _swf_stxr_d(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_int32_t r2)
 	}
     }
     else {
-	if (!jit_thumb_p() && jit_armv5e_p())
+	if (!jit_thumb_p() && jit_armv5e_p() && !(r2 & 1))
 	    STRD(r0, r1, r2);
 	else {
 	    rg1 = jit_get_reg(jit_class_gpr);
@@ -2488,6 +2558,7 @@ _swf_stxi_d(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0, jit_int32_t r1)
 		STRDI(rn(rg0), r0, i0);
 	    else
 		STRDIN(rn(rg0), r0, -i0);
+	    jit_unget_reg_pair(rg0);
 	}
 	else if (i0 >= 0 && i0 + 4 <= 4095) {
 	    rg0 = jit_get_reg(jit_class_gpr);
@@ -2519,9 +2590,11 @@ _swf_stxi_d(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0, jit_int32_t r1)
 	}
     }
     else {
-	if (!jit_thumb_p() && jit_armv5e_p() && i0 >= 0 && i0 <= 255)
+	if (!jit_thumb_p() && jit_armv5e_p() &&
+	    i0 >= 0 && i0 <= 255 && !(r1 & 1))
 	    STRDI(r1, r0, i0);
-	else if (!jit_thumb_p() && jit_armv5e_p() && i0 < 0 && i0 >= -255)
+	else if (!jit_thumb_p() && jit_armv5e_p() &&
+		 i0 < 0 && i0 >= -255 && !(r1 & 1))
 	    STRDIN(r1, r0, -i0);
 	else if (i0 >= 0 && i0 + 4 <= 4095) {
 	    stxi_i(i0, r0, r1);
