@@ -504,6 +504,12 @@ static void _vfp_subi_f(jit_state_t*,jit_int32_t,jit_int32_t,jit_float32_t);
 #  define vfp_subr_d(r0,r1,r2)		VSUB_F64(r0,r1,r2)
 #  define vfp_subi_d(r0,r1,i0)		_vfp_subi_d(_jit,r0,r1,i0)
 static void _vfp_subi_d(jit_state_t*,jit_int32_t,jit_int32_t,jit_float64_t);
+#  define vfp_rsbr_f(r0,r1,r2)		vfp_subr_f(r0,r2,r1)
+#  define vfp_rsbi_f(r0,r1,i0)		_vfp_rsbi_f(_jit,r0,r1,i0)
+static void _vfp_rsbi_f(jit_state_t*,jit_int32_t,jit_int32_t,jit_float32_t);
+#  define vfp_rsbr_d(r0,r1,r2)		vfp_subr_d(r0,r2,r1)
+#  define vfp_rsbi_d(r0,r1,i0)		_vfp_rsbi_d(_jit,r0,r1,i0)
+static void _vfp_rsbi_d(jit_state_t*,jit_int32_t,jit_int32_t,jit_float64_t);
 #  define vfp_mulr_f(r0,r1,r2)		VMUL_F32(r0,r1,r2)
 #  define vfp_muli_f(r0,r1,i0)		_vfp_muli_f(_jit,r0,r1,i0)
 static void _vfp_muli_f(jit_state_t*,jit_int32_t,jit_int32_t,jit_float32_t);
@@ -1245,8 +1251,11 @@ _vfp_movi_f(jit_state_t *_jit, jit_int32_t r0, jit_float32_t i0)
     jit_int32_t		code;
     u.f = i0;
     if (jit_fpr_p(r0)) {
-	if ((code = encode_vfp_double(1, 0, u.i, u.i)) != -1 ||
-	    (code = encode_vfp_double(1, 1, ~u.i, ~u.i)) != -1)
+	/* float arguments are packed, for others,
+	 * lightning only address even registers */
+	if (!(r0 & 1) && (r0 - 16) >= 0 &&
+	    ((code = encode_vfp_double(1, 0, u.i, u.i)) != -1 ||
+	     (code = encode_vfp_double(1, 1, ~u.i, ~u.i)) != -1))
 	    VIMM(code, r0);
 	else {
 	    reg = jit_get_reg(jit_class_gpr);
@@ -1455,6 +1464,8 @@ _vfp_b##name##i_d(jit_state_t *_jit,					\
 fopi(add)
 dopi(add)
 fopi(sub)
+fopi(rsb)
+dopi(rsb)
 dopi(sub)
 fopi(mul)
 dopi(mul)
