@@ -23,6 +23,9 @@
 
 #define rc(value)			jit_class_##value
 #define rn(reg)				(jit_regno(_rvs[jit_regno(reg)].spec))
+#define jit_arg_reg_p(i)		((i) >= 0 && (i) < 4)
+#define jit_arg_f_reg_p(i)		((i) >= 0 && (i) < 16)
+#define jit_arg_d_reg_p(i)		((i) >= 0 && (i) < 15)
 
 #define arm_patch_node			0x80000000
 #define arm_patch_word			0x40000000
@@ -55,6 +58,8 @@ typedef union _jit_thumb_t {
 static jit_int32_t _jit_get_reg_pair(jit_state_t*);
 #define jit_unget_reg_pair(rn)		_jit_unget_reg_pair(_jit,rn)
 static void _jit_unget_reg_pair(jit_state_t*,jit_int32_t);
+# define must_align_p(node)		_must_align_p(_jit, node)
+static jit_bool_t _must_align_p(jit_state_t*,jit_node_t*);
 #define load_const(uniq,r0,i0)		_load_const(_jit,uniq,r0,i0)
 static void _load_const(jit_state_t*,jit_bool_t,jit_int32_t,jit_word_t);
 #define flush_consts()			_flush_consts(_jit)
@@ -361,12 +366,29 @@ _jit_epilog(jit_state_t *_jit)
     _jitc->function = NULL;
 }
 
+jit_bool_t
+_jit_arg_register_p(jit_state_t *_jit, jit_node_t *u)
+{
+    if (u->code != jit_code_arg) {
+	if (u->code == jit_code_arg_f) {
+	    if (jit_cpu.abi)
+		return (jit_arg_f_reg_p(u->u.w));
+	}
+	else {
+	    assert(u->code == jit_code_arg_d);
+	    if (jit_cpu.abi)
+		return (jit_arg_d_reg_p(u->u.w));
+	}
+    }
+    return (jit_arg_reg_p(u->u.w));
+}
+
 jit_node_t *
 _jit_arg(jit_state_t *_jit)
 {
     jit_int32_t		offset;
     assert(_jitc->function);
-    if (_jitc->function->self.argi < 4)
+    if (jit_arg_reg_p(_jitc->function->self.argi))
 	offset = _jitc->function->self.argi++;
     else {
 	offset = _jitc->function->self.size;
@@ -375,25 +397,19 @@ _jit_arg(jit_state_t *_jit)
     return (jit_new_node_w(jit_code_arg, offset));
 }
 
-jit_bool_t
-_jit_arg_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (offset >= 0 && offset < 4);
-}
-
 jit_node_t *
 _jit_arg_f(jit_state_t *_jit)
 {
     jit_int32_t		offset;
     assert(_jitc->function);
     if (jit_cpu.abi && !(_jitc->function->self.call & jit_call_varargs)) {
-	if (_jitc->function->self.argf < 16) {
+	if (jit_arg_f_reg_p(_jitc->function->self.argf)) {
 	    offset = _jitc->function->self.argf++;
 	    goto done;
 	}
     }
     else {
-	if (_jitc->function->self.argi < 4) {
+	if (jit_arg_reg_p(_jitc->function->self.argi)) {
 	    offset = _jitc->function->self.argi++;
 	    goto done;
 	}
@@ -404,19 +420,13 @@ done:
     return (jit_new_node_w(jit_code_arg_f, offset));
 }
 
-jit_bool_t
-_jit_arg_f_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (jit_arg_reg_p(offset));
-}
-
 jit_node_t *
 _jit_arg_d(jit_state_t *_jit)
 {
     jit_int32_t		offset;
     assert(_jitc->function);
     if (jit_cpu.abi && !(_jitc->function->self.call & jit_call_varargs)) {
-	if (_jitc->function->self.argf < 15) {
+	if (jit_arg_d_reg_p(_jitc->function->self.argf)) {
 	    if (_jitc->function->self.argf & 1)
 		++_jitc->function->self.argf;
 	    offset = _jitc->function->self.argf;
@@ -425,9 +435,9 @@ _jit_arg_d(jit_state_t *_jit)
 	}
     }
     else {
-	if (_jitc->function->self.argi < 3) {
-	    if (_jitc->function->self.argi & 1)
-		++_jitc->function->self.argi;
+	if (_jitc->function->self.argi & 1)
+	    ++_jitc->function->self.argi;
+	if (jit_arg_reg_p(_jitc->function->self.argi)) {
 	    offset = _jitc->function->self.argi;
 	    _jitc->function->self.argi += 2;
 	    goto done;
@@ -441,18 +451,13 @@ done:
     return (jit_new_node_w(jit_code_arg_d, offset));
 }
 
-jit_bool_t
-_jit_arg_d_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (jit_arg_reg_p(offset));
-}
-
 void
 _jit_getarg_c(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg);
     if (jit_swf_p())
 	jit_ldxi_c(u, JIT_FP, v->u.w < 4 ? v->u.w << 2 : v->u.w);
-    else if (v->u.w < 4)
+    else if (jit_arg_reg_p(v->u.w))
 	jit_extr_c(u, JIT_RA0 - v->u.w);
     else
 	jit_ldxi_c(u, JIT_FP, v->u.w);
@@ -461,9 +466,10 @@ _jit_getarg_c(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 void
 _jit_getarg_uc(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg);
     if (jit_swf_p())
 	jit_ldxi_uc(u, JIT_FP, v->u.w < 4 ? v->u.w << 2 : v->u.w);
-    else if (v->u.w < 4)
+    else if (jit_arg_reg_p(v->u.w))
 	jit_extr_uc(u, JIT_RA0 - v->u.w);
     else
 	jit_ldxi_uc(u, JIT_FP, v->u.w);
@@ -472,9 +478,10 @@ _jit_getarg_uc(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 void
 _jit_getarg_s(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg);
     if (jit_swf_p())
 	jit_ldxi_s(u, JIT_FP, v->u.w < 4 ? v->u.w << 2 : v->u.w);
-    else if (v->u.w < 4)
+    else if (jit_arg_reg_p(v->u.w))
 	jit_extr_s(u, JIT_RA0 - v->u.w);
     else
 	jit_ldxi_s(u, JIT_FP, v->u.w);
@@ -483,9 +490,10 @@ _jit_getarg_s(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 void
 _jit_getarg_us(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg);
     if (jit_swf_p())
 	jit_ldxi_us(u, JIT_FP, v->u.w < 4 ? v->u.w << 2 : v->u.w);
-    else if (v->u.w < 4)
+    else if (jit_arg_reg_p(v->u.w))
 	jit_extr_us(u, JIT_RA0 - v->u.w);
     else
 	jit_ldxi_us(u, JIT_FP, v->u.w);
@@ -494,19 +502,54 @@ _jit_getarg_us(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 void
 _jit_getarg_i(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg);
     if (jit_swf_p())
 	jit_ldxi_i(u, JIT_FP, v->u.w < 4 ? v->u.w << 2 : v->u.w);
-    else if (v->u.w < 4)
+    else if (jit_arg_reg_p(v->u.w))
 	jit_movr(u, JIT_RA0 - v->u.w);
     else
 	jit_ldxi_i(u, JIT_FP, v->u.w);
 }
 
+void
+_jit_putargr(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg);
+    if (jit_swf_p())
+	jit_stxi(v->u.w < 4 ? v->u.w << 2 : v->u.w, JIT_FP, u);
+    else if (jit_arg_reg_p(v->u.w))
+	jit_movr(JIT_RA0 - v->u.w, u);
+    else
+	jit_stxi(v->u.w, JIT_FP, u);
+}
+
+void
+_jit_putargi(jit_state_t *_jit, jit_word_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg);
+    if (jit_swf_p()) {
+	regno = jit_get_reg(jit_class_gpr);
+	jit_movi(regno, u);
+	jit_stxi(v->u.w < 4 ? v->u.w << 2 : v->u.w, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+    else if (jit_arg_reg_p(v->u.w))
+	jit_movi(JIT_RA0 - v->u.w, u);
+    else {
+	regno = jit_get_reg(jit_class_gpr);
+	jit_movi(regno, u);
+	jit_stxi(v->u.w, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_getarg_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg_f);
     if (jit_cpu.abi) {
-	if (v->u.w < 16)
+	if (jit_arg_f_reg_p(v->u.w))
 	    jit_movr_f(u, JIT_FA0 - v->u.w);
 	else
 	    jit_ldxi_f(u, JIT_FP, v->u.w);
@@ -514,18 +557,71 @@ _jit_getarg_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
     else if (jit_swf_p())
 	jit_ldxi_f(u, JIT_FP, v->u.w < 4 ? v->u.w << 2 : v->u.w);
     else {
-	if (v->u.w < 4)
+	if (jit_arg_reg_p(v->u.w))
 	    jit_movr_w_f(u, JIT_RA0 - v->u.w);
 	else
 	    jit_ldxi_f(u, JIT_FP, v->u.w);
     }
 }
 
+void
+_jit_putargr_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg_f);
+    if (jit_cpu.abi) {
+	if (jit_arg_f_reg_p(v->u.w))
+	    jit_movr_f(JIT_FA0 - v->u.w, u);
+	else
+	    jit_stxi_f(v->u.w, JIT_FP, u);
+    }
+    else if (jit_swf_p())
+	jit_stxi_f(v->u.w < 4 ? v->u.w << 2 : v->u.w, JIT_FP, u);
+    else {
+	if (jit_arg_reg_p(v->u.w))
+	    jit_movr_f_w(JIT_RA0 - v->u.w, u);
+	else
+	    jit_stxi_f(v->u.w, JIT_FP, u);
+    }
+}
+
+void
+_jit_putargi_f(jit_state_t *_jit, jit_float32_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg_f);
+    if (jit_cpu.abi) {
+	if (jit_arg_f_reg_p(v->u.w))
+	    jit_movi_f(JIT_FA0 - v->u.w, u);
+	else {
+	    regno = jit_get_reg(jit_class_fpr);
+	    jit_movi_f(regno, u);
+	    jit_stxi_f(v->u.w, JIT_FP, regno);
+	    jit_unget_reg(regno);
+	}
+    }
+    else if (jit_swf_p()) {
+	regno = jit_get_reg(jit_class_fpr);
+	jit_movi_f(regno, u);
+	jit_stxi_f(v->u.w < 4 ? v->u.w << 2 : v->u.w, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+    else {
+	regno = jit_get_reg(jit_class_fpr);
+	jit_movi_f(regno, u);
+	if (jit_arg_reg_p(v->u.w))
+	    jit_movr_f_w(JIT_RA0 - v->u.w, regno);
+	else
+	    jit_stxi_f(v->u.w, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_getarg_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg_d);
     if (jit_cpu.abi) {
-	if (v->u.w < 16)
+	if (jit_arg_f_reg_p(v->u.w))
 	    jit_movr_d(u, JIT_FA0 - v->u.w);
 	else
 	    jit_ldxi_d(u, JIT_FP, v->u.w);
@@ -533,18 +629,70 @@ _jit_getarg_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
     else if (jit_swf_p())
 	jit_ldxi_d(u, JIT_FP, v->u.w < 4 ? v->u.w << 2 : v->u.w);
     else {
-	if (v->u.w < 4)
+	if (jit_arg_reg_p(v->u.w))
 	    jit_movr_ww_d(u, JIT_RA0 - v->u.w, JIT_RA0 - (v->u.w + 1));
 	else
 	    jit_ldxi_d(u, JIT_FP, v->u.w);
     }
 }
 
+void
+_jit_putargr_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg_d);
+    if (jit_cpu.abi) {
+	if (jit_arg_f_reg_p(v->u.w))
+	    jit_movr_d(JIT_FA0 - v->u.w, u);
+	else
+	    jit_stxi_d(v->u.w, JIT_FP, u);
+    }
+    else if (jit_swf_p())
+	jit_stxi_d(v->u.w < 4 ? v->u.w << 2 : v->u.w, JIT_FP, u);
+    else {
+	if (jit_arg_reg_p(v->u.w))
+	    jit_movr_d_ww(JIT_RA0 - v->u.w, JIT_RA0 - (v->u.w + 1), u);
+	else
+	    jit_stxi_d(v->u.w, JIT_FP, u);
+    }
+}
+
+void
+_jit_putargi_d(jit_state_t *_jit, jit_float64_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg_d);
+    if (jit_cpu.abi) {
+	if (jit_arg_f_reg_p(v->u.w))
+	    jit_movi_d(JIT_FA0 - v->u.w, u);
+	else {
+	    regno = jit_get_reg(jit_class_fpr);
+	    jit_movi_d(regno, u);
+	    jit_stxi_d(v->u.w, JIT_FP, regno);
+	    jit_unget_reg(regno);
+	}
+    }
+    else if (jit_swf_p()) {
+	regno = jit_get_reg(jit_class_fpr);
+	jit_movi_d(regno, u);
+	jit_stxi_d(v->u.w < 4 ? v->u.w << 2 : v->u.w, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+    else {
+	regno = jit_get_reg(jit_class_fpr);
+	jit_movi_d(regno, u);
+	if (jit_arg_reg_p(v->u.w))
+	    jit_movr_d_ww(JIT_RA0 - v->u.w, JIT_RA0 - (v->u.w + 1), regno);
+	else
+	    jit_stxi_d(v->u.w, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_pushargr(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
-    if (_jitc->function->call.argi < 4) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movr(JIT_RA0 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
@@ -559,7 +707,7 @@ _jit_pushargi(jit_state_t *_jit, jit_word_t u)
 {
     jit_int32_t		 regno;
     assert(_jitc->function);
-    if (_jitc->function->call.argi < 4) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movi(JIT_RA0 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
@@ -577,14 +725,14 @@ _jit_pushargr_f(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
     if (jit_cpu.abi && !(_jitc->function->call.call & jit_call_varargs)) {
-	if (_jitc->function->call.argf < 16) {
+	if (jit_arg_f_reg_p(_jitc->function->call.argf)) {
 	    jit_movr_f(JIT_FA0 - _jitc->function->call.argf, u);
 	    ++_jitc->function->call.argf;
 	    return;
 	}
     }
     else {
-	if (_jitc->function->call.argi < 4) {
+	if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	    jit_movr_f_w(JIT_RA0 - _jitc->function->call.argi, u);
 	    ++_jitc->function->call.argi;
 	    return;
@@ -600,7 +748,7 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
     jit_int32_t		regno;
     assert(_jitc->function);
     if (jit_cpu.abi && !(_jitc->function->call.call & jit_call_varargs)) {
-	if (_jitc->function->call.argf < 16) {
+	if (jit_arg_f_reg_p(_jitc->function->call.argf)) {
 	    /* cannot jit_movi_f in the argument register because
 	     * float arguments are packed, and that would cause
 	     * either an assertion in debug mode, or overwritting
@@ -614,7 +762,7 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
 	}
     }
     else {
-	if (_jitc->function->call.argi < 4) {
+	if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	    jit_movi_f_w(JIT_RA0 - _jitc->function->call.argi, u);
 	    ++_jitc->function->call.argi;
 	    return;
@@ -632,7 +780,7 @@ _jit_pushargr_d(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
     if (jit_cpu.abi && !(_jitc->function->call.call & jit_call_varargs)) {
-	if (_jitc->function->call.argf < 15) {
+	if (jit_arg_d_reg_p(_jitc->function->call.argf)) {
 	    if (_jitc->function->call.argf & 1)
 		++_jitc->function->call.argf;
 	    jit_movr_d(JIT_FA0 - _jitc->function->call.argf, u);
@@ -643,7 +791,7 @@ _jit_pushargr_d(jit_state_t *_jit, jit_int32_t u)
     else {
 	if (_jitc->function->call.argi & 1)
 	    ++_jitc->function->call.argi;
-	if (_jitc->function->call.argi < 4) {
+	if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	    jit_movr_d_ww(JIT_RA0 - _jitc->function->call.argi,
 			  JIT_RA0 - (_jitc->function->call.argi + 1),
 			  u);
@@ -663,7 +811,7 @@ _jit_pushargi_d(jit_state_t *_jit, jit_float64_t u)
     jit_int32_t		regno;
     assert(_jitc->function);
     if (jit_cpu.abi && !(_jitc->function->call.call & jit_call_varargs)) {
-	if (_jitc->function->call.argf < 15) {
+	if (jit_arg_d_reg_p(_jitc->function->call.argf)) {
 	    if (_jitc->function->call.argf & 1)
 		++_jitc->function->call.argf;
 	    jit_movi_d(JIT_FA0 - _jitc->function->call.argf, u);
@@ -674,7 +822,7 @@ _jit_pushargi_d(jit_state_t *_jit, jit_float64_t u)
     else {
 	if (_jitc->function->call.argi & 1)
 	    ++_jitc->function->call.argi;
-	if (_jitc->function->call.argi < 4) {
+	if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	    jit_movi_d_ww(JIT_RA0 - _jitc->function->call.argi,
 			  JIT_RA0 - (_jitc->function->call.argi + 1),
 			  u);
@@ -1038,13 +1186,27 @@ _emit_code(jit_state_t *_jit)
 	if (_jit->pc.uc >= _jitc->code.end)
 	    return (NULL);
 
+#if DEVEL_DISASSEMBLER
+	node->offset = _jit->pc.w;
+#endif
 	value = jit_classify(node->code);
 	jit_regarg_set(node, value);
 	switch (node->code) {
+	    case jit_code_align:
+		assert(!(node->u.w & (node->u.w - 1)) &&
+		       node->u.w <= sizeof(jit_word_t));
+		if (node->u.w == sizeof(jit_word_t) &&
+		    (word = _jit->pc.w & (sizeof(jit_word_t) - 1)))
+		    nop(sizeof(jit_word_t) - word);
+		break;
 	    case jit_code_note:		case jit_code_name:
+		if (must_align_p(node->next))
+		    nop(2);
 		node->u.w = _jit->pc.w;
 		break;
 	    case jit_code_label:
+		if (must_align_p(node->next))
+		    nop(2);
 		/* remember label is defined */
 		node->flag |= jit_flag_patch;
 		node->u.w = _jit->pc.w;
@@ -1061,6 +1223,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrw(subc,);
 		case_rrr(subx,);
 		case_rrw(subx,);
+		case_rrw(rsb,);
 		case_rrr(mul,);
 		case_rrw(mul,);
 		case_rrrr(qmul,);
@@ -1127,7 +1290,8 @@ _emit_code(jit_state_t *_jit)
 		case_wrr(stx, _s);
 		case_rrr(stx, _i);
 		case_wrr(stx, _i);
-		case_rr(hton,);
+		case_rr(hton, _us);
+		case_rr(hton, _ui);
 		case_rr(ext, _c);
 		case_rr(ext, _uc);
 		case_rr(ext, _s);
@@ -1214,6 +1378,7 @@ _emit_code(jit_state_t *_jit)
 		case_vvf(add);
 		case_vvv(sub, _f);
 		case_vvf(sub);
+		case_vvf(rsb);
 		case_vvv(mul, _f);
 		case_vvf(mul);
 		case_vvv(div, _f);
@@ -1299,6 +1464,7 @@ _emit_code(jit_state_t *_jit)
 		case_vvd(add);
 		case_vvv(sub, _d);
 		case_vvd(sub);
+		case_vvd(rsb);
 		case_vvv(mul, _d);
 		case_vvd(mul);
 		case_vvv(div, _d);
@@ -1385,15 +1551,19 @@ _emit_code(jit_state_t *_jit)
 		flush_consts();
 		break;
 	    case jit_code_jmpi:
-		temp = node->u.n;
-		assert(temp->code == jit_code_label ||
-		       temp->code == jit_code_epilog);
-		if (temp->flag & jit_flag_patch)
-		    jmpi(temp->u.w);
-		else {
-		    word = jmpi_p(_jit->pc.w);
-		    patch(word, node);
+		if (node->flag & jit_flag_node) {
+		    temp = node->u.n;
+		    assert(temp->code == jit_code_label ||
+			   temp->code == jit_code_epilog);
+		    if (temp->flag & jit_flag_patch)
+			jmpi(temp->u.w);
+		    else {
+			word = jmpi_p(_jit->pc.w, 1);
+			patch(word, node);
+		    }
 		}
+		else
+		    jmpi(node->u.w);
 		flush_consts();
 		break;
 	    case jit_code_callr:
@@ -1506,10 +1676,14 @@ _emit_code(jit_state_t *_jit)
 		abort();
 	}
 	jit_regarg_clr(node, value);
+	assert(_jitc->regarg == 0);
 	/* update register live state */
 	jit_reglive(node);
 
-	if (_jitc->consts.length && _jit->pc.uc - _jitc->consts.data >= 3968) {
+	if (_jitc->consts.length &&
+	    (_jit->pc.uc - _jitc->consts.data >= 3968 ||
+	     (jit_uword_t)_jit->pc.uc -
+	     (jit_uword_t)_jitc->consts.patches[0] >= 3968)) {
 	    /* longest sequence should be 64 bytes, but preventively
 	     * do not let it go past 128 remaining bytes before a flush */
 	    if (node->next &&
@@ -1568,10 +1742,7 @@ _emit_code(jit_state_t *_jit)
 	patch_at(_jitc->patches.ptr[offset].kind & ~arm_patch_node, word, value);
     }
 
-#if defined(__GNUC__)
-    word = sysconf(_SC_PAGE_SIZE);
-    __clear_cache(_jit->code.ptr, (void *)((_jit->pc.w + word) & -word));
-#endif
+    jit_flush(_jit->code.ptr, _jit->pc.uc);
 
     return (_jit->code.ptr);
 }
@@ -1582,6 +1753,20 @@ _emit_code(jit_state_t *_jit)
 #  include "jit_arm-vfp.c"
 #undef CODE
 
+void
+jit_flush(void *fptr, void *tptr)
+{
+#if defined(__GNUC__)
+    jit_uword_t		i, f, t, s;
+
+    s = sysconf(_SC_PAGE_SIZE);
+    f = (jit_uword_t)fptr & -s;
+    t = (((jit_uword_t)tptr) + s - 1) & -s;
+    for (i = f; i < t; i += s)
+	__clear_cache((void *)i, (void *)(i + s));
+#endif
+}
+
 void
 _emit_ldxi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
@@ -1661,6 +1846,30 @@ _jit_unget_reg_pair(jit_state_t *_jit, jit_int32_t reg)
     }
 }
 
+/*   A prolog must be aligned at mod 4 bytes boundary.
+ *   This condition was not being required to be tested by
+ * accident previously, but with the jit_frame and jit_tramp
+ * code it is required */
+static jit_bool_t
+_must_align_p(jit_state_t *_jit, jit_node_t *node)
+{
+    if (jit_thumb_p() && (_jit->pc.w & 3)) {
+	for (; node; node = node->next) {
+	    switch (node->code) {
+		case jit_code_note:
+		case jit_code_name:
+		case jit_code_label:
+		    break;
+		case jit_code_prolog:
+		    return (1);
+		default:
+		    return (0);
+	    }
+	}
+    }
+    return (0);
+}
+
 static void
 _load_const(jit_state_t *_jit, jit_bool_t uniq, jit_int32_t r0, jit_word_t i0)
 {
@@ -1796,7 +2005,8 @@ _patch(jit_state_t *_jit, jit_word_t instr, jit_node_t *node)
     }
     else {
 	flag = node->u.n->flag;
-	if (node->code == jit_code_calli)
+	if (node->code == jit_code_calli ||
+	    (node->code == jit_code_jmpi && !(node->flag & jit_flag_node)))
 	    kind = arm_patch_word;
 	else
 	    kind = arm_patch_jump;
