@@ -57,13 +57,20 @@ static jit_state_t		 *disasm_jit;
  * Implementation
  */
 void
-jit_init_debug(char *progname)
+jit_init_debug(const char *progname)
 {
 #if DISASSEMBLER
     bfd_init();
 
-    disasm_bfd = bfd_openr(progname, NULL);
-    assert(disasm_bfd);
+    if (progname)
+	disasm_bfd = bfd_openr(progname, NULL);
+    if (disasm_bfd == NULL) {
+#if defined(__linux__)
+	disasm_bfd = bfd_openr("/proc/self/exe", NULL);
+	if (disasm_bfd == NULL)
+#endif
+	    return;
+    }
     bfd_check_format(disasm_bfd, bfd_object);
     bfd_check_format(disasm_bfd, bfd_archive);
     disasm_print = disassembler(disasm_bfd);
@@ -72,32 +79,42 @@ jit_init_debug(char *progname)
 #  if defined(__i386__) || defined(__x86_64__)
     disasm_info.arch = bfd_arch_i386;
 #    if defined(__x86_64__)
+#      if __WORDSIZE == 32
+    disasm_info.mach = bfd_mach_x64_32;
+#      else
     disasm_info.mach = bfd_mach_x86_64;
+#      endif
 #    else
     disasm_info.mach = bfd_mach_i386_i386;
 #    endif
 #  endif
-#  if defined(__arm__)
-    /* FIXME add mapping for prolog switching to arm and possible jump
-     * before first prolog also in arm mode */
-    if (jit_cpu.thumb)
-	disasm_info.disassembler_options = "force-thumb";
-#  endif
 #  if defined(__powerpc__)
     disasm_info.arch = bfd_arch_powerpc;
     disasm_info.mach = bfd_mach_ppc64;
+#    if HAVE_DISASSEMBLE_INIT_FOR_TARGET
+    disassemble_init_for_target(&disasm_info);
+#    elif HAVE_DISASSEMBLE_INIT_POWERPC
     disassemble_init_powerpc(&disasm_info);
+#    endif
 #    if defined(__powerpc64__)
     disasm_info.disassembler_options = "64";
 #    endif
+#    if HAVE_DISASSEMBLE_INIT_FOR_TARGET
+    disassemble_init_for_target(&disasm_info);
+#    elif HAVE_DISASSEMBLE_INIT_POWERPC
     disassemble_init_powerpc(&disasm_info);
+#    endif
 #  endif
 #  if defined(__sparc__)
     disasm_info.endian = disasm_info.display_endian = BFD_ENDIAN_BIG;
 #  endif
-#  if defined(__s390x__)
+#  if defined(__s390__) || defined(__s390x__)
     disasm_info.arch = bfd_arch_s390;
+#    if __WORDSIZE == 32
+    disasm_info.mach = bfd_mach_s390_31;
+#    else
     disasm_info.mach = bfd_mach_s390_64;
+#    endif
     disasm_info.endian = disasm_info.display_endian = BFD_ENDIAN_BIG;
     disasm_info.disassembler_options = "zarch";
 #  endif
@@ -197,7 +214,15 @@ void
 _jit_disassemble(jit_state_t *_jit)
 {
 #if DISASSEMBLER
-    disassemble(_jit->code.ptr, _jit->pc.uc - _jit->code.ptr);
+    if (disasm_bfd) {
+#  if defined(__arm__)
+	/* FIXME add mapping for prolog switching to arm and possible jump
+	 * before first prolog also in arm mode */
+	disasm_info.disassembler_options = jit_cpu.thumb ? "force-thumb" : "";
+#  endif
+
+	disassemble(_jit->code.ptr, _jit->pc.uc - _jit->code.ptr);
+    }
 #endif
 }
 
@@ -302,6 +327,9 @@ _disassemble(jit_state_t *_jit, jit_pointer_t code, jit_int32_t length)
     bfd_vma		 pc = (jit_uword_t)code;
     bfd_vma		 end = (jit_uword_t)code + length;
     char		 buffer[address_buffer_length];
+#if DEVEL_DISASSEMBLER
+    jit_node_t		*node;
+#endif
 
 #if __arm__
     data_info = _jitc && _jitc->data_info.ptr;
@@ -313,13 +341,24 @@ _disassemble(jit_state_t *_jit, jit_pointer_t code, jit_int32_t length)
     old_file = old_name = NULL;
     old_line = 0;
     disasm_jit = _jit;
+#if DEVEL_DISASSEMBLER
+    node = _jitc->head;
+#endif
     while (pc < end) {
+#if DEVEL_DISASSEMBLER
+	while (node && node->offset < (jit_uword_t)pc)
+	    node = node->next;
+	while (node && node->offset == (jit_uword_t)pc) {
+	    jit_print_node(node);
+	    fputc('\n', stdout); 
+	    node = node->next;
+	}
+#endif
 #if __arm__
     again:
 	if (data_info) {
 	    while (_jitc->data_info.ptr[data_offset].code < pc) {
-		data_offset += 2;
-		if (data_offset >= _jitc->data_info.length) {
+		if (++data_offset >= _jitc->data_info.length) {
 		    data_info = 0;
 		    goto again;
 		}
