@@ -640,6 +640,8 @@ static void _f38(jit_state_t*,jit_int32_t,
 #define UADDCM(r1,r2,t)		f8(0x2,r2,r1,ADD_CF_NV,2,0,1,2,0,t)
 #define UXOR(r1,r2,t)		f8(0x2,r2,r1,LOG_CC_NV,0,1,1,2,0,t)
 #define XOR(r1,r2,t)		f8(0x2,r2,r1,LOG_CC_NV,0,1,0,2,0,t)
+#  define nop(c)			_nop(_jit,c)
+static void _nop(jit_state_t*,jit_int32_t);
 #define movr(r0,r1)		_movr(_jit,r0,r1)
 static void _movr(jit_state_t*,jit_int32_t,jit_int32_t);
 #define movi(r0,i0)		_movi(_jit,r0,i0)
@@ -653,7 +655,8 @@ static jit_word_t _movi_p(jit_state_t*,jit_int32_t,jit_word_t);
 #define extr_s(r0,r1)		EXTRWR(r1,31,16,r0)
 #define extr_us(r0,r1)		EXTRWR_U(r1,31,16,r0)
 #if __BYTE_ORDER == __BIG_ENDIAN
-#  define htonr(r0,r1)		movr(r0,r1)
+#  define htonr_us(r0,r1)	extr_us(r0,r1)
+#  define htonr_ui(r0,r1)	movr(r0,r1)
 #else
 #  error need htonr implementation
 #endif
@@ -675,6 +678,8 @@ static void _subci(jit_state_t*,jit_int32_t,jit_int32_t,jit_word_t);
 #define subxr(r0,r1,r2)		SUB_B(r1,r2,r0)
 #define subxi(r0,r1,i0)		_subxi(_jit,r0,r1,i0)
 static void _subxi(jit_state_t*,jit_int32_t,jit_int32_t,jit_word_t);
+#define rsbi(r0, r1, i0)	_rsbi(_jit, r0, r1, i0)
+static void _rsbi(jit_state_t*,jit_int32_t,jit_int32_t,jit_word_t);
 #define mulr(r0,r1,r2)		_mulr(_jit,r0,r1,r2)
 static void _mulr(jit_state_t*,jit_int32_t,jit_int32_t,jit_int32_t);
 #define muli(r0,r1,i0)		_muli(_jit,r0,r1,i0)
@@ -1585,6 +1590,14 @@ _f38(jit_state_t *_jit, jit_int32_t o,
     ii((o<<26)|((s>>5)<<9)|(u<<6)|(n<<1)|(s&0x1f));
 }
 
+static void
+_nop(jit_state_t *_jit, jit_int32_t i0)
+{
+    for (; i0 > 0; i0 -= 4)
+	NOP();
+    assert(i0 == 0);
+}
+
 static void
 _movr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
@@ -1619,7 +1632,7 @@ static void
 _addi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
     jit_int32_t		reg;
-    if (i0 >= -2048 && i0 <= 2047)
+    if (i0 >= -1024 && i0 <= 1023)
 	ADDI(i0, r1, r0);
     else {
 	reg = jit_get_reg(jit_class_gpr);
@@ -1679,6 +1692,20 @@ _subxi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
     jit_unget_reg(reg);
 }
 
+static void
+_rsbi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
+{
+    jit_int32_t		reg;
+    if (i0 >= -1024 && i0 <= 1023)
+	SUBI(i0, r1, r0);
+    else {
+	reg = jit_get_reg(jit_class_gpr);
+	movi(rn(reg), i0);
+	subr(r0, rn(reg), r1);
+	jit_unget_reg(reg);
+    }
+}
+
 static void
 _mulr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_int32_t r2)
 {
@@ -2621,6 +2648,13 @@ _prolog(jit_state_t *_jit, jit_node_t *node)
     jit_int32_t		regno;
     jit_word_t		offset;
 
+    if (_jitc->function->define_frame || _jitc->function->assume_frame) {
+	/* hppa stack grows up */
+	assert(_jitc->function->self.aoff <= _jitc->function->frame);
+	if (_jitc->function->assume_frame)
+	    return;
+	_jitc->function->self.aoff = _jitc->function->frame;
+    }
     _jitc->function->stack = ((_jitc->function->self.aoff -
 			       _jitc->function->self.alen -
 			       _jitc->function->self.size) + 63) & -64;
@@ -2651,6 +2685,8 @@ _epilog(jit_state_t *_jit, jit_node_t *node)
     jit_int32_t		regno;
     jit_word_t		offset;
 
+    if (_jitc->function->assume_frame)
+	return;
     /* Restore any modified callee save registers */
     offset = alloca_offset - 140;
     for (regno = 0; regno < jit_size(gr); regno++, offset += 4) {
