@@ -301,6 +301,12 @@ static void _subi_f(jit_state_t*,jit_int32_t,jit_int32_t,jit_float32_t*);
 #define subr_d(r0,r1,r2)		FSUB_D(r1,r2,r0)
 #define subi_d(r0,r1,i0)		_subi_d(_jit,r0,r1,i0)
 static void _subi_d(jit_state_t*,jit_int32_t,jit_int32_t,jit_float64_t*);
+#define rsbr_f(r0,r1,r2)		subr_f(r0,r2,r1)
+#define rsbi_f(r0,r1,i0)		_rsbi_f(_jit,r0,r1,i0)
+static void _rsbi_f(jit_state_t*,jit_int32_t,jit_int32_t,jit_float32_t*);
+#define rsbr_d(r0,r1,r2)		subr_d(r0,r2,r1)
+#define rsbi_d(r0,r1,i0)		_rsbi_d(_jit,r0,r1,i0)
+static void _rsbi_d(jit_state_t*,jit_int32_t,jit_int32_t,jit_float64_t*);
 #define mulr_f(r0,r1,r2)		FMPY_S(r1,r2,r0)
 #define muli_f(r0,r1,i0)		_muli_f(_jit,r0,r1,i0)
 static void _muli_f(jit_state_t*,jit_int32_t,jit_int32_t,jit_float32_t*);
@@ -741,6 +747,8 @@ fopi(add)
 dopi(add)
 fopi(sub)
 dopi(sub)
+fopi(rsb)
+dopi(rsb)
 fopi(mul)
 dopi(mul)
 fopi(div)
