@@ -22,6 +22,7 @@
 
 #define rc(value)			jit_class_##value
 #define rn(reg)				(jit_regno(_rvs[jit_regno(reg)].spec))
+#define jit_arg_reg_p(i)		(i >= 0 && i < 4)
 
 #define PROTO				1
 #  include "jit_hppa-cpu.c"
@@ -258,13 +259,21 @@ _jit_epilog(jit_state_t *_jit)
     _jitc->function = NULL;
 }
 
+jit_bool_t
+_jit_arg_register_p(jit_state_t *_jit, jit_node_t *u)
+{
+    assert(u->code == jit_code_arg ||
+	   u->code == jit_code_arg_f || u->code == jit_code_arg_d);
+    return (jit_arg_reg_p(u->u.w));
+}
+
 jit_node_t *
 _jit_arg(jit_state_t *_jit)
 {
     jit_int32_t		offset;
     assert(_jitc->function);
     _jitc->function->self.size -= sizeof(jit_word_t);
-    if (_jitc->function->self.argi < 4)
+    if (jit_arg_reg_p(_jitc->function->self.argi))
 	offset = _jitc->function->self.argi++;
     else
 	offset = _jitc->function->self.size;
@@ -277,7 +286,7 @@ _jit_arg_f(jit_state_t *_jit)
     jit_int32_t		offset;
     assert(_jitc->function);
     _jitc->function->self.size -= sizeof(jit_word_t);
-    if (_jitc->function->self.argi < 4)
+    if (jit_arg_reg_p(_jitc->function->self.argi))
 	offset = _jitc->function->self.argi++;
     else
 	offset = _jitc->function->self.size;
@@ -294,7 +303,7 @@ _jit_arg_d(jit_state_t *_jit)
 	_jitc->function->self.size -= sizeof(jit_word_t);
     }
     _jitc->function->self.size -= sizeof(jit_float64_t);
-    if (_jitc->function->self.argi < 4) {
+    if (jit_arg_reg_p(_jitc->function->self.argi)) {
 	offset = _jitc->function->self.argi + 1;
 	_jitc->function->self.argi += 2;
     }
@@ -307,15 +316,10 @@ _jit_arg_d(jit_state_t *_jit)
     return (jit_new_node_w(jit_code_arg_d, offset));
 }
 
-jit_bool_t
-_jit_arg_d_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (jit_arg_f_reg_p(offset));
-}
-
 void
 _jit_getarg_c(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg);
     if (v->u.w >= 0)
 	jit_extr_c(u, _R26 - v->u.w);
     else
@@ -325,6 +329,7 @@ _jit_getarg_c(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 void
 _jit_getarg_uc(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg);
     if (v->u.w >= 0)
 	jit_extr_uc(u, _R26 - v->u.w);
     else
@@ -334,6 +339,7 @@ _jit_getarg_uc(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 void
 _jit_getarg_s(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg);
     if (v->u.w >= 0)
 	jit_extr_s(u, _R26 - v->u.w);
     else
@@ -343,6 +349,7 @@ _jit_getarg_s(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 void
 _jit_getarg_us(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg);
     if (v->u.w >= 0)
 	jit_extr_us(u, _R26 - v->u.w);
     else
@@ -352,36 +359,114 @@ _jit_getarg_us(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 void
 _jit_getarg_i(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg);
     if (v->u.w >= 0)
 	jit_movr(u, _R26 - v->u.w);
     else
 	jit_ldxi_i(u, JIT_FP, v->u.w);
 }
 
+void
+_jit_putargr(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg);
+    if (v->u.w >= 0)
+	jit_movr(_R26 - v->u.w, u);
+    else
+	jit_stxi(v->u.w, JIT_FP, u);
+}
+
+void
+_jit_putargi(jit_state_t *_jit, jit_word_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg);
+    if (v->u.w >= 0)
+	jit_movi(_R26 - v->u.w, u);
+    else {
+	regno = jit_get_reg(jit_class_gpr);
+	jit_movi(regno, u);
+	jit_stxi(v->u.w, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_getarg_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg_f);
     if (v->u.w >= 0)
 	jit_movr_f(u, _F4 - v->u.w);
     else
 	jit_ldxi_f(u, JIT_FP, v->u.w);
 }
 
+void
+_jit_putargr_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg_f);
+    if (v->u.w >= 0)
+	jit_movr_f(_F4 - v->u.w, u);
+    else
+	jit_stxi_f(v->u.w, JIT_FP, u);
+}
+
+void
+_jit_putargi_f(jit_state_t *_jit, jit_float32_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg_f);
+    if (v->u.w >= 0)
+	jit_movi_f(_R26 - v->u.w, u);
+    else {
+	regno = jit_get_reg(jit_class_fpr);
+	jit_movi_f(regno, u);
+	jit_stxi_f(v->u.w, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_getarg_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg_d);
     if (v->u.w >= 0)
 	jit_movr_d(u, _F4 - v->u.w);
     else
 	jit_ldxi_d(u, JIT_FP, v->u.w);
 }
 
+void
+_jit_putargr_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg_d);
+    if (v->u.w >= 0)
+	jit_movr_d(_F4 - v->u.w, u);
+    else
+	jit_stxi_d(v->u.w, JIT_FP, u);
+}
+
+void
+_jit_putargi_d(jit_state_t *_jit, jit_float64_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg_d);
+    if (v->u.w >= 0)
+	jit_movi_d(_R26 - v->u.w, u);
+    else {
+	regno = jit_get_reg(jit_class_fpr);
+	jit_movi_d(regno, u);
+	jit_stxi_d(v->u.w, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_pushargr(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
     _jitc->function->call.size -= sizeof(jit_word_t);
-    if (_jitc->function->call.argi < 4) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movr(_R26 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
@@ -395,7 +480,7 @@ _jit_pushargi(jit_state_t *_jit, jit_word_t u)
     jit_int32_t		 regno;
     assert(_jitc->function);
     _jitc->function->call.size -= sizeof(jit_word_t);
-    if (_jitc->function->call.argi < 4) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movi(_R26 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
@@ -412,7 +497,7 @@ _jit_pushargr_f(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
     _jitc->function->call.size -= sizeof(jit_word_t);
-    if (_jitc->function->call.argi < 4) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movr_f(_F4 - _jitc->function->call.argi, u);
 #if !defined(__hpux)
 	if (_jitc->function->call.call & jit_call_varargs)
@@ -434,7 +519,7 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
     jit_int32_t		 regno;
     assert(_jitc->function);
     _jitc->function->call.size -= sizeof(jit_word_t);
-    if (_jitc->function->call.argi < 4) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movi_f(_F4 - _jitc->function->call.argi, u);
 #if !defined(__hpux)
 	if (_jitc->function->call.call & jit_call_varargs)
@@ -464,7 +549,7 @@ _jit_pushargr_d(jit_state_t *_jit, jit_int32_t u)
 	++_jitc->function->call.argi;
 	_jitc->function->call.size -= sizeof(jit_word_t);
     }
-    if (_jitc->function->call.argi < 4) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movr_d(_F4 - (_jitc->function->call.argi + 1), u);
 #if !defined(__hpux)
 	if (_jitc->function->call.call & jit_call_varargs)
@@ -496,7 +581,7 @@ _jit_pushargi_d(jit_state_t *_jit, jit_float64_t u)
 	++_jitc->function->call.argi;
 	_jitc->function->call.size -= sizeof(jit_word_t);
     }
-    if (_jitc->function->call.argi < 4) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movi_d(_F4 - (_jitc->function->call.argi + 1), u);
 #if !defined(__hpux)
 	if (_jitc->function->call.call & jit_call_varargs)
@@ -719,9 +804,19 @@ _emit_code(jit_state_t *_jit)
 	if (_jit->pc.uc >= _jitc->code.end)
 	    return (NULL);
 
+#if DEVEL_DISASSEMBLER
+	node->offset = _jit->pc.w;
+#endif
 	value = jit_classify(node->code);
 	jit_regarg_set(node, value);
 	switch (node->code) {
+	    case jit_code_align:
+		assert(!(node->u.w & (node->u.w - 1)) &&
+		       node->u.w <= sizeof(jit_word_t));
+		if (node->u.w == sizeof(jit_word_t) &&
+		    (word = _jit->pc.w & (sizeof(jit_word_t) - 1)))
+		    nop(sizeof(jit_word_t) - word);
+		break;
 	    case jit_code_note:		case jit_code_name:
 		node->u.w = _jit->pc.w;
 		break;
@@ -742,6 +837,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrw(subc,);
 		case_rrr(subx,);
 		case_rrw(subx,);
+		case_rrw(rsb,);
 		case_rrr(mul,);
 		case_rrw(mul,);
 		case_rrrr(qmul,);
@@ -796,7 +892,8 @@ _emit_code(jit_state_t *_jit)
 		case_rr(ext, _uc);
 		case_rr(ext, _s);
 		case_rr(ext, _us);
-		case_rr(hton,);
+		case_rr(hton, _us);
+		case_rr(hton, _ui);
 		case_rrr(lt,);
 		case_rrw(lt,);
 		case_rrr(lt, _u);
@@ -917,8 +1014,10 @@ _emit_code(jit_state_t *_jit)
 		case_rrf(add, _d, 64);
 		case_rrr(sub, _f);
 		case_rrf(sub, _f, 32);
+		case_rrf(rsb, _f, 32);
 		case_rrr(sub, _d);
 		case_rrf(sub, _d, 64);
+		case_rrf(rsb, _d, 64);
 		case_rrr(mul, _f);
 		case_rrf(mul, _f, 32);
 		case_rrr(mul, _d);
@@ -1059,16 +1158,19 @@ _emit_code(jit_state_t *_jit)
 		jmpr(rn(node->u.w));
 		break;
 	    case jit_code_jmpi:
-		temp = node->u.n;
-		assert(temp->code == jit_code_label ||
-		       temp->code == jit_code_epilog);
-		/* no support for jump outside jit */
-		if (temp->flag & jit_flag_patch)
-		    jmpi(temp->u.w);
-		else {
-		    word = jmpi_p(_jit->pc.w);
-		    patch(word, node);
+		if (node->flag & jit_flag_node) {
+		    temp = node->u.n;
+		    assert(temp->code == jit_code_label ||
+			   temp->code == jit_code_epilog);
+		    if (temp->flag & jit_flag_patch)
+			jmpi(temp->u.w);
+		    else {
+			word = jmpi_p(_jit->pc.w);
+			patch(word, node);
+		    }
 		}
+		else
+		    jmpi(node->u.w);
 		break;
 	    case jit_code_callr:
 		callr(rn(node->u.w));
@@ -1126,6 +1228,7 @@ _emit_code(jit_state_t *_jit)
 		abort();
 	}
 	jit_regarg_clr(node, value);
+	assert(_jitc->regarg == 0);
 	/* update register live state */
 	jit_reglive(node);
     }
@@ -1148,6 +1251,23 @@ _emit_code(jit_state_t *_jit)
 	patch_at(_jitc->patches.ptr[offset].inst, word);
     }
 
+    jit_flush(_jit->code.ptr, _jit->pc.uc);
+
+    return (_jit->code.ptr);
+}
+
+#define CODE				1
+#  include "jit_hppa-cpu.c"
+#  include "jit_hppa-fpu.c"
+#undef CODE
+
+void
+jit_flush(void *fptr, void *tptr)
+{
+    jit_word_t		f, t, s;
+    s = sysconf(_SC_PAGE_SIZE);
+    f = (jit_word_t)fptr & -s;
+    t = (((jit_word_t)tptr) + s - 1) & -s;
 #if defined(__hppa)
 /* --- parisc2.0.pdf ---
 		Programming Note
@@ -1189,9 +1309,7 @@ at most two cachelines.
      * on this software.
      */
     {
-	jit_word_t	f = (jit_word_t)_jit->code.ptr;
 	jit_word_t	n = f + 32;
-	jit_word_t	t = f + _jit->code.length;
 	register int	u, v;
 	for (; f <= t; n = f + 32, f += 64) {
 	    asm volatile ("fdc 0(0,%0)"
@@ -1220,18 +1338,10 @@ at most two cachelines.
     }
 #else
     /* This is supposed to work but appears to fail on multiprocessor systems */
-    word = sysconf(_SC_PAGE_SIZE);
-    __clear_cache(_jit->code.ptr, (void *)((_jit->pc.w + word) & -word));
+    __clear_cache((void *)f, (void *)t);
 #endif
-
-    return (_jit->code.ptr);
 }
 
-#define CODE				1
-#  include "jit_hppa-cpu.c"
-#  include "jit_hppa-fpu.c"
-#undef CODE
-
 void
 _emit_ldxi(jit_state_t *_jit, jit_gpr_t r0, jit_gpr_t r1, jit_word_t i0)
 {
