@@ -22,6 +22,18 @@
 
 #define rc(value)			jit_class_##value
 #define rn(reg)				(jit_regno(_rvs[jit_regno(reg)].spec))
+#define jit_arg_reg_p(i)		((i) >= 0 && (i) < 8)
+#if __BYTE_ORDER == __LITTLE_ENDIAN
+#  define C_DISP			0
+#  define S_DISP			0
+#  define I_DISP			0
+#  define F_DISP			0
+#else
+#  define C_DISP			8 - sizeof(jit_int8_t)
+#  define S_DISP			8 - sizeof(jit_int16_t)
+#  define I_DISP			8 - sizeof(jit_int32_t)
+#  define F_DISP			8 - sizeof(jit_float32_t)
+#endif
 
 /*
  * Prototypes
@@ -350,13 +362,21 @@ _jit_epilog(jit_state_t *_jit)
     _jitc->function = NULL;
 }
 
+jit_bool_t
+_jit_arg_register_p(jit_state_t *_jit, jit_node_t *u)
+{
+    assert(u->code == jit_code_arg ||
+	   u->code == jit_code_arg_f || u->code == jit_code_arg_d);
+    return (jit_arg_reg_p(u->u.w));
+}
+
 jit_node_t *
 _jit_arg(jit_state_t *_jit)
 {
     jit_int32_t		offset;
 
     assert(_jitc->function);
-    if (_jitc->function->self.argi < 8)
+    if (jit_arg_reg_p(_jitc->function->self.argi))
 	offset = _jitc->function->self.argi++;
     else {
 	offset = _jitc->function->self.size;
@@ -365,19 +385,13 @@ _jit_arg(jit_state_t *_jit)
     return (jit_new_node_w(jit_code_arg, offset));
 }
 
-jit_bool_t
-_jit_arg_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (offset >= 0 && offset < 8);
-}
-
 jit_node_t *
 _jit_arg_f(jit_state_t *_jit)
 {
     jit_int32_t		offset;
 
     assert(_jitc->function);
-    if (_jitc->function->self.argi < 8)
+    if (jit_arg_reg_p(_jitc->function->self.argi))
 	offset = _jitc->function->self.argi++;
     else {
 	offset = _jitc->function->self.size;
@@ -392,7 +406,7 @@ _jit_arg_d(jit_state_t *_jit)
     jit_int32_t		offset;
 
     assert(_jitc->function);
-    if (_jitc->function->self.argi < 8)
+    if (jit_arg_reg_p(_jitc->function->self.argi))
 	offset = _jitc->function->self.argi++;
     else {
 	offset = _jitc->function->self.size;
@@ -404,132 +418,173 @@ _jit_arg_d(jit_state_t *_jit)
 void
 _jit_getarg_c(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 8) {
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_c(u, _R32 + v->u.w);
-    }
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_c(u, JIT_FP, v->u.w);
-#else
-	jit_ldxi_c(u, JIT_FP,
-		   v->u.w + (__WORDSIZE >> 3) - sizeof(jit_int8_t));
-#endif
-    }
+    else
+	jit_ldxi_c(u, JIT_FP, v->u.w + C_DISP);
 }
 
 void
 _jit_getarg_uc(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 8)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_uc(u, _R32 + v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_uc(u, JIT_FP, v->u.w);
-#else
-	jit_ldxi_uc(u, JIT_FP,
-		    v->u.w + (__WORDSIZE >> 3) - sizeof(jit_uint8_t));
-#endif
-    }
+    else
+	jit_ldxi_uc(u, JIT_FP, v->u.w + C_DISP);
 }
 
 void
 _jit_getarg_s(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 8)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_s(u, _R32 + v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_s(u, JIT_FP, v->u.w);
-#else
-	jit_ldxi_s(u, JIT_FP,
-		   v->u.w + (__WORDSIZE >> 3) - sizeof(jit_int16_t));
-#endif
-    }
+    else
+	jit_ldxi_s(u, JIT_FP, v->u.w + S_DISP);
 }
 
 void
 _jit_getarg_us(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 8)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_us(u, _R32 + v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_us(u, JIT_FP, v->u.w);
-#else
-	jit_ldxi_us(u, JIT_FP,
-		    v->u.w + (__WORDSIZE >> 3) - sizeof(jit_uint16_t));
-#endif
-    }
+    else
+	jit_ldxi_us(u, JIT_FP, v->u.w + S_DISP);
 }
 
 void
 _jit_getarg_i(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 8)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_i(u, _R32 + v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_i(u, JIT_FP, v->u.w);
-#else
-	jit_ldxi_i(u, JIT_FP,
-		   v->u.w + (__WORDSIZE >> 3) - sizeof(jit_int32_t));
-#endif
-    }
+    else
+	jit_ldxi_i(u, JIT_FP, v->u.w + I_DISP);
 }
 
 void
 _jit_getarg_ui(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 8)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_ui(u, _R32 + v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_ui(u, JIT_FP, v->u.w);
-#else
-	jit_ldxi_ui(u, JIT_FP,
-		    v->u.w + (__WORDSIZE >> 3) - sizeof(jit_uint32_t));
-#endif
-    }
+    else
+	jit_ldxi_ui(u, JIT_FP, v->u.w + I_DISP);
 }
 
 void
 _jit_getarg_l(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 8)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_movr(u, _R32 + v->u.w);
     else
 	jit_ldxi(u, JIT_FP, v->u.w);
 }
 
+void
+_jit_putargr(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
+	jit_movr(_R32 + v->u.w, u);
+    else
+	jit_stxi(v->u.w, JIT_FP, u);
+}
+
+void
+_jit_putargi(jit_state_t *_jit, jit_word_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
+	jit_movi(_R32 + v->u.w, u);
+    else {
+	regno = jit_get_reg(jit_class_gpr);
+	jit_movi(regno, u);
+	jit_stxi(v->u.w, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_getarg_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 8)
+    assert(v->code == jit_code_arg_f);
+    if (jit_arg_reg_p(v->u.w))
 	jit_movr_f(u, _F8 + v->u.w);
+    else
+	jit_ldxi_f(u, JIT_FP, v->u.w + F_DISP);
+}
+
+void
+_jit_putargr_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg_f);
+    if (jit_arg_reg_p(v->u.w))
+	jit_movr_f(_F8 + v->u.w, u);
+    else
+	jit_stxi_f(v->u.w, JIT_FP, u + F_DISP);
+}
+
+void
+_jit_putargi_f(jit_state_t *_jit, jit_float32_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg_f);
+    if (jit_arg_reg_p(v->u.w))
+	jit_movi_f(_F8 + v->u.w, u);
     else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_f(u, JIT_FP, v->u.w);
-#else
-	jit_ldxi_f(u, JIT_FP,
-		   v->u.w + (__WORDSIZE >> 3) - sizeof(jit_float32_t));
-#endif
+	regno = jit_get_reg(jit_class_fpr);
+	jit_movi_f(regno, u);
+	jit_stxi_f(v->u.w, JIT_FP, regno + F_DISP);
+	jit_unget_reg(regno);
     }
 }
 
 void
 _jit_getarg_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 8)
+    assert(v->code == jit_code_arg_d);
+    if (jit_arg_reg_p(v->u.w))
 	jit_movr_d(u, _F8 + v->u.w);
     else
 	jit_ldxi_d(u, JIT_FP, v->u.w);
 }
 
+void
+_jit_putargr_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg_d);
+    if (jit_arg_reg_p(v->u.w))
+	jit_movr_d(_F8 + v->u.w, u);
+    else
+	jit_stxi_d(v->u.w, JIT_FP, u);
+}
+
+void
+_jit_putargi_d(jit_state_t *_jit, jit_float64_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg_d);
+    if (jit_arg_reg_p(v->u.w))
+	jit_movi_d(_F8 + v->u.w, u);
+    else {
+	regno = jit_get_reg(jit_class_fpr);
+	jit_movi_d(regno, u);
+	jit_stxi_d(v->u.w, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_pushargr(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
-    if (_jitc->function->call.argi < 8) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movr(_OUT0 + _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
@@ -544,7 +599,7 @@ _jit_pushargi(jit_state_t *_jit, jit_word_t u)
 {
     jit_int32_t		 regno;
     assert(_jitc->function);
-    if (_jitc->function->call.argi < 8) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movi(_OUT0 + _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
@@ -561,7 +616,7 @@ void
 _jit_pushargr_f(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
-    if (_jitc->function->call.argi < 8) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	if (!(_jitc->function->call.call & jit_call_varargs))
 	    jit_movr_f(_F8 + _jitc->function->call.argi, u);
 	else
@@ -569,11 +624,8 @@ _jit_pushargr_f(jit_state_t *_jit, jit_int32_t u)
 	++_jitc->function->call.argi;
     }
     else {
-	jit_stxi_f(_jitc->function->call.size + params_offset
-#if __BYTE_ORDER == __BIG_ENDIAN
-		   + sizeof(jit_float32_t)
-#endif
-		   , JIT_SP, u);
+	jit_stxi_f(_jitc->function->call.size + params_offset + F_DISP,
+		   JIT_SP, u);
 	_jitc->function->call.size += sizeof(jit_word_t);
     }
 }
@@ -583,7 +635,7 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
 {
     jit_int32_t		 regno;
     assert(_jitc->function);
-    if (_jitc->function->call.argi < 8) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	if (!(_jitc->function->call.call & jit_call_varargs))
 	    jit_movi_f(_F8 + _jitc->function->call.argi, u);
 	else
@@ -593,11 +645,8 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
     else {
 	regno = jit_get_reg(jit_class_fpr);
 	jit_movi_f(regno, u);
-	jit_stxi_f(_jitc->function->call.size + params_offset
-#if __BYTE_ORDER == __BIG_ENDIAN
-		   + sizeof(jit_float32_t)
-#endif
-		   , JIT_SP, regno);
+	jit_stxi_f(_jitc->function->call.size + params_offset + F_DISP,
+		   JIT_SP, regno);
 	_jitc->function->call.size += sizeof(jit_word_t);
 	jit_unget_reg(regno);
     }
@@ -607,7 +656,7 @@ void
 _jit_pushargr_d(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
-    if (_jitc->function->call.argi < 8) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	if (!(_jitc->function->call.call & jit_call_varargs))
 	    jit_movr_d(_F8 + _jitc->function->call.argi, u);
 	else
@@ -625,7 +674,7 @@ _jit_pushargi_d(jit_state_t *_jit, jit_float64_t u)
 {
     jit_int32_t		 regno;
     assert(_jitc->function);
-    if (_jitc->function->call.argi < 8) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	if (!(_jitc->function->call.call & jit_call_varargs))
 	    jit_movi_d(_F8 + _jitc->function->call.argi, u);
 	else
@@ -762,10 +811,18 @@ _emit_code(jit_state_t *_jit)
     undo.patch_offset = 0;
 
     undo.prolog_offset = 0;
-    /* code may start with a jump so add an initial function descriptor */
-    word = _jit->pc.w + 16;
-    il(word);			/* addr */
-    il(0);			/* gp */
+    for (node = _jitc->head; node; node = node->next)
+	if (node->code != jit_code_label &&
+	    node->code != jit_code_note &&
+	    node->code != jit_code_name)
+	    break;
+    if (node && (node->code != jit_code_prolog ||
+		 !(_jitc->functions.ptr + node->w.w)->assume_frame)) {
+	/* code may start with a jump so add an initial function descriptor */
+	word = _jit->pc.w + 16;
+	il(word);		/* addr */
+	il(0);			/* gp */
+    }
 #define case_rr(name, type)						\
 	    case jit_code_##name##r##type:				\
 		name##r##type(rn(node->u.w), rn(node->v.w));		\
@@ -856,9 +913,18 @@ _emit_code(jit_state_t *_jit)
 	value = jit_classify(node->code);
 #if GET_JIT_SIZE
 	sync();
+#endif
+#if DEVEL_DISASSEMBLER
+	sync();
+	node->offset = _jit->pc.w;
 #endif
 	jit_regarg_set(node, value);
 	switch (node->code) {
+	    case jit_code_align:
+		assert(!(node->u.w & (node->u.w - 1)) &&
+		       node->u.w <= sizeof(jit_word_t));
+		/* nothing done */
+		break;
 	    case jit_code_note:		case jit_code_name:
 		sync();
 		node->u.w = _jit->pc.w;
@@ -879,6 +945,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrw(sub,);
 		case_rrr(subx,);
 		case_rrw(subx,);
+		case_rrw(rsb,);
 		case_rrr(subc,);
 		case_rrw(subc,);
 		case_rrr(mul,);
@@ -931,7 +998,9 @@ _emit_code(jit_state_t *_jit)
 		else
 		    movi(rn(node->u.w), node->v.w);
 		break;
-		case_rr(hton,);
+		case_rr(hton, _us);
+		case_rr(hton, _ui);
+		case_rr(hton, _ul);
 		case_rr(ext, _c);
 		case_rr(ext, _uc);
 		case_rr(ext, _s);
@@ -1050,6 +1119,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrf(add, _f, 32);
 		case_rrr(sub, _f);
 		case_rrf(sub, _f, 32);
+		case_rrf(rsb, _f, 32);
 		case_rrr(mul, _f);
 		case_rrf(mul, _f, 32);
 		case_rrr(div, _f);
@@ -1132,6 +1202,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrf(add, _d, 64);
 		case_rrr(sub, _d);
 		case_rrf(sub, _d, 64);
+		case_rrf(rsb, _d, 64);
 		case_rrr(mul, _d);
 		case_rrf(mul, _d, 64);
 		case_rrr(div, _d);
@@ -1214,17 +1285,21 @@ _emit_code(jit_state_t *_jit)
 		jmpr(rn(node->u.w));
 		break;
 	    case jit_code_jmpi:
-		if (_jit->pc.uc == _jit->code.ptr + 16)
-		    _jitc->jump = 1;
-		temp = node->u.n;
-		assert(temp->code == jit_code_label ||
-		       temp->code == jit_code_epilog);
-		if (temp->flag & jit_flag_patch)
-		    jmpi(temp->u.w);
-		else {
-		    word = jmpi_p(_jit->pc.w);
-		    patch(word, node);
+		if (node->flag & jit_flag_node) {
+		    if (_jit->pc.uc == _jit->code.ptr + 16)
+			_jitc->jump = 1;
+		    temp = node->u.n;
+		    assert(temp->code == jit_code_label ||
+			   temp->code == jit_code_epilog);
+		    if (temp->flag & jit_flag_patch)
+			jmpi(temp->u.w);
+		    else {
+			word = jmpi_p(_jit->pc.w);
+			patch(word, node);
+		    }
 		}
+		else
+		    jmpi(node->u.w);
 		break;
 	    case jit_code_callr:
 		callr(rn(node->u.w));
@@ -1252,7 +1327,7 @@ _emit_code(jit_state_t *_jit)
 		undo.prolog_offset = _jitc->prolog.offset;
 	    restart_function:
 		_jitc->again = 0;
-		if (_jitc->jump) {
+		if (_jitc->jump && !_jitc->function->assume_frame) {
 		    /* remember prolog to hide offset adjustment for a jump
 		     * to the start of a function, what is expected to be
 		     * a common practice as first jit instruction */
@@ -1336,6 +1411,12 @@ _emit_code(jit_state_t *_jit)
 	sync();
 #endif
 	jit_regarg_clr(node, value);
+	if (jit_carry == _NOREG)
+	    assert(jit_regset_cmp_ui(&_jitc->regarg, 0) == 0);
+	else {
+	    assert(jit_regset_scan1(&_jitc->regarg, 0) == jit_carry);
+	    assert(jit_regset_scan1(&_jitc->regarg, jit_carry + 1) == ULONG_MAX);
+	}
 	/* update register live state */
 	jit_reglive(node);
     }
@@ -1359,10 +1440,7 @@ _emit_code(jit_state_t *_jit)
 	patch_at(node->code, _jitc->patches.ptr[offset].inst, word);
     }
 
-#if defined(__GNUC__)
-    word = sysconf(_SC_PAGE_SIZE);
-    __clear_cache(_jit->code.ptr, (void *)((_jit->pc.w + word) & -word));
-#endif
+    jit_flush(_jit->code.ptr, _jit->pc.uc);
 
     return (_jit->code.ptr);
 }
@@ -1372,6 +1450,19 @@ _emit_code(jit_state_t *_jit)
 #  include "jit_ia64-fpu.c"
 #undef CODE
 
+void
+jit_flush(void *fptr, void *tptr)
+{
+#if defined(__GNUC__)
+    jit_word_t		f, t, s;
+
+    s = sysconf(_SC_PAGE_SIZE);
+    f = (jit_word_t)fptr & -s;
+    t = (((jit_word_t)tptr) + s - 1) & -s;
+    __clear_cache((void *)f, (void *)t);
+#endif
+}
+
 /* Use r2 that is reserved to not require a jit_get_reg call, also note
  * that addil needs a register that first in 2 bits, so, if using a
  * register other than r2 must be less than r8 (or check for a smaller
