@@ -39,7 +39,7 @@ static jit_free_func_ptr jit_free_ptr = jit_default_free_func;
  * Implementation
  */
 jit_pointer_t
-jit_memcpy(jit_pointer_t dst, jit_pointer_t src, jit_word_t size)
+jit_memcpy(jit_pointer_t dst, const void * src, jit_word_t size)
 {
     if (size)
 	return (memcpy(dst, src, size));
@@ -47,7 +47,7 @@ jit_memcpy(jit_pointer_t dst, jit_pointer_t src, jit_word_t size)
 }
 
 jit_pointer_t
-jit_memmove(jit_pointer_t dst, jit_pointer_t src , jit_word_t size)
+jit_memmove(jit_pointer_t dst, const void *src , jit_word_t size)
 {
     if (size)
 	return (memmove(dst, src, size));
