@@ -42,6 +42,16 @@
 #else
 #  define WORD_ADJUST			0
 #endif
+#define jit_arg_reg_p(i)		((i) >= 0 && (i) < NUM_WORD_ARGS)
+#if __BYTE_ORDER == __LITTLE_ENDIAN
+#  define C_DISP			0
+#  define S_DISP			0
+#  define I_DISP			0
+#else
+#  define C_DISP			STACK_SLOT - sizeof(jit_int8_t)
+#  define S_DISP			STACK_SLOT - sizeof(jit_int16_t)
+#  define I_DISP			STACK_SLOT - sizeof(jit_int32_t)
+#endif
 
 /*
  * Prototypes
@@ -275,13 +285,26 @@ _jit_epilog(jit_state_t *_jit)
     _jitc->function = NULL;
 }
 
+jit_bool_t
+_jit_arg_register_p(jit_state_t *_jit, jit_node_t *u)
+{
+    if (u->code == jit_code_arg)
+	return (jit_arg_reg_p(u->u.w));
+    assert(u->code == jit_code_arg_f || u->code == jit_code_arg_d);
+#if NEW_ABI
+    return (jit_arg_reg_p(u->u.w));
+#else
+    return (u->u.w < 8);
+#endif
+}
+
 jit_node_t *
 _jit_arg(jit_state_t *_jit)
 {
     jit_int32_t		offset;
     assert(_jitc->function);
 #if NEW_ABI
-    if (_jitc->function->self.argi < NUM_WORD_ARGS)
+    if (jit_arg_reg_p(_jitc->function->self.argi))
 	offset = _jitc->function->self.argi++;
     else {
 	offset = _jitc->function->self.size;
@@ -297,19 +320,13 @@ _jit_arg(jit_state_t *_jit)
     return (jit_new_node_w(jit_code_arg, offset));
 }
 
-jit_bool_t
-_jit_arg_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (offset >= 0 && offset < NUM_WORD_ARGS);
-}
-
 jit_node_t *
 _jit_arg_f(jit_state_t *_jit)
 {
     jit_int32_t		offset;
     assert(_jitc->function);
 #if NEW_ABI
-    if (_jitc->function->self.argi < NUM_WORD_ARGS)
+    if (jit_arg_reg_p(_jitc->function->self.argi))
 	offset = _jitc->function->self.argi++;
     else {
 	offset = _jitc->function->self.size;
@@ -317,7 +334,7 @@ _jit_arg_f(jit_state_t *_jit)
     }
 #else
     offset = (_jitc->function->self.size - stack_framesize) >> STACK_SHIFT;
-    if (offset < 4) {
+    if (offset < NUM_WORD_ARGS) {
 	if (!_jitc->function->self.argi) {
 	    if (offset == 0)
 		offset = 4;
@@ -334,20 +351,13 @@ _jit_arg_f(jit_state_t *_jit)
     return (jit_new_node_w(jit_code_arg_f, offset));
 }
 
-jit_bool_t
-_jit_arg_f_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    /* 0-3 integer register, 4-7 float register */
-    return (offset >= 0 && offset < 8);
-}
-
 jit_node_t *
 _jit_arg_d(jit_state_t *_jit)
 {
     jit_int32_t		offset;
     assert(_jitc->function);
 #if NEW_ABI
-    if (_jitc->function->self.argi < NUM_WORD_ARGS)
+    if (jit_arg_reg_p(_jitc->function->self.argi))
 	offset = _jitc->function->self.argi++;
     else {
 	offset = _jitc->function->self.size;
@@ -359,7 +369,7 @@ _jit_arg_d(jit_state_t *_jit)
 	_jitc->function->self.argi = 1;
     }
     offset = (_jitc->function->self.size - stack_framesize) >> STACK_SHIFT;
-    if (offset < 4) {
+    if (offset < NUM_WORD_ARGS) {
 	if (!_jitc->function->self.argi)
 	    offset += 4;
     }
@@ -370,117 +380,114 @@ _jit_arg_d(jit_state_t *_jit)
     return (jit_new_node_w(jit_code_arg_d, offset));
 }
 
-jit_bool_t
-_jit_arg_d_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (jit_arg_f_reg_p(offset));
-}
-
 void
 _jit_getarg_c(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < NUM_WORD_ARGS)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_c(u, _A0 - v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_c(u, _FP, v->u.w);
-#else
-	jit_ldxi_c(u, JIT_FP, v->u.w + STACK_SLOT - sizeof(jit_int8_t));
-#endif
-    }
+    else
+	jit_ldxi_c(u, _FP, v->u.w + C_DISP);
 }
 
 void
 _jit_getarg_uc(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < NUM_WORD_ARGS)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_uc(u, _A0 - v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_uc(u, _FP, v->u.w);
-#else
-	jit_ldxi_uc(u, JIT_FP, v->u.w + STACK_SLOT - sizeof(jit_uint8_t));
-#endif
-    }
+    else
+	jit_ldxi_uc(u, _FP, v->u.w + C_DISP);
 }
 
 void
 _jit_getarg_s(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < NUM_WORD_ARGS)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_s(u, _A0 - v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_s(u, _FP, v->u.w);
-#else
-	jit_ldxi_s(u, JIT_FP, v->u.w + STACK_SLOT - sizeof(jit_int16_t));
-#endif
-    }
+    else
+	jit_ldxi_s(u, _FP, v->u.w + S_DISP);
 }
 
 void
 _jit_getarg_us(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < NUM_WORD_ARGS)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_us(u, _A0 - v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_us(u, _FP, v->u.w);
-#else
-	jit_ldxi_us(u, JIT_FP, v->u.w + STACK_SLOT - sizeof(jit_uint16_t));
-#endif
-    }
+    else
+	jit_ldxi_us(u, _FP, v->u.w + S_DISP);
 }
 
 void
 _jit_getarg_i(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < NUM_WORD_ARGS) {
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w)) {
 #if __WORDSIZE == 64
 	jit_extr_i(u, _A0 - v->u.w);
 #else
 	jit_movr(u, _A0 - v->u.w);
 #endif
     }
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_i(u, _FP, v->u.w);
-#else
-	jit_ldxi_i(u, JIT_FP, v->u.w + STACK_SLOT - sizeof(jit_int32_t));
-#endif
-    }
+    else
+	jit_ldxi_i(u, _FP, v->u.w + I_DISP);
 }
 
 #if __WORDSIZE == 64
 void
 _jit_getarg_ui(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < NUM_WORD_ARGS)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_ui(u, _A0 - v->u.w);
-    else {
-#  if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_ui(u, _FP, v->u.w);
-#  else
-	jit_ldxi_ui(u, JIT_FP, v->u.w + STACK_SLOT - sizeof(jit_uint32_t));
-#  endif
-    }
+    else
+	jit_ldxi_ui(u, _FP, v->u.w + I_DISP);
 }
 
 void
 _jit_getarg_l(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < NUM_WORD_ARGS)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_movr(u, _A0 - v->u.w);
     else
 	jit_ldxi_l(u, _FP, v->u.w);
 }
 #endif
 
+void
+_jit_putargr(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
+	jit_movr(_A0 - v->u.w, u);
+    else
+	jit_stxi(v->u.w, _FP, u);
+}
+
+void
+_jit_putargi(jit_state_t *_jit, jit_word_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
+	jit_movi(_A0 - v->u.w, u);
+    else {
+	regno = jit_get_reg(jit_class_gpr);
+	jit_movi(regno, u);
+	jit_stxi(v->u.w, _FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_getarg_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg_f);
 #if NEW_ABI
-    if (v->u.w < NUM_WORD_ARGS)
+    if (jit_arg_reg_p(v->u.w))
 	jit_movr_f(u, _F12 - v->u.w);
 #else
     if (v->u.w < 4)
@@ -492,11 +499,55 @@ _jit_getarg_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 	jit_ldxi_f(u, _FP, v->u.w);
 }
 
+void
+_jit_putargr_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg_f);
+#if NEW_ABI
+    if (jit_arg_reg_p(v->u.w))
+	jit_movr_f(_F12 - v->u.w, u);
+#else
+    if (v->u.w < 4)
+	jit_movr_f_w(_A0 - v->u.w, u);
+    else if (v->u.w < 8)
+	jit_movr_f(_F12 - ((v->u.w - 4) >> 1), u);
+#endif
+    else
+	jit_stxi_f(v->u.w, _FP, u);
+}
+
+void
+_jit_putargi_f(jit_state_t *_jit, jit_float32_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg_f);
+#if NEW_ABI
+    if (jit_arg_reg_p(v->u.w))
+	jit_movi_f(_F12 - v->u.w, u);
+#else
+    if (v->u.w < 4) {
+	regno = jit_get_reg(jit_class_fpr);
+	jit_movi_f(regno, u);
+	jit_movr_f_w(_A0 - ((v->u.w - 4) >> 1), regno);
+	jit_unget_reg(regno);
+    }
+    else if (v->u.w < 8)
+	jit_movi_f(_F12 - ((v->u.w - 4) >> 1), u);
+#endif
+    else {
+	regno = jit_get_reg(jit_class_fpr);
+	jit_movi_f(regno, u);
+	jit_stxi_f(v->u.w, _FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_getarg_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg_d);
 #if NEW_ABI
-    if (v->u.w < NUM_WORD_ARGS)
+    if (jit_arg_reg_p(v->u.w))
 	jit_movr_d(u, _F12 - v->u.w);
 #else
     if (v->u.w < 4)
@@ -508,12 +559,55 @@ _jit_getarg_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 	jit_ldxi_d(u, _FP, v->u.w);
 }
 
+void
+_jit_putargr_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg_d);
+#if NEW_ABI
+    if (jit_arg_reg_p(v->u.w))
+	jit_movr_d(_F12 - v->u.w, u);
+#else
+    if (v->u.w < 4)
+	jit_movr_d_ww(_A0 - v->u.w, _A0 - (v->u.w + 1), u);
+    else if (v->u.w < 8)
+	jit_movr_d(_F12 - ((v->u.w - 4) >> 1), u);
+#endif
+    else
+	jit_stxi_d(v->u.w, _FP, u);
+}
+
+void
+_jit_putargi_d(jit_state_t *_jit, jit_float64_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg_d);
+#if NEW_ABI
+    if (jit_arg_reg_p(v->u.w))
+	jit_movi_d(_F12 - v->u.w, u);
+#else
+    if (v->u.w < 4) {
+	regno = jit_get_reg(jit_class_fpr);
+	jit_movi_d(regno, u);
+	jit_movr_d_ww(_A0 - v->u.w, _A0 - (v->u.w + 1), regno);
+	jit_unget_reg(regno);
+    }
+    else if (v->u.w < 8)
+	jit_movi_d(_F12 - ((v->u.w - 4) >> 1), u);
+#endif
+    else {
+	regno = jit_get_reg(jit_class_fpr);
+	jit_movi_d(regno, u);
+	jit_stxi_d(v->u.w, _FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_pushargr(jit_state_t *_jit, jit_int32_t u)
 {
 #if NEW_ABI
     assert(_jitc->function);
-    if (_jitc->function->call.argi < NUM_WORD_ARGS) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movr(_A0 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
@@ -526,7 +620,7 @@ _jit_pushargr(jit_state_t *_jit, jit_int32_t u)
     assert(_jitc->function);
     offset = _jitc->function->call.size >> STACK_SHIFT;
     _jitc->function->call.argi = 1;
-    if (offset < NUM_WORD_ARGS)
+    if (jit_arg_reg_p(offset))
 	jit_movr(_A0 - offset, u);
     else
 	jit_stxi(_jitc->function->call.size, JIT_SP, u);
@@ -540,7 +634,7 @@ _jit_pushargi(jit_state_t *_jit, jit_word_t u)
 #if NEW_ABI
     jit_int32_t		regno;
     assert(_jitc->function);
-    if (_jitc->function->call.argi < NUM_WORD_ARGS) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movi(_A0 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
@@ -557,7 +651,7 @@ _jit_pushargi(jit_state_t *_jit, jit_word_t u)
     assert(_jitc->function);
     offset = _jitc->function->call.size >> STACK_SHIFT;
     ++_jitc->function->call.argi;
-    if (offset < NUM_WORD_ARGS)
+    if (jit_arg_reg_p(offset))
 	jit_movi(_A0 - offset, u);
     else {
 	regno = jit_get_reg(jit_class_gpr);
@@ -574,7 +668,7 @@ _jit_pushargr_f(jit_state_t *_jit, jit_int32_t u)
 {
 #if NEW_ABI
     assert(_jitc->function);
-    if (_jitc->function->call.argi < NUM_WORD_ARGS) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	if (!(_jitc->function->call.call & jit_call_varargs))
 	    jit_movr_f(_F12 - _jitc->function->call.argi, u);
 	else
@@ -609,7 +703,7 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
     jit_int32_t		regno;
 #if NEW_ABI
     assert(_jitc->function);
-    if (_jitc->function->call.argi < NUM_WORD_ARGS) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	if (!(_jitc->function->call.call & jit_call_varargs))
 	    jit_movi_f(_F12 - _jitc->function->call.argi, u);
 	else
@@ -650,7 +744,7 @@ _jit_pushargr_d(jit_state_t *_jit, jit_int32_t u)
 {
 #if NEW_ABI
     assert(_jitc->function);
-    if (_jitc->function->call.argi < NUM_WORD_ARGS) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	if (!(_jitc->function->call.call & jit_call_varargs))
 	    jit_movr_d(_F12 - _jitc->function->call.argi, u);
 	else
@@ -693,7 +787,7 @@ _jit_pushargi_d(jit_state_t *_jit, jit_float64_t u)
     jit_int32_t		regno;
 #if NEW_ABI
     assert(_jitc->function);
-    if (_jitc->function->call.argi < NUM_WORD_ARGS) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	if (!(_jitc->function->call.call & jit_call_varargs))
 	    jit_movi_d(_F12 - _jitc->function->call.argi, u);
 	else
@@ -975,9 +1069,19 @@ _emit_code(jit_state_t *_jit)
 	if (_jit->pc.uc >= _jitc->code.end)
 	    return (NULL);
 
+#if DEVEL_DISASSEMBLER
+	node->offset = _jit->pc.w;
+#endif
 	value = jit_classify(node->code);
 	jit_regarg_set(node, value);
 	switch (node->code) {
+	    case jit_code_align:
+		assert(!(node->u.w & (node->u.w - 1)) &&
+		       node->u.w <= sizeof(jit_word_t));
+		if (node->u.w == sizeof(jit_word_t) &&
+		    (word = _jit->pc.w & (sizeof(jit_word_t) - 1)))
+		    nop(sizeof(jit_word_t) - word);
+		break;
 	    case jit_code_note:		case jit_code_name:
 		node->u.w = _jit->pc.w;
 		break;
@@ -998,6 +1102,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrw(subc,);
 		case_rrr(subx,);
 		case_rrw(subx,);
+		case_rrw(rsb,);
 		case_rrr(mul,);
 		case_rrw(mul,);
 		case_rrrr(qmul,);
@@ -1086,7 +1191,11 @@ _emit_code(jit_state_t *_jit)
 		case_rrr(stx, _l);
 		case_wrr(stx, _l);
 #endif
-		case_rr(hton,);
+		case_rr(hton, _us);
+		case_rr(hton, _ui);
+#if __WORDSIZE == 64
+		case_rr(hton, _ul);
+#endif
 		case_rr(ext, _c);
 		case_rr(ext, _uc);
 		case_rr(ext, _s);
@@ -1179,6 +1288,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrf(add, _f, 32);
 		case_rrr(sub, _f);
 		case_rrf(sub, _f, 32);
+		case_rrf(rsb, _f, 32);
 		case_rrr(mul, _f);
 		case_rrf(mul, _f, 32);
 		case_rrr(div, _f);
@@ -1261,6 +1371,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrf(add, _d, 64);
 		case_rrr(sub, _d);
 		case_rrf(sub, _d, 64);
+		case_rrf(rsb, _d, 64);
 		case_rrr(mul, _d);
 		case_rrf(mul, _d, 64);
 		case_rrr(div, _d);
@@ -1343,15 +1454,19 @@ _emit_code(jit_state_t *_jit)
 		jmpr(rn(node->u.w));
 		break;
 	    case jit_code_jmpi:
-		temp = node->u.n;
-		assert(temp->code == jit_code_label ||
-		       temp->code == jit_code_epilog);
-		if (temp->flag & jit_flag_patch)
-		    jmpi(temp->u.w);
-		else {
-		    word = jmpi(_jit->pc.w);
-		    patch(word, node);
+		if (node->flag & jit_flag_node) {
+		    temp = node->u.n;
+		    assert(temp->code == jit_code_label ||
+			   temp->code == jit_code_epilog);
+		    if (temp->flag & jit_flag_patch)
+			jmpi(temp->u.w);
+		    else {
+			word = jmpi(_jit->pc.w);
+			patch(word, node);
+		    }
 		}
+		else
+		    jmpi(node->u.w);
 		break;
 	    case jit_code_callr:
 		callr(rn(node->u.w));
@@ -1453,6 +1568,7 @@ _emit_code(jit_state_t *_jit)
 	    }
 	}
 	jit_regarg_clr(node, value);
+	assert(_jitc->regarg == jit_carry == _NOREG ? 0 : (1 << jit_carry));
 	/* update register live state */
 	jit_reglive(node);
     }
@@ -1473,9 +1589,7 @@ _emit_code(jit_state_t *_jit)
 	patch_at(_jitc->patches.ptr[offset].inst, word);
     }
 
-#if defined(__linux__)
-    _flush_cache((char *)_jit->code.ptr, _jit->pc.uc - _jit->code.ptr, ICACHE);
-#endif
+    jit_flush(_jit->code.ptr, _jit->pc.uc);
 
     return (_jit->code.ptr);
 }
@@ -1485,6 +1599,19 @@ _emit_code(jit_state_t *_jit)
 #  include "jit_mips-fpu.c"
 #undef CODE
 
+void
+jit_flush(void *fptr, void *tptr)
+{
+#if defined(__linux__)
+    jit_word_t		f, t, s;
+
+    s = sysconf(_SC_PAGE_SIZE);
+    f = (jit_word_t)fptr & -s;
+    t = (((jit_word_t)tptr) + s - 1) & -s;
+    _flush_cache((void *)f, t - f, ICACHE);
+#endif
+}
+
 void
 _emit_ldxi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
