@@ -26,7 +26,11 @@
 #    define fits_uint32_p(im)		1
 #  else
 #    define gpr_save_area		144	/* r14~r31 = 18 * 8 */
-#    define params_offset		48
+#    if ABI_ELFv2
+#      define params_offset		32
+#    else
+#      define params_offset		48
+#    endif
 #    define can_sign_extend_int_p(im)					\
 	(((im) >= 0 && (long)(im) <=  0x7fffffffL) ||			\
 	 ((im) <  0 && (long)(im) >= -0x80000000L))
@@ -49,6 +53,7 @@
 #  define _SP_REGNO			1
 #  define _R2_REGNO			2
 #  define _R11_REGNO			11
+#  define _R12_REGNO			12
 #  define _FP_REGNO			31
 #  if __WORDSIZE == 32
 #    define ldr(r0,r1)			ldr_i(r0,r1)
@@ -269,7 +274,42 @@ static void _FXS(jit_state_t*,int,int,int,int,int,int,int);
 #  define LD(d,a,s)			FDs(58,d,a,s)
 #  define LDX(d,a,b)			FX(31,d,a,b,21)
 #  define MCRF(d,s)			FXL(19,d<<2,(s)<<2,0)
-#  define MCRXR(d)			FX(31,d<<2,0,0,512)
+#  if DEBUG
+/* In case instruction is emulated, check the kernel can handle it.
+   Will only generate it if DEBUG is enabled.
+"""
+Chapter 6. Optional Facilities and Instructions that are being
+Phased Out of the Architecture
+...
+6.1 Move To Condition Register from XER
+The mcrxr instruction is being phased out of the archi-
+tecture. Its description is included here as an aid to
+constructing operating system code to emulate it.
+
+Move to Condition Register from XER
+X-form
+mcrxr BF
+31	BF	//	///	///	512	/
+0	6	9	11	16	21	31
+CR(4xBF:4xBF+3) <- XER(32:35)
+XER(32:35) <- 0b0000
+The contents of XER(32:35) are copied to Condition Reg-
+ister field BF. XER(32:35) are set to zero.
+Special Registers Altered:
+CR field BF XER(32:35)
+
+Programming Note
+Warning: This instruction has been phased out of
+the architecture. Attempting to execute this
+instruction will cause the system illegal instruction
+error handler to be invoked
+"""
+ */
+#    define MCRXR(d)			FX(31,d<<2,0,0,512)
+#  else
+#    define MCRXR(cr)			_MCRXR(_jit,cr);
+static void _MCRXR(jit_state_t*, jit_int32_t);
+#  endif
 #  define MFCR(d)			FX(31,d,0,0,19)
 #  define MFMSR(d)			FX(31,d,0,0,83)
 #  define MFSPR(d,s)			FXFX(31,d,s<<5,339)
@@ -325,8 +365,8 @@ static void _FXS(jit_state_t*,int,int,int,int,int,int,int);
 #  define NOP()				ORI(0,0,0)
 #  define ORIS(d,a,u)			FDu(25,a,d,u)
 #  define RFI()				FXL(19,0,0,50)
-#  define RLWIMI(d,s,h,b,e)		FM(20,s,a,h,b,e,0)
-#  define RLWIMI_(d,s,h,b,e)		FM(20,s,a,h,b,e,1)
+#  define RLWIMI(d,s,h,b,e)		FM(20,s,d,h,b,e,0)
+#  define RLWIMI_(d,s,h,b,e)		FM(20,s,d,h,b,e,1)
 #  define INSLWI(a,s,n,b)		RLWIMI(a,s,32-b,b,b+n-1)
 #  define INSRWI(a,s,n,b)		RLWIMI(a,s,32-(b+n),b,(b+n)-1)
 #  define RLWINM(a,s,h,b,e)		FM(21,s,a,h,b,e,0)
@@ -447,10 +487,12 @@ static void _FXS(jit_state_t*,int,int,int,int,int,int,int);
 #  define TWI(t,a,s)			FDs(3,t,a,s)
 #  define TWGTI(a,s)			TWI(8,a,s)
 #  define TWLLEI(a,s)			TWI(6,a,s)
-#  define XOR(d,a,b)			FXO(31,a,d,b,0,316)
-#  define XOR_(d,a,b)			FXO_(31,a,d,b,0,316)
-#  define XORI(s,a,u)			FDu(26,s,a,u)
-#  define XORIS(s,a,u)			FDu(27,s,a,u)
+#  define XOR(d,a,b)			FX(31,a,d,b,316)
+#  define XOR_(d,a,b)			FX_(31,a,d,b,316)
+#  define XORI(s,a,u)			FDu(26,a,s,u)
+#  define XORIS(s,a,u)			FDu(27,a,s,u)
+#  define nop(c)			_nop(_jit,c)
+static void _nop(jit_state_t*,jit_int32_t);
 #  define movr(r0,r1)			_movr(_jit,r0,r1)
 static void _movr(jit_state_t*,jit_int32_t,jit_int32_t);
 #  define movi(r0,i0)			_movi(_jit,r0,i0)
@@ -468,9 +510,22 @@ static jit_word_t _movi_p(jit_state_t*,jit_int32_t,jit_word_t);
 #    define extr_ui(r0,r1)		CLRLDI(r0,r1,32)
 #  endif
 #  if __BYTE_ORDER == __BIG_ENDIAN
-#    define htonr(r0,r1)		movr(r0,r1)
+#    define htonr_us(r0,r1)		extr_us(r0,r1)
+#    if __WORDSIZE == 32
+#      define htonr_ui(r0,r1)		movr(r0,r1)
+#    else
+#      define htonr_ui(r0,r1)		extr_ui(r0,r1)
+#      define htonr_ul(r0,r1)		movr(r0,r1)
+#    endif
 #  else
-#    error need htonr implementation
+#    define htonr_us(r0,r1)		_htonr_us(_jit,r0,r1)
+static void _htonr_us(jit_state_t*,jit_int32_t,jit_int32_t);
+#    define htonr_ui(r0,r1)		_htonr_ui(_jit,r0,r1)
+static void _htonr_ui(jit_state_t*,jit_int32_t,jit_int32_t);
+#    if __WORDSIZE == 64
+#      define htonr_ul(r0,r1)		_htonr_ul(_jit,r0,r1)
+static void _htonr_ul(jit_state_t*,jit_int32_t,jit_int32_t);
+#    endif
 #  endif
 #  define addr(r0,r1,r2)		ADD(r0,r1,r2)
 #  define addi(r0,r1,i0)		_addi(_jit,r0,r1,i0)
@@ -490,6 +545,8 @@ static void _subci(jit_state_t*,jit_int32_t,jit_int32_t,jit_word_t);
 #  define subxr(r0,r1,r2)		SUBFE(r0,r2,r1)
 #  define subxi(r0,r1,i0)		_subxi(_jit,r0,r1,i0)
 static void _subxi(jit_state_t*,jit_int32_t,jit_int32_t,jit_word_t);
+#  define rsbi(r0, r1, i0)		_rsbi(_jit, r0, r1, i0)
+static void _rsbi(jit_state_t*,jit_int32_t,jit_int32_t,jit_word_t);
 #  if __WORDSIZE == 32
 #    define mulr(r0,r1,r2)		MULLW(r0,r1,r2)
 #    define mullr(r0,r1,r2)		MULLW(r0,r1,r2)
@@ -976,6 +1033,32 @@ _FXS(jit_state_t *_jit, int o, int s, int a, int h, int x, int i, int r)
 }
 #endif
 
+#if !DEBUG
+/*
+ * Use the sequence commented at
+ * http://tenfourfox.blogspot.com/2011/04/attention-g5-owners-your-javascript-no.html
+ */
+static void
+_MCRXR(jit_state_t *_jit, jit_int32_t cr)
+{
+    jit_int32_t		reg;
+    reg = jit_get_reg(jit_class_gpr|jit_class_nospill);
+    MFXER(rn(reg));
+    MTCRF(128, rn(reg));
+    RLWINM(rn(reg), rn(reg), 0, 0, 28);
+    MTXER(rn(reg));
+    jit_unget_reg(reg);
+}
+#endif
+
+static void
+_nop(jit_state_t *_jit, jit_int32_t i0)
+{
+    for (; i0 > 0; i0 -= 4)
+	NOP();
+    assert(i0 == 0);
+}
+
 static void
 _movr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
@@ -1032,6 +1115,48 @@ _movi_p(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
     return (word);
 }
 
+#  if __BYTE_ORDER == __LITTLE_ENDIAN
+static void
+_htonr_us(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+{
+    jit_int32_t		t0;
+    t0 = jit_get_reg(jit_class_gpr);
+    rshi(rn(t0), r1, 8);
+    andi(r0, r1, 0xff);
+    andi(rn(t0), rn(t0), 0xff);
+    lshi(r0, r0, 8);
+    orr(r0, r0, rn(t0));
+    jit_unget_reg(t0);
+}
+
+static void
+_htonr_ui(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+{
+    jit_int32_t		reg;
+    reg = jit_get_reg(jit_class_gpr);
+    ROTLWI(rn(reg), r1, 8);
+    RLWIMI(rn(reg), r1, 24, 0, 7);
+    RLWIMI(rn(reg), r1, 24, 16, 23);
+    CLRLDI(r0, rn(reg), 32);
+    jit_unget_reg(reg);
+}
+
+#    if __WORDSIZE == 64
+static void
+_htonr_ul(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+{
+    jit_int32_t		reg;
+    reg = jit_get_reg(jit_class_gpr);
+    rshi_u(rn(reg), r1, 32);
+    htonr_ui(r0, r1);
+    htonr_ui(rn(reg), rn(reg));
+    lshi(r0, r0, 32);
+    orr(r0, r0, rn(reg));
+    jit_unget_reg(reg);
+}
+#    endif
+#  endif
+
 static void
 _addi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
@@ -1109,6 +1234,13 @@ _subxi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
     jit_unget_reg(reg);
 }
 
+static void
+_rsbi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
+{
+    subi(r0, r1, i0);
+    negr(r0, r0);
+}
+
 static void
 _muli(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
@@ -2966,8 +3098,13 @@ _stxi_l(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0, jit_int32_t r1)
 static void
 _jmpr(jit_state_t *_jit, jit_int32_t r0)
 {
+#if 0
     MTLR(r0);
     BLR();
+#else
+    MTCTR(r0);
+    BCTR();
+#endif
 }
 
 /* pc relative jump */
@@ -2976,11 +3113,16 @@ _jmpi(jit_state_t *_jit, jit_word_t i0)
 {
     jit_int32_t		reg;
     jit_word_t		w, d;
-    reg = jit_get_reg(jit_class_gpr|jit_class_nospill);
     w = _jit->pc.w;
     d = (i0 - w) & ~3;
-    B(d);
-    jit_unget_reg(reg);
+    if (can_sign_extend_jump_p(d))
+	B(d);
+    else {
+	reg = jit_get_reg(jit_class_gpr|jit_class_nospill);
+	w = movi_p(rn(reg), i0);
+	jmpr(rn(reg));
+	jit_unget_reg(reg);
+    }
     return (w);
 }
 
@@ -3001,6 +3143,9 @@ static void
 _callr(jit_state_t *_jit, jit_int32_t r0)
 {
 #  if __powerpc__
+#    if ABI_ELFv2
+    movr(_R12_REGNO, r0);
+#    else
     stxi(sizeof(void*) * 5, _SP_REGNO, _R2_REGNO);
     /* FIXME Pretend to not know about r11? */
     if (r0 == _R0_REGNO) {
@@ -3013,12 +3158,13 @@ _callr(jit_state_t *_jit, jit_int32_t r0)
 	ldxi(_R11_REGNO, r0, sizeof(void*) * 2);
     }
     ldr(r0, r0);
+#    endif
 #  endif
 
     MTCTR(r0);
     BCTRL();
 
-#  if __powerpc__
+#  if __powerpc__ && !ABI_ELFv2
     ldxi(_R2_REGNO, _SP_REGNO, sizeof(void*) * 5);
 #  endif
 }
@@ -3072,6 +3218,13 @@ _prolog(jit_state_t *_jit, jit_node_t *node)
     unsigned long	regno;
     jit_word_t		offset;
 
+    if (_jitc->function->define_frame || _jitc->function->assume_frame) {
+	jit_int32_t	frame = -_jitc->function->frame;
+	assert(_jitc->function->self.aoff >= frame);
+	if (_jitc->function->assume_frame)
+	    return;
+	_jitc->function->self.aoff = frame;
+    }
     _jitc->function->stack = ((_jitc->function->self.alen +
 			      _jitc->function->self.size -
 			      _jitc->function->self.aoff) + 15) & -16;
@@ -3124,6 +3277,8 @@ _epilog(jit_state_t *_jit, jit_node_t *node)
     unsigned long	regno;
     jit_word_t		offset;
 
+    if (_jitc->function->assume_frame)
+	return;
 #if __ppc__
     LWZ(_SP_REGNO, _SP_REGNO, 0);
     ldxi(_R0_REGNO, _SP_REGNO, 8);
@@ -3181,7 +3336,7 @@ _patch_at(jit_state_t *_jit, jit_word_t instr, jit_word_t label)
 	    u.i[0] = (u.i[0] & ~0xfffd) | (d & 0xfffe);
 	    break;
 	case 18:					/* Bx */
-#if __powerpc__
+#if __powerpc__ && !ABI_ELFv2
 	    if (_jitc->jump && (!(u.i[0] & 1))) {	/* jmpi label */
 		/* zero is used for toc and env, so, quick check
 		 * if this is a "jmpi main" like initial jit
@@ -3209,6 +3364,47 @@ _patch_at(jit_state_t *_jit, jit_word_t instr, jit_word_t label)
 	    u.i[0] = (u.i[0] & ~0x3fffffd) | (d & 0x3fffffe);
 	    break;
 	case 15:					/* LI */
+#if __WORDSIZE == 32
+#  define MTCTR_OFF		2
+#  define BCTR_OFF		3
+#else
+#  define MTCTR_OFF		6
+#  define BCTR_OFF		7
+#endif
+	    /* movi reg label; jmpr reg */
+	    if (_jitc->jump &&
+#if 0
+		/* check for MLTR(reg) */
+		(u.i[MTCTR_OFF] >> 26) == 31 &&
+		((u.i[MTCTR_OFF] >> 16) & 0x3ff) == 8 &&
+		((u.i[MTCTR_OFF] >> 1) & 0x3ff) == 467 &&
+		/* check for BLR */
+		u.i[BCTR_OFF] == 0x4e800020) {
+#else
+		/* check for MTCTR(reg) */
+		(u.i[MTCTR_OFF] >> 26) == 31 &&
+		((u.i[MTCTR_OFF] >> 16) & 0x3ff) == 9 &&
+		((u.i[MTCTR_OFF] >> 1) & 0x3ff) == 467 &&
+		/* check for BCTR */
+		u.i[BCTR_OFF] == 0x4e800420) {
+#endif
+		/* zero is used for toc and env, so, quick check
+		 * if this is a "jmpi main" like initial jit
+		 * instruction */
+		if (((long *)label)[1] == 0 && ((long *)label)[2] == 0) {
+		    for (d = 0; d < _jitc->prolog.offset; d++) {
+			/* not so pretty, but hides powerpc
+			 * specific abi intrinsics and/or
+			 * implementation from user */
+			if (_jitc->prolog.ptr[d] == label) {
+			    label += sizeof(void*) * 3;
+			    break;
+			}
+		    }
+		}
+	    }
+#undef BCTR_OFF
+#undef MTCTR_OFF
 #if __WORDSIZE == 32
 	    assert(!(u.i[0] & 0x1f0000));
 	    u.i[0] = (u.i[0] & ~0xffff) | ((label >> 16) & 0xffff);
