@@ -19,6 +19,19 @@
 
 #define rc(value)			jit_class_##value
 #define rn(reg)				(jit_regno(_rvs[jit_regno(reg)].spec))
+#define jit_arg_reg_p(i)		((i) >= 0 && (i) < 8)
+#define jit_arg_f_reg_p(i)		((i) >= 0 && (i) < 13)
+#if __BYTE_ORDER == __LITTLE_ENDIAN
+#  define C_DISP			0
+#  define S_DISP			0
+#  define I_DISP			0
+#  define F_DISP			0
+#else
+#  define C_DISP			(__WORDSIZE >> 3) - sizeof(jit_int8_t)
+#  define S_DISP			(__WORDSIZE >> 3) - sizeof(jit_int16_t)
+#  define I_DISP			(__WORDSIZE >> 3) - sizeof(jit_int32_t)
+#  define F_DISP			(__WORDSIZE >> 3) - sizeof(jit_float32_t)
+#endif
 
 /*
  * Prototypes
@@ -252,12 +265,21 @@ _jit_epilog(jit_state_t *_jit)
     _jitc->function = NULL;
 }
 
+jit_bool_t
+_jit_arg_register_p(jit_state_t *_jit, jit_node_t *u)
+{
+    if (u->code == jit_code_arg)
+	return (jit_arg_reg_p(u->u.w));
+    assert(u->code == jit_code_arg_f || u->code == jit_code_arg_d);
+    return (jit_arg_f_reg_p(u->u.w));
+}
+
 jit_node_t *
 _jit_arg(jit_state_t *_jit)
 {
     jit_int32_t		offset;
     assert(_jitc->function);
-    if (_jitc->function->self.argi < 8)
+    if (jit_arg_reg_p(_jitc->function->self.argi))
 	offset = _jitc->function->self.argi++;
     else
 	offset = _jitc->function->self.size;
@@ -265,184 +287,223 @@ _jit_arg(jit_state_t *_jit)
     return (jit_new_node_w(jit_code_arg, offset));
 }
 
-jit_bool_t
-_jit_arg_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (offset >= 0 && offset < 8);
-}
-
 jit_node_t *
 _jit_arg_f(jit_state_t *_jit)
 {
     jit_int32_t		offset;
     assert(_jitc->function);
-    if (_jitc->function->self.argf < 13)
+    if (jit_arg_f_reg_p(_jitc->function->self.argf))
 	offset = _jitc->function->self.argf++;
-    else {
+    else
+	offset = _jitc->function->self.size + F_DISP;
+    if (jit_arg_reg_p(_jitc->function->self.argi)) {
 #if __WORDSIZE == 32
-	offset = _jitc->function->self.size;
+	_jitc->function->self.argi += 2;
 #else
-	offset = _jitc->function->self.size + 4;
+	_jitc->function->self.argi++;
 #endif
     }
     _jitc->function->self.size += sizeof(jit_word_t);
     return (jit_new_node_w(jit_code_arg_f, offset));
 }
 
-jit_bool_t
-_jit_arg_f_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (offset >= 0 && offset < 13);
-}
-
 jit_node_t *
 _jit_arg_d(jit_state_t *_jit)
 {
     jit_int32_t		offset;
     assert(_jitc->function);
-    if (_jitc->function->self.argf < 13)
+    if (jit_arg_f_reg_p(_jitc->function->self.argf))
 	offset = _jitc->function->self.argf++;
     else
 	offset = _jitc->function->self.size;
+    if (jit_arg_reg_p(_jitc->function->self.argi)) {
+#if __WORDSIZE == 32
+	_jitc->function->self.argi += 2;
+#else
+	_jitc->function->self.argi++;
+#endif
+    }
     _jitc->function->self.size += sizeof(jit_float64_t);
     return (jit_new_node_w(jit_code_arg_d, offset));
 }
 
-jit_bool_t
-_jit_arg_d_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (jit_arg_f_reg_p(offset));
-}
-
 void
 _jit_getarg_c(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 8)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_c(u, JIT_RA0 - v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_c(u, JIT_FP, v->u.w);
-#else
-	jit_ldxi_c(u, JIT_FP,
-		   v->u.w + (__WORDSIZE >> 3) - sizeof(jit_int8_t));
-#endif
-    }
+    else
+	jit_ldxi_c(u, JIT_FP, v->u.w + C_DISP);
 }
 
 void
 _jit_getarg_uc(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 8)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_uc(u, JIT_RA0 - v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_uc(u, JIT_FP, v->u.w);
-#else
-	jit_ldxi_uc(u, JIT_FP,
-		    v->u.w + (__WORDSIZE >> 3) - sizeof(jit_uint8_t));
-#endif
-    }
+    else
+	jit_ldxi_uc(u, JIT_FP, v->u.w + C_DISP);
 }
 
 void
 _jit_getarg_s(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 8)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_s(u, JIT_RA0 - v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_s(u, JIT_FP, v->u.w);
-#else
-	jit_ldxi_s(u, JIT_FP,
-		   v->u.w + (__WORDSIZE >> 3) - sizeof(jit_int16_t));
-#endif
-    }
+    else
+	jit_ldxi_s(u, JIT_FP, v->u.w + S_DISP);
 }
 
 void
 _jit_getarg_us(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 8)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_us(u, JIT_RA0 - v->u.w);
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_us(u, JIT_FP, v->u.w);
-#else
-	jit_ldxi_us(u, JIT_FP,
-		    v->u.w + (__WORDSIZE >> 3) - sizeof(jit_uint16_t));
-#endif
-    }
+    else
+	jit_ldxi_us(u, JIT_FP, v->u.w + S_DISP);
 }
 
 void
 _jit_getarg_i(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 8) {
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w)) {
 #if __WORDSIZE == 32
 	jit_movr(u, JIT_RA0 - v->u.w);
 #else
 	jit_extr_i(u, JIT_RA0 - v->u.w);
 #endif
     }
-    else {
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_i(u, JIT_FP, v->u.w);
-#else
-	jit_ldxi_i(u, JIT_FP,
-		   v->u.w + (__WORDSIZE >> 3) - sizeof(jit_int32_t));
-#endif
-    }
+    else
+	jit_ldxi_i(u, JIT_FP, v->u.w + I_DISP);
 }
 
 #if __WORDSIZE == 64
 void
 _jit_getarg_ui(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 8)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_ui(u, JIT_RA0 - v->u.w);
-    else {
-#  if __BYTE_ORDER == __LITTLE_ENDIAN
-	jit_ldxi_ui(u, JIT_FP, v->u.w);
-#  else
-	jit_ldxi_ui(u, JIT_FP,
-		    v->u.w + (__WORDSIZE >> 3) - sizeof(jit_uint32_t));
-#  endif
-    }
+    else
+	jit_ldxi_ui(u, JIT_FP, v->u.w + I_DISP);
 }
 
 void
 _jit_getarg_l(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 8)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_movr(u, JIT_RA0 - v->u.w);
     else
 	jit_ldxi_l(u, JIT_FP, v->u.w);
 }
 #endif
 
+void
+_jit_putargr(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
+	jit_movr(JIT_RA0 - v->u.w, u);
+    else
+	jit_stxi(v->u.w, JIT_FP, u);
+}
+
+void
+_jit_putargi(jit_state_t *_jit, jit_word_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
+	jit_movi(JIT_RA0 - v->u.w, u);
+    else {
+	regno = jit_get_reg(jit_class_gpr);
+	jit_movi(regno, u);
+	jit_stxi(v->u.w, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_getarg_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 13)
+    assert(v->code == jit_code_arg_f);
+    if (jit_arg_f_reg_p(v->u.w))
 	jit_movr_d(u, JIT_FA0 - v->u.w);
     else
 	jit_ldxi_f(u, JIT_FP, v->u.w);
 }
 
+void
+_jit_putargr_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg_f);
+    if (jit_arg_f_reg_p(v->u.w))
+	jit_movr_d(JIT_FA0 - v->u.w, u);
+    else
+	jit_stxi_f(v->u.w, JIT_FP, u);
+}
+
+void
+_jit_putargi_f(jit_state_t *_jit, jit_float32_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg_f);
+    if (jit_arg_f_reg_p(v->u.w))
+	jit_movi_d(JIT_FA0 - v->u.w, u);
+    else {
+	regno = jit_get_reg(jit_class_fpr);
+	jit_movi_d(regno, u);
+	jit_stxi_f(v->u.w, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_getarg_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 13)
+    assert(v->code == jit_code_arg_d);
+    if (jit_arg_f_reg_p(v->u.w))
 	jit_movr_d(u, JIT_FA0 - v->u.w);
     else
 	jit_ldxi_d(u, JIT_FP, v->u.w);
 }
 
+void
+_jit_putargr_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg_d);
+    if (jit_arg_f_reg_p(v->u.w))
+	jit_movr_d(JIT_FA0 - v->u.w, u);
+    else
+	jit_stxi_d(v->u.w, JIT_FP, u);
+}
+
+void
+_jit_putargi_d(jit_state_t *_jit, jit_float64_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg_d);
+    if (jit_arg_f_reg_p(v->u.w))
+	jit_movi_d(JIT_FA0 - v->u.w, u);
+    else {
+	regno = jit_get_reg(jit_class_fpr);
+	jit_movi_d(regno, u);
+	jit_stxi_d(v->u.w, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_pushargr(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
-    if (_jitc->function->call.argi < 8) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movr(JIT_RA0 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
@@ -456,7 +517,7 @@ _jit_pushargi(jit_state_t *_jit, jit_word_t u)
 {
     jit_int32_t		 regno;
     assert(_jitc->function);
-    if (_jitc->function->call.argi < 8) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movi(JIT_RA0 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
@@ -473,12 +534,12 @@ void
 _jit_pushargr_f(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
-    if (_jitc->function->call.argf < 13) {
+    if (jit_arg_f_reg_p(_jitc->function->call.argf)) {
 	jit_movr_d(JIT_FA0 - _jitc->function->call.argf, u);
 	++_jitc->function->call.argf;
 	if (!(_jitc->function->call.call & jit_call_varargs)) {
 	    /* in case of excess arguments */
-	    if (_jitc->function->call.argi < 8) {
+	    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 #if __WORDSIZE == 32
 		_jitc->function->call.argi += 2;
 #else
@@ -489,7 +550,7 @@ _jit_pushargr_f(jit_state_t *_jit, jit_int32_t u)
 	    return;
 	}
     }
-    if (_jitc->function->call.argi < 8) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	/* use reserved 8 bytes area */
 	jit_stxi_d(alloca_offset - 8, JIT_FP, u);
 	jit_ldxi(JIT_RA0 - _jitc->function->call.argi, JIT_FP,
@@ -501,14 +562,9 @@ _jit_pushargr_f(jit_state_t *_jit, jit_int32_t u)
 	_jitc->function->call.argi++;
 #endif
     }
-    else {
-#if __WORDSIZE == 32
-	jit_stxi_f(_jitc->function->call.size + params_offset, JIT_SP, u);
-#else
-	jit_stxi_f(_jitc->function->call.size + params_offset + 4,
+    else
+	jit_stxi_f(_jitc->function->call.size + params_offset + F_DISP,
 		   JIT_SP, u);
-#endif
-    }
     _jitc->function->call.size += sizeof(jit_word_t);
 }
 
@@ -518,12 +574,12 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
     jit_int32_t		 regno;
 
     assert(_jitc->function);
-    if (_jitc->function->call.argf < 13) {
+    if (jit_arg_f_reg_p(_jitc->function->call.argf)) {
 	jit_movi_d(JIT_FA0 - _jitc->function->call.argf, u);
 	++_jitc->function->call.argf;
 	if (!(_jitc->function->call.call & jit_call_varargs)) {
 	    /* in case of excess arguments */
-	    if (_jitc->function->call.argi < 8) {
+	    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 #if __WORDSIZE == 32
 		_jitc->function->call.argi += 2;
 #else
@@ -536,7 +592,7 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
     }
     regno = jit_get_reg(jit_class_fpr);
     jit_movi_f(regno, u);
-    if (_jitc->function->call.argi < 8) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	/* use reserved 8 bytes area */
 	jit_stxi_d(alloca_offset - 8, JIT_FP, regno);
 	jit_ldxi(JIT_RA0 - _jitc->function->call.argi, JIT_FP,
@@ -548,14 +604,9 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
 	_jitc->function->call.argi++;
 #endif
     }
-    else {
-#if __WORDSIZE == 32
-	jit_stxi_f(_jitc->function->call.size + params_offset, JIT_SP, regno);
-#else
-	jit_stxi_f(_jitc->function->call.size + params_offset + 4,
+    else
+	jit_stxi_f(_jitc->function->call.size + params_offset + F_DISP,
 		   JIT_SP, regno);
-#endif
-    }
     _jitc->function->call.size += sizeof(jit_word_t);
     jit_unget_reg(regno);
 }
@@ -564,12 +615,12 @@ void
 _jit_pushargr_d(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
-    if (_jitc->function->call.argf < 13) {
+    if (jit_arg_f_reg_p(_jitc->function->call.argf)) {
 	jit_movr_d(JIT_FA0 - _jitc->function->call.argf, u);
 	++_jitc->function->call.argf;
 	if (!(_jitc->function->call.call & jit_call_varargs)) {
 	    /* in case of excess arguments */
-	    if (_jitc->function->call.argi < 8) {
+	    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 #if __WORDSIZE == 32
 		_jitc->function->call.argi += 2;
 #else
@@ -580,7 +631,7 @@ _jit_pushargr_d(jit_state_t *_jit, jit_int32_t u)
 	    return;
 	}
     }
-    if (_jitc->function->call.argi < 8) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	/* use reserved 8 bytes area */
 	jit_stxi_d(alloca_offset - 8, JIT_FP, u);
 	jit_ldxi(JIT_RA0 - _jitc->function->call.argi, JIT_FP,
@@ -603,12 +654,12 @@ _jit_pushargi_d(jit_state_t *_jit, jit_float64_t u)
     jit_int32_t		 regno;
 
     assert(_jitc->function);
-    if (_jitc->function->call.argf < 13) {
+    if (jit_arg_f_reg_p(_jitc->function->call.argf)) {
 	jit_movi_d(JIT_FA0 - _jitc->function->call.argf, u);
 	++_jitc->function->call.argf;
 	if (!(_jitc->function->call.call & jit_call_varargs)) {
 	    /* in case of excess arguments */
-	    if (_jitc->function->call.argi < 8) {
+	    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 #if __WORDSIZE == 32
 		_jitc->function->call.argi += 2;
 #else
@@ -621,7 +672,7 @@ _jit_pushargi_d(jit_state_t *_jit, jit_float64_t u)
     }
     regno = jit_get_reg(jit_class_fpr);
     jit_movi_d(regno, u);
-    if (_jitc->function->call.argi < 8) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	/* use reserved 8 bytes area */
 	jit_stxi_d(alloca_offset - 8, JIT_FP, regno);
 	jit_ldxi(JIT_RA0 - _jitc->function->call.argi, JIT_FP,
@@ -776,13 +827,21 @@ _emit_code(jit_state_t *_jit)
     undo.node = NULL;
     undo.patch_offset = 0;
 
-#if __powerpc__
+#if __powerpc__ && !ABI_ELFv2
     undo.prolog_offset = 0;
-    /* code may start with a jump so add an initial function descriptor */
-    word = _jit->pc.w + sizeof(void*) * 3;
-    iw(word);				/* addr */
-    iw(0);				/* toc */
-    iw(0);				/* env */
+    for (node = _jitc->head; node; node = node->next)
+	if (node->code != jit_code_label &&
+	    node->code != jit_code_note &&
+	    node->code != jit_code_name)
+	    break;
+    if (node && (node->code != jit_code_prolog ||
+		 !(_jitc->functions.ptr + node->w.w)->assume_frame)) {
+	/* code may start with a jump so add an initial function descriptor */
+	word = _jit->pc.w + sizeof(void*) * 3;
+	iw(word);			/* addr */
+	iw(0);				/* toc */
+	iw(0);				/* env */
+    }
 #endif
 
 #define case_rr(name, type)						\
@@ -872,9 +931,19 @@ _emit_code(jit_state_t *_jit)
 	if (_jit->pc.uc >= _jitc->code.end)
 	    return (NULL);
 
+#if DEVEL_DISASSEMBLER
+	node->offset = _jit->pc.w;
+#endif
 	value = jit_classify(node->code);
 	jit_regarg_set(node, value);
 	switch (node->code) {
+	    case jit_code_align:
+		assert(!(node->u.w & (node->u.w - 1)) &&
+		       node->u.w <= sizeof(jit_word_t));
+		if (node->u.w == sizeof(jit_word_t) &&
+		    (word = _jit->pc.w & (sizeof(jit_word_t) - 1)))
+		    nop(sizeof(jit_word_t) - word);
+		break;
 	    case jit_code_note:		case jit_code_name:
 		node->u.w = _jit->pc.w;
 		break;
@@ -895,6 +964,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrw(subc,);
 		case_rrr(subx,);
 		case_rrw(subx,);
+		case_rrw(rsb,);
 		case_rrr(mul,);
 		case_rrw(mul,);
 		case_rrrr(qmul,);
@@ -933,7 +1003,11 @@ _emit_code(jit_state_t *_jit)
 		case_rr(ext, _i);
 		case_rr(ext, _ui);
 #  endif
-		case_rr(hton,);
+		case_rr(hton, _us);
+		case_rr(hton, _ui);
+#  if __WORDSIZE == 64
+		case_rr(hton, _ul);
+#  endif
 		case_rr(neg,);
 		case_rr(com,);
 		case_rr(mov,);
@@ -1082,6 +1156,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrf(add, _f, 32);
 		case_rrr(sub, _f);
 		case_rrf(sub, _f, 32);
+		case_rrf(rsb, _f, 32);
 		case_rrr(mul, _f);
 		case_rrf(mul, _f, 32);
 		case_rrr(div, _f);
@@ -1164,6 +1239,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrf(add, _d, 64);
 		case_rrr(sub, _d);
 		case_rrf(sub, _d, 64);
+		case_rrf(rsb, _d, 64);
 		case_rrr(mul, _d);
 		case_rrf(mul, _d, 64);
 		case_rrr(div, _d);
@@ -1236,20 +1312,23 @@ _emit_code(jit_state_t *_jit)
 		jmpr(rn(node->u.w));
 		break;
 	    case jit_code_jmpi:
-#if __powerpc__
-		if (_jit->pc.uc == _jit->code.ptr + sizeof(void*) * 3)
-		    _jitc->jump = 1;
+		if (node->flag & jit_flag_node) {
+#if __powerpc__ && !ABI_ELFv2
+		    if (_jit->pc.uc == _jit->code.ptr + sizeof(void*) * 3)
+			_jitc->jump = 1;
 #endif
-		temp = node->u.n;
-		assert(temp->code == jit_code_label ||
-		       temp->code == jit_code_epilog);
-		/* no support for jump outside jit */
-		if (temp->flag & jit_flag_patch)
-		    jmpi(temp->u.w);
-		else {
-		    word = jmpi(_jit->pc.w);
-		    patch(word, node);
+		    temp = node->u.n;
+		    assert(temp->code == jit_code_label ||
+			   temp->code == jit_code_epilog);
+		    if (temp->flag & jit_flag_patch)
+			jmpi(temp->u.w);
+		    else {
+			word = jmpi(_jit->pc.w);
+			patch(word, node);
+		    }
 		}
+		else
+		    (void)jmpi_p(node->u.w);
 		break;
 	    case jit_code_callr:
 		callr(rn(node->u.w));
@@ -1271,13 +1350,13 @@ _emit_code(jit_state_t *_jit)
 		undo.node = node;
 		undo.word = _jit->pc.w;
 		undo.patch_offset = _jitc->patches.offset;
-#if __powerpc__
+#if __powerpc__ && !ABI_ELFv2
 		undo.prolog_offset = _jitc->prolog.offset;
 #endif
 	    restart_function:
 		_jitc->again = 0;
-#if __powerpc__
-		if (_jitc->jump) {
+#if __powerpc__ && !ABI_ELFv2
+		if (_jitc->jump && !_jitc->function->assume_frame) {
 		    /* remember prolog to hide offset adjustment for a jump
 		     * to the start of a function, what is expected to be
 		     * a common practice as first jit instruction */
@@ -1311,7 +1390,7 @@ _emit_code(jit_state_t *_jit)
 		    node = undo.node;
 		    _jit->pc.w = undo.word;
 		    _jitc->patches.offset = undo.patch_offset;
-#if __powerpc__
+#if __powerpc__ && !ABI_ELFv2
 		    _jitc->prolog.offset = undo.prolog_offset;
 #endif
 		    goto restart_function;
@@ -1330,6 +1409,7 @@ _emit_code(jit_state_t *_jit)
 		abort();
 	}
 	jit_regarg_clr(node, value);
+	assert(_jitc->regarg == 0);
 	/* update register live state */
 	jit_reglive(node);
     }
@@ -1350,8 +1430,7 @@ _emit_code(jit_state_t *_jit)
 	patch_at(_jitc->patches.ptr[offset].inst, word);
     }
 
-    word = sysconf(_SC_PAGE_SIZE);
-    __clear_cache(_jit->code.ptr, (void *)((_jit->pc.w + word) & -word));
+    jit_flush(_jit->code.ptr, _jit->pc.uc);
 
     return (_jit->code.ptr);
 }
@@ -1361,6 +1440,19 @@ _emit_code(jit_state_t *_jit)
 #  include "jit_ppc-fpu.c"
 #undef CODE
 
+void
+jit_flush(void *fptr, void *tptr)
+{
+#if defined(__GNUC__)
+    jit_word_t		f, t, s;
+
+    s = sysconf(_SC_PAGE_SIZE);
+    f = (jit_word_t)fptr & -s;
+    t = (((jit_word_t)tptr) + s - 1) & -s;
+    __clear_cache((void *)f, (void *)t);
+#endif
+}
+
 void
 _emit_ldxi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
