@@ -452,6 +452,8 @@ static void _subci(jit_state_t*, jit_int32_t, jit_int32_t, jit_word_t);
 #  define subxr(r0, r1, r2)		SUBXcc(r1, r2, r0)
 #  define subxi(r0, r1, i0)		_subxi(_jit, r0, r1, i0)
 static void _subxi(jit_state_t*, jit_int32_t, jit_int32_t, jit_word_t);
+#  define rsbi(r0, r1, i0)		_rsbi(_jit, r0, r1, i0)
+static void _rsbi(jit_state_t*,jit_int32_t,jit_int32_t,jit_word_t);
 #  define mulr(r0, r1, r2)		UMUL(r1, r2, r0)
 #  define muli(r0, r1, i0)		_muli(_jit, r0, r1, i0)
 static void _muli(jit_state_t*, jit_int32_t, jit_int32_t, jit_word_t);
@@ -506,7 +508,8 @@ static void _xori(jit_state_t*, jit_int32_t, jit_int32_t, jit_word_t);
 #  define rshi(r0, r1, i0)		SRAI(r1, i0, r0)
 #  define rshr_u(r0, r1, r2)		SRL(r1, r2, r0)
 #  define rshi_u(r0, r1, i0)		SRLI(r1, i0, r0)
-#  define htonr(r0,r1)			movr(r0,r1)
+#  define htonr_us(r0,r1)		extr_us(r0,r1)
+#  define htonr_ui(r0,r1)		movr(r0,r1)
 #  define extr_c(r0,r1)			_extr_c(_jit,r0,r1)
 static void _extr_c(jit_state_t*,jit_int32_t,jit_int32_t);
 #  define extr_uc(r0,r1)		andi(r0, r1, 0xff)
@@ -809,8 +812,9 @@ _f1(jit_state_t *_jit, jit_int32_t op, jit_int32_t disp30)
 static void
 _nop(jit_state_t *_jit, jit_int32_t i0)
 {
-    while (i0--)
+    for (; i0 > 0; i0 -= 4)
 	NOP();
+    assert(i0 == 0);
 }
 
 static void
@@ -926,6 +930,13 @@ _subxi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
     }
 }
 
+static void
+_rsbi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
+{
+    subi(r0, r1, i0);
+    negr(r0, r0);
+}
+
 static void
 _muli(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
@@ -1324,7 +1335,7 @@ _ldxi_s(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
     else {
 	reg = jit_get_reg(jit_class_gpr);
 	movi(rn(reg), i0);
-	ldxr_c(r0, r1, rn(reg));
+	ldxr_s(r0, r1, rn(reg));
 	jit_unget_reg(reg);
     }
 }
@@ -1620,6 +1631,13 @@ _calli_p(jit_state_t *_jit, jit_word_t i0)
 static void
 _prolog(jit_state_t *_jit, jit_node_t *node)
 {
+    if (_jitc->function->define_frame || _jitc->function->assume_frame) {
+	jit_int32_t	frame = -_jitc->function->frame;
+	assert(_jitc->function->self.aoff >= frame);
+	if (_jitc->function->assume_frame)
+	    return;
+	_jitc->function->self.aoff = frame;
+    }
     /* align at 16 bytes boundary */
     _jitc->function->stack = ((stack_framesize +
 			      _jitc->function->self.alen -
@@ -1649,6 +1667,8 @@ _prolog(jit_state_t *_jit, jit_node_t *node)
 static void
 _epilog(jit_state_t *_jit, jit_node_t *node)
 {
+    if (_jitc->function->assume_frame)
+	return;
     /* (most) other backends do not save incoming arguments, so,
      * only save locals here */
     if (jit_regset_tstbit(&_jitc->function->regset, _L0))
