@@ -19,6 +19,8 @@
 
 #define rc(value)			jit_class_##value
 #define rn(reg)				(jit_regno(_rvs[jit_regno(reg)].spec))
+#define jit_arg_reg_p(i)		((i) >= 0 && (i) < 6)
+#define jit_arg_d_reg_p(i)		((i) >= 0 && (i) < 5)
 
 /*
  * Prototypes
@@ -225,12 +227,21 @@ _jit_epilog(jit_state_t *_jit)
     _jitc->function = NULL;
 }
 
+jit_bool_t
+_jit_arg_register_p(jit_state_t *_jit, jit_node_t *u)
+{
+    if (u->code == jit_code_arg || u->code == jit_code_arg_f)
+	return (jit_arg_reg_p(u->u.w));
+    assert(u->code == jit_code_arg_d);
+    return (jit_arg_d_reg_p(u->u.w));
+}
+
 jit_node_t *
 _jit_arg(jit_state_t *_jit)
 {
     jit_int32_t		offset;
     assert(_jitc->function);
-    if (_jitc->function->self.argi < 6)
+    if (jit_arg_reg_p(_jitc->function->self.argi))
 	offset = _jitc->function->self.argi++;
     else {
 	offset = _jitc->function->self.size;
@@ -239,18 +250,12 @@ _jit_arg(jit_state_t *_jit)
     return (jit_new_node_w(jit_code_arg, offset));
 }
 
-jit_bool_t
-_jit_arg_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (offset >= 0 && offset < 6);
-}
-
 jit_node_t *
 _jit_arg_f(jit_state_t *_jit)
 {
     jit_int32_t		offset;
     assert(_jitc->function);
-    if (_jitc->function->self.argi < 6)
+    if (jit_arg_reg_p(_jitc->function->self.argi))
 	offset = _jitc->function->self.argi++;
     else {
 	offset = _jitc->function->self.size;
@@ -259,22 +264,16 @@ _jit_arg_f(jit_state_t *_jit)
     return (jit_new_node_w(jit_code_arg_f, offset));
 }
 
-jit_bool_t
-_jit_arg_f_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (0);
-}
-
 jit_node_t *
 _jit_arg_d(jit_state_t *_jit)
 {
     jit_int32_t		offset;
     assert(_jitc->function);
-    if (_jitc->function->self.argi < 5) {
+    if (jit_arg_d_reg_p(_jitc->function->self.argi)) {
 	offset = _jitc->function->self.argi;
 	_jitc->function->self.argi += 2;
     }
-    else if (_jitc->function->self.argi < 6) {
+    else if (jit_arg_reg_p(_jitc->function->self.argi)) {
 	offset = _jitc->function->self.argi++;
 	_jitc->function->self.size += sizeof(jit_float32_t);
     }
@@ -285,74 +284,91 @@ _jit_arg_d(jit_state_t *_jit)
     return (jit_new_node_w(jit_code_arg_d, offset));
 }
 
-jit_bool_t
-_jit_arg_d_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (0);
-}
-
 void
 _jit_getarg_c(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 6)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_c(u, _I0 + v->u.w);
-    else {
-	assert(__BYTE_ORDER == __BIG_ENDIAN && __WORDSIZE == 32);
+    else
 	jit_ldxi_c(u, JIT_FP,
 		   v->u.w + (__WORDSIZE >> 3) - sizeof(jit_int8_t));
-    }
 }
 
 void
 _jit_getarg_uc(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 6)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_uc(u, _I0 + v->u.w);
-    else {
-	assert(__BYTE_ORDER == __BIG_ENDIAN && __WORDSIZE == 32);
+    else
 	jit_ldxi_uc(u, JIT_FP,
 		    v->u.w + (__WORDSIZE >> 3) - sizeof(jit_uint8_t));
-    }
 }
 
 void
 _jit_getarg_s(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 6)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_s(u, _I0 + v->u.w);
-    else {
-	assert(__BYTE_ORDER == __BIG_ENDIAN && __WORDSIZE == 32);
+    else
 	jit_ldxi_s(u, JIT_FP,
 		   v->u.w + (__WORDSIZE >> 3) - sizeof(jit_int16_t));
-    }
 }
 
 void
 _jit_getarg_us(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 6)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_extr_us(u, _I0 + v->u.w);
-    else {
-	assert(__BYTE_ORDER == __BIG_ENDIAN && __WORDSIZE == 32);
+    else
 	jit_ldxi_us(u, JIT_FP,
 		    v->u.w + (__WORDSIZE >> 3) - sizeof(jit_uint16_t));
-    }
 }
 
 void
 _jit_getarg_i(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 6)
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
 	jit_movr(u, _I0 + v->u.w);
     else
 	jit_ldxi_i(u, JIT_FP, v->u.w);
 }
 
+void
+_jit_putargr(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
+	jit_movr(_I0 + v->u.w, u);
+    else
+	jit_stxi(v->u.w, JIT_FP, u);
+}
+
+void
+_jit_putargi(jit_state_t *_jit, jit_word_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
+	jit_movi(_I0 + v->u.w, u);
+    else {
+	regno = jit_get_reg(jit_class_gpr);
+	jit_movi(regno, u);
+	jit_stxi(v->u.w, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_getarg_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg_f);
     assert(_jitc->function);
-    if (v->u.w < 6) {
+    if (jit_arg_reg_p(v->u.w)) {
 	jit_stxi(-4, JIT_FP, _I0 + v->u.w);
 	jit_ldxi_f(u, JIT_FP, -4);
     }
@@ -360,16 +376,45 @@ _jit_getarg_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 	jit_ldxi_f(u, JIT_FP, v->u.w);
 }
 
+void
+_jit_putargr_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg_f);
+    if (jit_arg_reg_p(v->u.w)) {
+	jit_stxi_f(-4, JIT_FP, u);
+	jit_ldxi(_I0 + v->u.w, JIT_FP, -4);
+    }
+    else
+	jit_stxi_f(v->u.w, JIT_FP, u);
+}
+
+void
+_jit_putargi_f(jit_state_t *_jit, jit_float32_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg_f);
+    regno = jit_get_reg(jit_class_fpr);
+    jit_movi_f(regno, u);
+    if (jit_arg_reg_p(v->u.w)) {
+	jit_stxi_f(-4, JIT_FP, regno);
+	jit_ldxi(_I0 + v->u.w, JIT_FP, -4);
+    }
+    else
+	jit_stxi_f(v->u.w, JIT_FP, regno);
+    jit_unget_reg(regno);
+}
+
 void
 _jit_getarg_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
+    assert(v->code == jit_code_arg_d);
     assert(_jitc->function);
-    if (v->u.w < 5) {
+    if (jit_arg_d_reg_p(v->u.w)) {
 	jit_stxi(-8, JIT_FP, _I0 + v->u.w);
 	jit_stxi(-4, JIT_FP, _I0 + v->u.w + 1);
 	jit_ldxi_d(u, JIT_FP, -8);
     }
-    else if (v->u.w < 6) {
+    else if (jit_arg_reg_p(v->u.w)) {
 	jit_stxi(-8, JIT_FP, _I0 + v->u.w);
 	jit_ldxi_f(u, JIT_FP, -8);
 	jit_ldxi_f(u + 1, JIT_FP, stack_framesize);
@@ -380,10 +425,76 @@ _jit_getarg_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
     }
 }
 
+void
+_jit_putargr_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg_d);
+    if (jit_arg_d_reg_p(v->u.w)) {
+	jit_stxi_d(-8, JIT_FP, u);
+	jit_ldxi(_I0 + v->u.w, JIT_FP, -8);
+	jit_ldxi(_I0 + v->u.w + 1, JIT_FP, -4);
+    }
+    else if (jit_arg_reg_p(v->u.w)) {
+	regno = jit_get_reg(jit_class_gpr);
+	jit_stxi_d(-8, JIT_FP, u);
+	jit_ldxi(_I0 + v->u.w, JIT_FP, -8);
+	jit_ldxi(regno, JIT_FP, -4);
+	jit_stxi(stack_framesize, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+    else if ((v->u.w & 7) == 0)
+	jit_stxi_d(v->u.w, JIT_FP, u);
+    else {
+	jit_stxi_d(-8, JIT_FP, u);
+	regno = jit_get_reg(jit_class_gpr);
+	jit_ldxi(regno, JIT_FP, -8);
+	jit_stxi(v->u.w, JIT_FP, regno);
+	jit_ldxi(regno, JIT_FP, -4);
+	jit_stxi(v->u.w + 4, JIT_FP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
+void
+_jit_putargi_d(jit_state_t *_jit, jit_float64_t u, jit_node_t *v)
+{
+    jit_int32_t		regno, gpr;
+    assert(v->code == jit_code_arg_d);
+    regno = jit_get_reg(jit_class_fpr);
+    jit_movi_d(regno, u);
+    if (jit_arg_d_reg_p(v->u.w)) {
+	jit_stxi_d(-8, JIT_FP, regno);
+	jit_ldxi(_I0 + v->u.w, JIT_FP, -8);
+	jit_ldxi(_I0 + v->u.w + 1, JIT_FP, -4);
+    }
+    else if (jit_arg_reg_p(v->u.w)) {
+	gpr = jit_get_reg(jit_class_gpr);
+	jit_stxi_d(-8, JIT_FP, regno);
+	jit_ldxi(_I0 + v->u.w, JIT_FP, -8);
+	jit_ldxi(gpr, JIT_FP, -4);
+	jit_stxi(stack_framesize, JIT_FP, gpr);
+	jit_unget_reg(gpr);
+    }
+    else if ((v->u.w & 7) == 0)
+	jit_stxi_d(v->u.w, JIT_FP, regno);
+    else {
+	jit_stxi_d(-8, JIT_FP, regno);
+	gpr = jit_get_reg(jit_class_gpr);
+	jit_ldxi(gpr, JIT_FP, -8);
+	jit_stxi(v->u.w, JIT_FP, gpr);
+	jit_ldxi(gpr, JIT_FP, -4);
+	jit_stxi(v->u.w + 4, JIT_FP, gpr);
+	jit_unget_reg(gpr);
+    }
+    jit_unget_reg(regno);
+}
+
+
 void
 _jit_pushargr(jit_state_t *_jit, jit_int32_t u)
 {
-    if (_jitc->function->call.argi < 6) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movr(_O0 + _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
@@ -397,7 +508,7 @@ void
 _jit_pushargi(jit_state_t *_jit, jit_word_t u)
 {
     jit_int32_t		regno;
-    if (_jitc->function->call.argi < 6) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movi(_O0 + _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
     }
@@ -413,7 +524,7 @@ _jit_pushargi(jit_state_t *_jit, jit_word_t u)
 void
 _jit_pushargr_f(jit_state_t *_jit, jit_int32_t u)
 {
-    if (_jitc->function->call.argi < 6) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_stxi_f(-4, JIT_FP, u);
 	jit_ldxi(_O0 + _jitc->function->call.argi, JIT_FP, -4);
 	++_jitc->function->call.argi;
@@ -430,7 +541,7 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
     jit_int32_t		regno;
     regno = jit_get_reg(jit_class_fpr);
     jit_movi_f(regno, u);
-    if (_jitc->function->call.argi < 6) {
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_stxi_f(-4, JIT_FP, regno);
 	jit_ldxi(_O0 + _jitc->function->call.argi, JIT_FP, -4);
 	++_jitc->function->call.argi;
@@ -445,13 +556,13 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
 void
 _jit_pushargr_d(jit_state_t *_jit, jit_int32_t u)
 {
-    if (_jitc->function->call.argi < 5) {
+    if (jit_arg_d_reg_p(_jitc->function->call.argi)) {
 	jit_stxi_d(-8, JIT_FP, u);
 	jit_ldxi(_O0 + _jitc->function->call.argi, JIT_FP, -8);
 	jit_ldxi(_O0 + _jitc->function->call.argi + 1, JIT_FP, -4);
 	_jitc->function->call.argi += 2;
     }
-    else if (_jitc->function->call.argi < 6) {
+    else if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_stxi_f(-8, JIT_FP, u);
 	jit_ldxi(_O0 + _jitc->function->call.argi, JIT_FP, -8);
 	++_jitc->function->call.argi;
@@ -473,13 +584,13 @@ _jit_pushargi_d(jit_state_t *_jit, jit_float64_t u)
     jit_int32_t		regno;
     regno = jit_get_reg(jit_class_fpr);
     jit_movi_d(regno, u);
-    if (_jitc->function->call.argi < 5) {
+    if (jit_arg_d_reg_p(_jitc->function->call.argi)) {
 	jit_stxi_d(-8, JIT_FP, regno);
 	jit_ldxi(_O0 + _jitc->function->call.argi, JIT_FP, -8);
 	jit_ldxi(_O0 + _jitc->function->call.argi + 1, JIT_FP, -4);
 	_jitc->function->call.argi += 2;
     }
-    else if (_jitc->function->call.argi < 6) {
+    else if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_stxi_f(-8, JIT_FP, regno);
 	jit_ldxi(_O0 + _jitc->function->call.argi, JIT_FP, -8);
 	++_jitc->function->call.argi;
@@ -712,14 +823,24 @@ _emit_code(jit_state_t *_jit)
 	if (_jit->pc.uc >= _jitc->code.end)
 	    return (NULL);
 
+#if DEVEL_DISASSEMBLER
+	node->offset = _jit->pc.w;
+#endif
 	value = jit_classify(node->code);
 	jit_regarg_set(node, value);
 	switch (node->code) {
+	    case jit_code_align:
+		assert(!(node->u.w & (node->u.w - 1)) &&
+		       node->u.w <= sizeof(jit_word_t));
+		if (node->u.w == sizeof(jit_word_t) &&
+		    (word = _jit->pc.w & (sizeof(jit_word_t) - 1)))
+		    nop(sizeof(jit_word_t) - word);
+		break;
 	    case jit_code_note:		case jit_code_name:
 		node->u.w = _jit->pc.w;
 		break;
 	    case jit_code_label:
-		if (node->link &&
+		if ((node->link || (node->flag & jit_flag_use)) &&
 		    (word = _jit->pc.w & (sizeof(jit_word_t) - 1)))
 		    nop(sizeof(jit_word_t) - word);
 		/* remember label is defined */
@@ -738,6 +859,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrw(subc,);
 		case_rrr(subx,);
 		case_rrw(subx,);
+		case_rrw(rsb,);
 		case_rrr(mul,);
 		case_rrw(mul,);
 		case_rrrr(qmul,);
@@ -822,7 +944,8 @@ _emit_code(jit_state_t *_jit)
 		case_wrr(stx, _s);
 		case_rrr(stx, _i);
 		case_wrr(stx, _i);
-		case_rr(hton,);
+		case_rr(hton, _us);
+		case_rr(hton, _ui);
 		case_rr(ext, _c);
 		case_rr(ext, _uc);
 		case_rr(ext, _s);
@@ -891,6 +1014,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrf(add, _f, 32);
 		case_rrr(sub, _f);
 		case_rrf(sub, _f, 32);
+		case_rrf(rsb, _f, 32);
 		case_rrr(mul, _f);
 		case_rrf(mul, _f, 32);
 		case_rrr(div, _f);
@@ -973,6 +1097,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrf(add, _d, 64);
 		case_rrr(sub, _d);
 		case_rrf(sub, _d, 64);
+		case_rrf(rsb, _d, 64);
 		case_rrr(mul, _d);
 		case_rrf(mul, _d, 64);
 		case_rrr(div, _d);
@@ -1055,15 +1180,19 @@ _emit_code(jit_state_t *_jit)
 		jmpr(rn(node->u.w));
 		break;
 	    case jit_code_jmpi:
-		temp = node->u.n;
-		assert(temp->code == jit_code_label ||
-		       temp->code == jit_code_epilog);
-		if (temp->flag & jit_flag_patch)
-		    jmpi(temp->u.w);
-		else {
-		    word = jmpi_p(_jit->pc.w);
-		    patch(word, node);
+		if (node->flag & jit_flag_node) {
+		    temp = node->u.n;
+		    assert(temp->code == jit_code_label ||
+			   temp->code == jit_code_epilog);
+		    if (temp->flag & jit_flag_patch)
+			jmpi(temp->u.w);
+		    else {
+			word = jmpi_p(_jit->pc.w);
+			patch(word, node);
+		    }
 		}
+		else
+		    jmpi(node->u.w);
 		break;
 	    case jit_code_callr:
 		callr(rn(node->u.w));
@@ -1118,6 +1247,7 @@ _emit_code(jit_state_t *_jit)
 		abort();
 	}
 	jit_regarg_clr(node, value);
+	assert(_jitc->regarg == 0);
 	/* update register live state */
 	jit_reglive(node);
     }
@@ -1141,6 +1271,8 @@ _emit_code(jit_state_t *_jit)
 	patch_at(_jitc->patches.ptr[offset].inst, word);
     }
 
+    jit_flush(_jit->code.ptr, _jit->pc.uc);
+
     return (_jit->code.ptr);
 }
 
@@ -1149,6 +1281,11 @@ _emit_code(jit_state_t *_jit)
 #  include "jit_sparc-fpu.c"
 #undef CODE
 
+void
+jit_flush(void *fptr, void *tptr)
+{
+}
+
 void
 _emit_ldxi(jit_state_t *_jit, jit_gpr_t r0, jit_gpr_t r1, jit_word_t i0)
 {
