@@ -21,7 +21,8 @@
 #define USE_INC_DEC			0
 
 #if PROTO
-#  if __WORDSIZE == 32
+#  if __X32 || __X64_32
+#    define WIDE			0
 #    define ldi(u, v)			ldi_i(u, v)
 #    define ldxi(u, v, w)		ldxi_i(u, v, w)
 #    define sti(u, v)			sti_i(u, v)
@@ -29,20 +30,24 @@
 #    define can_sign_extend_int_p(im)	1
 #    define can_zero_extend_int_p(im)	1
 #    define fits_uint32_p(im)		1
-#    define reg8_p(rn)							\
-      ((rn) >= _RAX_REGNO && (rn) <= _RBX_REGNO)
 #  else
+#    define WIDE			1
 #    define ldi(u, v)			ldi_l(u, v)
 #    define ldxi(u, v, w)		ldxi_l(u, v, w)
 #    define sti(u, v)			sti_l(u, v)
 #    define stxi(u, v, w)		stxi_l(u, v, w)
 #    define can_sign_extend_int_p(im)					\
-	(((im) >= 0 && (long)(im) <=  0x7fffffffL) ||			\
-	 ((im) <  0 && (long)(im) >  -0x80000000L))
-#  define can_zero_extend_int_p(im)					\
-    ((im) >= 0 && (im) < 0x80000000L)
-#    define fits_uint32_p(im)		((im & 0xffffffff00000000L) == 0)
-#    define reg8_p(rn)			1
+	(((im) >= 0 && (long long)(im) <=  0x7fffffffLL) ||		\
+	 ((im) <  0 && (long long)(im) >  -0x80000000LL))
+#    define can_zero_extend_int_p(im)					\
+	((im) >= 0 && (im) < 0x80000000LL)
+#    define fits_uint32_p(im)		(((im) & 0xffffffff00000000LL) == 0)
+#  endif
+#  if __X32 || __CYGWIN__ || __X64_32
+#      define reg8_p(rn)						\
+      ((rn) >= _RAX_REGNO && (rn) <= _RBX_REGNO)
+#  else
+#      define reg8_p(rn)		1
 #  endif
 #  define _RAX_REGNO			0
 #  define _RCX_REGNO			1
@@ -60,8 +65,8 @@
 #  define _R13_REGNO			13
 #  define _R14_REGNO			14
 #  define _R15_REGNO			15
-#  define r7(reg)			(reg & 7)
-#  define r8(reg)			(reg & 15)
+#  define r7(reg)			((reg) & 7)
+#  define r8(reg)			((reg) & 15)
 #  define _SCL1				0x00
 #  define _SCL2				0x01
 #  define _SCL4				0x02
@@ -122,8 +127,10 @@
 #  define ic(c)				*_jit->pc.uc++ = c
 #  define is(s)				*_jit->pc.us++ = s
 #  define ii(i)				*_jit->pc.ui++ = i
-#  if __WORDSIZE == 64
+#  if __X64 && !__X64_32
 #    define il(l)			*_jit->pc.ul++ = l
+#  else
+#    define il(l)			ii(l)
 #  endif
 #  define patch_abs(instr, label)					\
 	*(jit_word_t *)(instr - sizeof(jit_word_t)) = label
@@ -197,6 +204,8 @@ static void _subxr(jit_state_t*,jit_int32_t,jit_int32_t,jit_int32_t);
 #  define isubxi(r0, i0)		alui(X86_SBB, r0, i0)
 #  define subxi(r0, r1, i0)		_subxi(_jit, r0, r1, i0)
 static void _subxi(jit_state_t*,jit_int32_t,jit_int32_t,jit_word_t);
+#  define rsbi(r0, r1, i0)		_rsbi(_jit, r0, r1, i0)
+static void _rsbi(jit_state_t*,jit_int32_t,jit_int32_t,jit_word_t);
 #  define imulr(r0, r1)			_imulr(_jit, r0, r1)
 static void _imulr(jit_state_t*, jit_int32_t, jit_int32_t);
 #  define imuli(r0, r1, i0)		_imuli(_jit, r0, r1, i0)
@@ -356,21 +365,27 @@ static void _movcr_u(jit_state_t*,jit_int32_t,jit_int32_t);
 static void _movsr(jit_state_t*,jit_int32_t,jit_int32_t);
 #  define movsr_u(r0, r1)		_movsr_u(_jit, r0, r1)
 static void _movsr_u(jit_state_t*,jit_int32_t,jit_int32_t);
-#  if __WORDSIZE == 64
+#  if __X64 && !__X64_32
 #    define movir(r0, r1)		_movir(_jit, r0, r1)
 static void _movir(jit_state_t*,jit_int32_t,jit_int32_t);
 #    define movir_u(r0, r1)		_movir_u(_jit, r0, r1)
 static void _movir_u(jit_state_t*,jit_int32_t,jit_int32_t);
 #  endif
-#define htonr(r0, r1)			_htonr(_jit, r0, r1)
-static void _htonr(jit_state_t*,jit_int32_t,jit_int32_t);
+#  define htonr_us(r0, r1)		_htonr_us(_jit, r0, r1)
+static void _htonr_us(jit_state_t*,jit_int32_t,jit_int32_t);
+#  define htonr_ui(r0, r1)		_htonr_ui(_jit, r0, r1)
+static void _htonr_ui(jit_state_t*,jit_int32_t,jit_int32_t);
+#  if __X64 && !__X64_32
+#define htonr_ul(r0, r1)		_htonr_ul(_jit, r0, r1)
+static void _htonr_ul(jit_state_t*,jit_int32_t,jit_int32_t);
+#endif
 #  define extr_c(r0, r1)		_extr_c(_jit, r0, r1)
 static void _extr_c(jit_state_t*,jit_int32_t,jit_int32_t);
 #  define extr_uc(r0, r1)		_extr_uc(_jit, r0, r1)
 static void _extr_uc(jit_state_t*,jit_int32_t,jit_int32_t);
 #  define extr_s(r0, r1)		movsr(r0, r1)
 #  define extr_us(r0, r1)		movsr_u(r0, r1)
-#  if __WORDSIZE == 64
+#  if __X64 && !__X64_32
 #    define extr_i(r0, r1)		movir(r0, r1)
 #    define extr_ui(r0, r1)		movir_u(r0, r1)
 #  endif
@@ -390,19 +405,28 @@ static void _ldi_s(jit_state_t*, jit_int32_t, jit_word_t);
 static void _ldr_us(jit_state_t*, jit_int32_t, jit_int32_t);
 #  define ldi_us(r0, i0)		_ldi_us(_jit, r0, i0)
 static void _ldi_us(jit_state_t*, jit_int32_t, jit_word_t);
-#  define ldr_i(r0, r1)			_ldr_i(_jit, r0, r1)
+#  if __X32 || !__X64_32
+#    define ldr_i(r0, r1)		_ldr_i(_jit, r0, r1)
 static void _ldr_i(jit_state_t*, jit_int32_t, jit_int32_t);
-#  define ldi_i(r0, i0)			_ldi_i(_jit, r0, i0)
+#    define ldi_i(r0, i0)		_ldi_i(_jit, r0, i0)
 static void _ldi_i(jit_state_t*, jit_int32_t, jit_word_t);
-#  if __WORDSIZE == 64
-#    define ldr_ui(r0, r1)		_ldr_ui(_jit, r0, r1)
+#  endif
+#  if __X64
+#    if __X64_32
+#      define ldr_i(r0, r1)		_ldr_ui(_jit, r0, r1)
+#      define ldi_i(r0, i0)		_ldi_ui(_jit, r0, i0)
+#    else
+#      define ldr_ui(r0, r1)		_ldr_ui(_jit, r0, r1)
+#      define ldi_ui(r0, i0)		_ldi_ui(_jit, r0, i0)
+#    endif
 static void _ldr_ui(jit_state_t*, jit_int32_t, jit_int32_t);
-#    define ldi_ui(r0, i0)		_ldi_ui(_jit, r0, i0)
 static void _ldi_ui(jit_state_t*, jit_int32_t, jit_word_t);
-#    define ldr_l(r0, r1)		_ldr_l(_jit, r0, r1)
+#    if !__X64_32
+#      define ldr_l(r0, r1)		_ldr_l(_jit, r0, r1)
 static void _ldr_l(jit_state_t*, jit_int32_t, jit_int32_t);
-#    define ldi_l(r0, i0)		_ldi_l(_jit, r0, i0)
+#      define ldi_l(r0, i0)		_ldi_l(_jit, r0, i0)
 static void _ldi_l(jit_state_t*, jit_int32_t, jit_word_t);
+#    endif
 #  endif
 #  define ldxr_c(r0, r1, r2)		_ldxr_c(_jit, r0, r1, r2)
 static void _ldxr_c(jit_state_t*, jit_int32_t, jit_int32_t, jit_int32_t);
@@ -420,19 +444,28 @@ static void _ldxi_s(jit_state_t*, jit_int32_t, jit_int32_t, jit_word_t);
 static void _ldxr_us(jit_state_t*, jit_int32_t, jit_int32_t, jit_int32_t);
 #  define ldxi_us(r0, r1, i0)		_ldxi_us(_jit, r0, r1, i0)
 static void _ldxi_us(jit_state_t*, jit_int32_t, jit_int32_t, jit_word_t);
-#  define ldxr_i(r0, r1, r2)		_ldxr_i(_jit, r0, r1, r2)
+#  if __X32 || !__X64_32
+#    define ldxr_i(r0, r1, r2)		_ldxr_i(_jit, r0, r1, r2)
 static void _ldxr_i(jit_state_t*, jit_int32_t, jit_int32_t, jit_int32_t);
-#  define ldxi_i(r0, r1, i0)		_ldxi_i(_jit, r0, r1, i0)
+#    define ldxi_i(r0, r1, i0)		_ldxi_i(_jit, r0, r1, i0)
 static void _ldxi_i(jit_state_t*, jit_int32_t, jit_int32_t, jit_word_t);
-#  if __WORDSIZE == 64
-#  define ldxr_ui(r0, r1, r2)		_ldxr_ui(_jit, r0, r1, r2)
+#  endif
+#  if __X64
+#    if __X64_32
+#      define ldxr_i(r0, r1, r2)	_ldxr_ui(_jit, r0, r1, r2)
+#      define ldxi_i(r0, r1, i0)	_ldxi_ui(_jit, r0, r1, i0)
+#    else
+#      define ldxr_ui(r0, r1, r2)	_ldxr_ui(_jit, r0, r1, r2)
+#      define ldxi_ui(r0, r1, i0)	_ldxi_ui(_jit, r0, r1, i0)
+#    endif
 static void _ldxr_ui(jit_state_t*, jit_int32_t, jit_int32_t, jit_int32_t);
-#  define ldxi_ui(r0, r1, i0)		_ldxi_ui(_jit, r0, r1, i0)
 static void _ldxi_ui(jit_state_t*, jit_int32_t, jit_int32_t, jit_word_t);
-#    define ldxr_l(r0, r1, r2)		_ldxr_l(_jit, r0, r1, r2)
+#    if !__X64_32
+#      define ldxr_l(r0, r1, r2)	_ldxr_l(_jit, r0, r1, r2)
 static void _ldxr_l(jit_state_t*, jit_int32_t, jit_int32_t, jit_int32_t);
-#    define ldxi_l(r0, r1, i0)		_ldxi_l(_jit, r0, r1, i0)
+#      define ldxi_l(r0, r1, i0)	_ldxi_l(_jit, r0, r1, i0)
 static void _ldxi_l(jit_state_t*, jit_int32_t, jit_int32_t, jit_word_t);
+#    endif
 #  endif
 #  define str_c(r0, r1)			_str_c(_jit, r0, r1)
 static void _str_c(jit_state_t*, jit_int32_t, jit_int32_t);
@@ -446,7 +479,7 @@ static void _sti_s(jit_state_t*, jit_word_t, jit_int32_t);
 static void _str_i(jit_state_t*, jit_int32_t, jit_int32_t);
 #  define sti_i(i0, r0)			_sti_i(_jit, i0, r0)
 static void _sti_i(jit_state_t*, jit_word_t, jit_int32_t);
-#  if __WORDSIZE == 64
+#  if __X64 && !__X64_32
 #    define str_l(r0, r1)		_str_l(_jit, r0, r1)
 static void _str_l(jit_state_t*, jit_int32_t, jit_int32_t);
 #    define sti_l(i0, r0)		_sti_l(_jit, i0, r0)
@@ -464,7 +497,7 @@ static void _stxi_s(jit_state_t*, jit_word_t, jit_int32_t, jit_int32_t);
 static void _stxr_i(jit_state_t*, jit_int32_t, jit_int32_t, jit_int32_t);
 #  define stxi_i(i0, r0, r1)		_stxi_i(_jit, i0, r0, r1)
 static void _stxi_i(jit_state_t*, jit_word_t, jit_int32_t, jit_int32_t);
-#  if __WORDSIZE == 64
+#  if __X64 && !__X64_32
 #    define stxr_l(r0, r1, r2)		_stxr_l(_jit, r0, r1, r2)
 static void _stxr_l(jit_state_t*, jit_int32_t, jit_int32_t, jit_int32_t);
 #    define stxi_l(i0, r0, r1)		_stxi_l(_jit, i0, r0, r1)
@@ -634,6 +667,13 @@ static void _prolog(jit_state_t*, jit_node_t*);
 static void _epilog(jit_state_t*, jit_node_t*);
 #  define patch_at(node, instr, label)	_patch_at(_jit, node, instr, label)
 static void _patch_at(jit_state_t*, jit_node_t*, jit_word_t, jit_word_t);
+#  if !defined(HAVE_FFSL)
+#    if __X32
+#      define ffsl(i)			ffs(i)
+#    else
+static int ffsl(long);
+#    endif
+#  endif
 #endif
 
 #if CODE
@@ -641,7 +681,7 @@ static void
 _rex(jit_state_t *_jit, jit_int32_t l, jit_int32_t w,
      jit_int32_t r, jit_int32_t x, jit_int32_t b)
 {
-#if __WORDSIZE == 64
+#if __X64
     jit_int32_t	v = 0x40 | (w << 3);
 
     if (r != _NOREG)
@@ -661,7 +701,7 @@ _rx(jit_state_t *_jit, jit_int32_t rd, jit_int32_t md,
 {
     if (ri == _NOREG) {
 	if (rb == _NOREG) {
-#if __WORDSIZE == 32
+#if __X32
 	    mrm(0x00, r7(rd), 0x05);
 #else
 	    mrm(0x00, r7(rd), 0x04);
@@ -770,7 +810,7 @@ static void
 _lea(jit_state_t *_jit, jit_int32_t md, jit_int32_t rb,
      jit_int32_t ri, jit_int32_t ms, jit_int32_t rd)
 {
-    rex(0, 1, rd, ri, rb);
+    rex(0, WIDE, rd, ri, rb);
     ic(0x8d);
     rx(rd, md, rb, ri, ms);
 }
@@ -778,21 +818,21 @@ _lea(jit_state_t *_jit, jit_int32_t md, jit_int32_t rb,
 static void
 _pushr(jit_state_t *_jit, jit_int32_t r0)
 {
-    rex(0, 0, 0, 0, r0);
+    rex(0, WIDE, 0, 0, r0);
     ic(0x50 | r7(r0));
 }
 
 static void
 _popr(jit_state_t *_jit, jit_int32_t r0)
 {
-    rex(0, 0, 0, 0, r0);
+    rex(0, WIDE, 0, 0, r0);
     ic(0x58 | r7(r0));
 }
 
 static void
 _xchgr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
-    rex(0, 1, r1, _NOREG, r0);
+    rex(0, WIDE, r1, _NOREG, r0);
     ic(0x87);
     mrm(0x03, r7(r1), r7(r0));
 }
@@ -800,7 +840,7 @@ _xchgr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 static void
 _testr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
-    rex(0, 1, r1, _NOREG, r0);
+    rex(0, WIDE, r1, _NOREG, r0);
     ic(0x85);
     mrm(0x03, r7(r1), r7(r0));
 }
@@ -808,7 +848,7 @@ _testr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 static void
 _testi(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
 {
-    rex(0, 1, _NOREG, _NOREG, r0);
+    rex(0, WIDE, _NOREG, _NOREG, r0);
     if (r0 == _RAX_REGNO)
 	ic(0xa9);
     else {
@@ -830,7 +870,7 @@ _cc(jit_state_t *_jit, jit_int32_t code, jit_int32_t r0)
 static void
 _alur(jit_state_t *_jit, jit_int32_t code, jit_int32_t r0, jit_int32_t r1)
 {
-    rex(0, 1, r1, _NOREG, r0);
+    rex(0, WIDE, r1, _NOREG, r0);
     ic(code | 0x01);
     mrm(0x03, r7(r1), r7(r0));
 }
@@ -840,7 +880,7 @@ _alui(jit_state_t *_jit, jit_int32_t code, jit_int32_t r0, jit_word_t i0)
 {
     jit_int32_t		reg;
     if (can_sign_extend_int_p(i0)) {
-	rex(0, 1, _NOREG, _NOREG, r0);
+	rex(0, WIDE, _NOREG, _NOREG, r0);
 	if ((jit_int8_t)i0 == i0) {
 	    ic(0x83);
 	    ic(0xc0 | code | r7(r0));
@@ -1097,10 +1137,17 @@ _subxi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
     }
 }
 
+static void
+_rsbi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
+{
+    subi(r0, r1, i0);
+    negr(r0, r0);
+}
+
 static void
 _imulr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
-    rex(0, 1, r0, _NOREG, r1);
+    rex(0, WIDE, r0, _NOREG, r1);
     ic(0x0f);
     ic(0xaf);
     mrm(0x03, r7(r0), r7(r1));
@@ -1111,7 +1158,7 @@ _imuli(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
     jit_int32_t		reg;
     if (can_sign_extend_int_p(i0)) {
-	rex(0, 1, r0, _NOREG, r1);
+	rex(0, WIDE, r0, _NOREG, r1);
 	if ((jit_int8_t)i0 == i0) {
 	    ic(0x6b);
 	    mrm(0x03, r7(r0), r7(r1));
@@ -1277,7 +1324,7 @@ _iqmuli(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1,
 static void
 _sign_extend_rdx_rax(jit_state_t *_jit)
 {
-    rex(0, 1, 0, 0, 0);
+    rex(0, WIDE, 0, 0, 0);
     ic(0x99);
 }
 
@@ -1685,7 +1732,7 @@ _xori(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 static void
 _irotshr(jit_state_t *_jit, jit_int32_t code, jit_int32_t r0)
 {
-    rex(0, 1, _RCX_REGNO, _NOREG, r0);
+    rex(0, WIDE, _RCX_REGNO, _NOREG, r0);
     ic(0xd3);
     mrm(0x03, code, r7(r0));
 }
@@ -1741,7 +1788,7 @@ _rotshr(jit_state_t *_jit, jit_int32_t code,
 static void
 _irotshi(jit_state_t *_jit, jit_int32_t code, jit_int32_t r0, jit_word_t i0)
 {
-    rex(0, 1, _NOREG, _NOREG, r0);
+    rex(0, WIDE, _NOREG, _NOREG, r0);
     if (i0 == 1) {
 	ic(0xd1);
 	mrm(0x03, code, r7(r0));
@@ -1776,7 +1823,7 @@ _lshi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 static void
 _unr(jit_state_t *_jit, jit_int32_t code, jit_int32_t r0)
 {
-    rex(0, 1, _NOREG, _NOREG, r0);
+    rex(0, WIDE, _NOREG, _NOREG, r0);
     ic(0xf7);
     mrm(0x03, code, r7(r0));
 }
@@ -1804,8 +1851,8 @@ static void
 _incr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
     movr(r0, r1);
-#  if __WORDSIZE == 64
-    rex(0, 1, _NOREG, _NOREG, r0);
+#  if __X64
+    rex(0, WIDE, _NOREG, _NOREG, r0);
     ic(0xff);
     ic(0xc0 | r7(r0));
 #  else
@@ -1817,8 +1864,8 @@ static void
 _decr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
     movr(r0, r1);
-#  if __WORDSIZE == 64
-    rex(0, 1, _NOREG, _NOREG, r0);
+#  if __X64
+    rex(0, WIDE, _NOREG, _NOREG, r0);
     ic(0xff);
     ic(0xc8 | r7(r0));
 #  else
@@ -1871,7 +1918,7 @@ _ci(jit_state_t *_jit,
 	reg = jit_get_reg(jit_class_gpr|jit_class_rg8);
 	ixorr(rn(reg), rn(reg));
 	icmpi(r1, i0);
-	cc(code, reg);
+	cc(code, rn(reg));
 	movr(r0, rn(reg));
 	jit_unget_reg(reg);
     }
@@ -2067,17 +2114,21 @@ _movr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 static void
 _imovi(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
 {
-#if __WORDSIZE == 64
+#if __X64
+#  if !__X64_32
     if (fits_uint32_p(i0)) {
+#  endif
 	rex(0, 0, _NOREG, _NOREG, r0);
 	ic(0xb8 | r7(r0));
 	ii(i0);
+#  if !__X64_32
     }
     else {
 	rex(0, 1, _NOREG, _NOREG, r0);
 	ic(0xb8 | r7(r0));
 	il(i0);
     }
+#  endif
 #else
     ic(0xb8 | r7(r0));
     ii(i0);
@@ -2096,21 +2147,16 @@ _movi(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
 static jit_word_t
 _movi_p(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
 {
-#if __WORDSIZE == 64
-    rex(0, 1, _NOREG, _NOREG, r0);
+    rex(0, WIDE, _NOREG, _NOREG, r0);
     ic(0xb8 | r7(r0));
     il(i0);
-#else
-    ic(0xb8 | r7(r0));
-    ii(i0);
-#endif
     return (_jit->pc.w);
 }
 
 static void
 _movcr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
-    rex(0, 1, r0, _NOREG, r1);
+    rex(0, WIDE, r0, _NOREG, r1);
     ic(0x0f);
     ic(0xbe);
     mrm(0x03, r7(r0), r7(r1));
@@ -2119,7 +2165,7 @@ _movcr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 static void
 _movcr_u(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
-    rex(0, 1, r0, _NOREG, r1);
+    rex(0, WIDE, r0, _NOREG, r1);
     ic(0x0f);
     ic(0xb6);
     mrm(0x03, r7(r0), r7(r1));
@@ -2128,7 +2174,7 @@ _movcr_u(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 static void
 _movsr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
-    rex(0, 1, r0, _NOREG, r1);
+    rex(0, WIDE, r0, _NOREG, r1);
     ic(0x0f);
     ic(0xbf);
     mrm(0x03, r7(r0), r7(r1));
@@ -2137,13 +2183,13 @@ _movsr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 static void
 _movsr_u(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
-    rex(0, 1, r0, _NOREG, r1);
+    rex(0, WIDE, r0, _NOREG, r1);
     ic(0x0f);
     ic(0xb7);
     mrm(0x03, r7(r0), r7(r1));
 }
 
-#if __WORDSIZE == 64
+#if __X64
 static void
 _movir(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
@@ -2162,13 +2208,35 @@ _movir_u(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 #endif
 
 static void
-_htonr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+_htonr_us(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+{
+    extr_us(r0, r1);
+    ic(0x66);
+    rex(0, 0, _NOREG, _NOREG, r0);
+    ic(0xc1);
+    mrm(0x03, X86_ROR, r7(r0));
+    ic(8);
+}
+
+static void
+_htonr_ui(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+{
+    movr(r0, r1);
+    rex(0, 0, _NOREG, _NOREG, r0);
+    ic(0x0f);
+    ic(0xc8 | r7(r0));
+}
+
+#if __X64 && !__X64_32
+static void
+_htonr_ul(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
     movr(r0, r1);
     rex(0, 1, _NOREG, _NOREG, r0);
     ic(0x0f);
     ic(0xc8 | r7(r0));
 }
+#endif
 
 static void
 _extr_c(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
@@ -2201,7 +2269,7 @@ _extr_uc(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 static void
 _ldr_c(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
-    rex(0, 1, r0, _NOREG, r1);
+    rex(0, WIDE, r0, _NOREG, r1);
     ic(0x0f);
     ic(0xbe);
     rx(r0, 0, r1, _NOREG, _SCL1);
@@ -2212,7 +2280,7 @@ _ldi_c(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
 {
     jit_int32_t		reg;
     if (can_sign_extend_int_p(i0)) {
-	rex(0, 1, r0, _NOREG, _NOREG);
+	rex(0, WIDE, r0, _NOREG, _NOREG);
 	ic(0x0f);
 	ic(0xbe);
 	rx(r0, i0, _NOREG, _NOREG, _SCL1);
@@ -2228,7 +2296,7 @@ _ldi_c(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
 static void
 _ldr_uc(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
-    rex(0, 1, r0, _NOREG, r1);
+    rex(0, WIDE, r0, _NOREG, r1);
     ic(0x0f);
     ic(0xb6);
     rx(r0, 0, r1, _NOREG, _SCL1);
@@ -2239,7 +2307,7 @@ _ldi_uc(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
 {
     jit_int32_t		reg;
     if (can_sign_extend_int_p(i0)) {
-	rex(0, 1, r0, _NOREG, _NOREG);
+	rex(0, WIDE, r0, _NOREG, _NOREG);
 	ic(0x0f);
 	ic(0xb6);
 	rx(r0, i0, _NOREG, _NOREG, _SCL1);
@@ -2255,7 +2323,7 @@ _ldi_uc(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
 static void
 _ldr_s(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
-    rex(0, 1, r0, _NOREG, r1);
+    rex(0, WIDE, r0, _NOREG, r1);
     ic(0x0f);
     ic(0xbf);
     rx(r0, 0, r1, _NOREG, _SCL1);
@@ -2266,7 +2334,7 @@ _ldi_s(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
 {
     jit_int32_t		reg;
     if (can_sign_extend_int_p(i0)) {
-	rex(0, 1, r0, _NOREG, _NOREG);
+	rex(0, WIDE, r0, _NOREG, _NOREG);
 	ic(0x0f);
 	ic(0xbf);
 	rx(r0, i0, _NOREG, _NOREG, _SCL1);
@@ -2282,7 +2350,7 @@ _ldi_s(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
 static void
 _ldr_us(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
-    rex(0, 1, r0, _NOREG, r1);
+    rex(0, WIDE, r0, _NOREG, r1);
     ic(0x0f);
     ic(0xb7);
     rx(r0, 0, r1, _NOREG, _SCL1);
@@ -2293,7 +2361,7 @@ _ldi_us(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
 {
     jit_int32_t		reg;
     if (can_sign_extend_int_p(i0)) {
-	rex(0, 1, r0, _NOREG, _NOREG);
+	rex(0, WIDE, r0, _NOREG, _NOREG);
 	ic(0x0f);
 	ic(0xb7);
 	rx(r0, i0, _NOREG, _NOREG, _SCL1);
@@ -2306,11 +2374,12 @@ _ldi_us(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
     }
 }
 
+#if __X32 || !__X64_32
 static void
 _ldr_i(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
-#if __WORDSIZE == 64
-    rex(0, 1, r0, _NOREG, r1);
+#if __X64
+    rex(0, WIDE, r0, _NOREG, r1);
     ic(0x63);
 #else
     ic(0x8b);
@@ -2323,8 +2392,8 @@ _ldi_i(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
 {
     jit_int32_t		reg;
     if (can_sign_extend_int_p(i0)) {
-#if __WORDSIZE == 64
-	rex(0, 1, r0, _NOREG, _NOREG);
+#if __X64
+	rex(0, WIDE, r0, _NOREG, _NOREG);
 	ic(0x63);
 #else
 	ic(0x8b);
@@ -2338,8 +2407,9 @@ _ldi_i(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
 	jit_unget_reg(reg);
     }
 }
+#endif
 
-#if __WORDSIZE == 64
+#if __X64
 static void
 _ldr_ui(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
@@ -2365,6 +2435,7 @@ _ldi_ui(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
     }
 }
 
+#  if !__X64_32
 static void
 _ldr_l(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
@@ -2389,12 +2460,13 @@ _ldi_l(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
 	jit_unget_reg(reg);
     }
 }
+#  endif
 #endif
 
 static void
 _ldxr_c(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_int32_t r2)
 {
-    rex(0, 1, r0, r1, r2);
+    rex(0, WIDE, r0, r1, r2);
     ic(0x0f);
     ic(0xbe);
     rx(r0, 0, r2, r1, _SCL1);
@@ -2405,7 +2477,7 @@ _ldxi_c(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
     jit_int32_t		reg;
     if (can_sign_extend_int_p(i0)) {
-	rex(0, 1, r0, _NOREG, r1);
+	rex(0, WIDE, r0, _NOREG, r1);
 	ic(0x0f);
 	ic(0xbe);
 	rx(r0, i0, r1, _NOREG, _SCL1);
@@ -2421,7 +2493,7 @@ _ldxi_c(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 static void
 _ldxr_uc(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_int32_t r2)
 {
-    rex(0, 1, r0, r1, r2);
+    rex(0, WIDE, r0, r1, r2);
     ic(0x0f);
     ic(0xb6);
     rx(r0, 0, r2, r1, _SCL1);
@@ -2432,7 +2504,7 @@ _ldxi_uc(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
     jit_int32_t		reg;
     if (can_sign_extend_int_p(i0)) {
-	rex(0, 1, r0, _NOREG, r1);
+	rex(0, WIDE, r0, _NOREG, r1);
 	ic(0x0f);
 	ic(0xb6);
 	rx(r0, i0, r1, _NOREG, _SCL1);
@@ -2448,7 +2520,7 @@ _ldxi_uc(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 static void
 _ldxr_s(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_int32_t r2)
 {
-    rex(0, 1, r0, r1, r2);
+    rex(0, WIDE, r0, r1, r2);
     ic(0x0f);
     ic(0xbf);
     rx(r0, 0, r2, r1, _SCL1);
@@ -2459,7 +2531,7 @@ _ldxi_s(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
     jit_int32_t		reg;
     if (can_sign_extend_int_p(i0)) {
-	rex(0, 1, r0, _NOREG, r1);
+	rex(0, WIDE, r0, _NOREG, r1);
 	ic(0x0f);
 	ic(0xbf);
 	rx(r0, i0, r1, _NOREG, _SCL1);
@@ -2475,7 +2547,7 @@ _ldxi_s(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 static void
 _ldxr_us(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_int32_t r2)
 {
-    rex(0, 1, r0, r1, r2);
+    rex(0, WIDE, r0, r1, r2);
     ic(0x0f);
     ic(0xb7);
     rx(r0, 0, r2, r1, _SCL1);
@@ -2486,7 +2558,7 @@ _ldxi_us(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
     jit_int32_t		reg;
     if (can_sign_extend_int_p(i0)) {
-	rex(0, 1, r0, _NOREG, r1);
+	rex(0, WIDE, r0, _NOREG, r1);
 	ic(0x0f);
 	ic(0xb7);
 	rx(r0, i0, r1, _NOREG, _SCL1);
@@ -2499,11 +2571,12 @@ _ldxi_us(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
     }
 }
 
+#if __X64 || !__X64_32
 static void
 _ldxr_i(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_int32_t r2)
 {
-#if __WORDSIZE == 64
-    rex(0, 1, r0, r1, r2);
+#if __X64
+    rex(0, WIDE, r0, r1, r2);
     ic(0x63);
 #else
     ic(0x8b);
@@ -2516,8 +2589,8 @@ _ldxi_i(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
     jit_int32_t		reg;
     if (can_sign_extend_int_p(i0)) {
-#if __WORDSIZE == 64
-	rex(0, 1, r0, _NOREG, r1);
+#if __X64
+	rex(0, WIDE, r0, _NOREG, r1);
 	ic(0x63);
 #else
 	ic(0x8b);
@@ -2531,8 +2604,9 @@ _ldxi_i(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 	jit_unget_reg(reg);
     }
 }
+#endif
 
-#if __WORDSIZE == 64
+#if __X64
 static void
 _ldxr_ui(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_int32_t r2)
 {
@@ -2558,6 +2632,7 @@ _ldxi_ui(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
     }
 }
 
+#  if !__X64_32
 static void
 _ldxr_l(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_int32_t r2)
 {
@@ -2582,6 +2657,7 @@ _ldxi_l(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 	jit_unget_reg(reg);
     }
 }
+#  endif
 #endif
 
 static void
@@ -2596,6 +2672,7 @@ _str_c(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
     else {
 	reg = jit_get_reg(jit_class_gpr|jit_class_rg8);
 	movr(rn(reg), r1);
+	rex(0, 0, rn(reg), _NOREG, r0);
 	ic(0x88);
 	rx(rn(reg), 0, r0, _NOREG, _SCL1);
 	jit_unget_reg(reg);
@@ -2615,6 +2692,7 @@ _sti_c(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0)
 	else {
 	    reg = jit_get_reg(jit_class_gpr|jit_class_rg8);
 	    movr(rn(reg), r0);
+	    rex(0, 0, rn(reg), _NOREG, _NOREG);
 	    ic(0x88);
 	    rx(rn(reg), i0, _NOREG, _NOREG, _SCL1);
 	    jit_unget_reg(reg);
@@ -2680,7 +2758,7 @@ _sti_i(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0)
     }
 }
 
-#if __WORDSIZE == 64
+#if __X64 && !__X64_32
 static void
 _str_l(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
@@ -2719,6 +2797,7 @@ _stxr_c(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_int32_t r2)
     else {
 	reg = jit_get_reg(jit_class_gpr|jit_class_rg8);
 	movr(rn(reg), r2);
+	rex(0, 0, rn(reg), r1, r0);
 	ic(0x88);
 	rx(rn(reg), 0, r0, r1, _SCL1);
 	jit_unget_reg(reg);
@@ -2738,6 +2817,7 @@ _stxi_c(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0, jit_int32_t r1)
 	else {
 	    reg = jit_get_reg(jit_class_gpr|jit_class_rg8);
 	    movr(rn(reg), r1);
+	    rex(0, 0, rn(reg), _NOREG, r0);
 	    ic(0x88);
 	    rx(rn(reg), i0, r0, _NOREG, _SCL1);
 	    jit_unget_reg(reg);
@@ -2803,7 +2883,7 @@ _stxi_i(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0, jit_int32_t r1)
     }
 }
 
-#if __WORDSIZE == 64
+#if __X64 && !__X64_32
 static void
 _stxr_l(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_int32_t r2)
 {
@@ -3044,6 +3124,7 @@ _bmsi(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0, jit_word_t i1)
 	reg = jit_get_reg(jit_class_gpr);
 	movi(rn(reg), i1);
 	testr(r0, rn(reg));
+	jit_unget_reg(reg);
     }
     jnz(i0);
     return (_jit->pc.w);
@@ -3067,6 +3148,7 @@ _bmci(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0, jit_word_t i1)
 	reg = jit_get_reg(jit_class_gpr);
 	movi(rn(reg), i1);
 	testr(r0, rn(reg));
+	jit_unget_reg(reg);
     }
     jz(i0);
     return (_jit->pc.w);
@@ -3259,7 +3341,7 @@ _bxsubi_u(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0, jit_word_t i1)
 static void
 _callr(jit_state_t *_jit, jit_int32_t r0)
 {
-#if __WORDSIZE == 32
+#if __X32
     emms();
 #endif
     rex(0, 0, _NOREG, _NOREG, r0);
@@ -3271,7 +3353,7 @@ static jit_word_t
 _calli(jit_state_t *_jit, jit_word_t i0)
 {
     jit_word_t		word;
-#if __WORDSIZE == 64
+#if __X64
     jit_int32_t		reg;
 
     reg = jit_get_reg(jit_class_gpr);
@@ -3292,7 +3374,7 @@ _calli(jit_state_t *_jit, jit_word_t i0)
 static void
 _jmpr(jit_state_t *_jit, jit_int32_t r0)
 {
-    rex(0, 1, _NOREG, _NOREG, r0);
+    rex(0, WIDE, _NOREG, _NOREG, r0);
     ic(0xff);
     mrm(0x03, 0x04, r7(r0));
 }
@@ -3310,17 +3392,28 @@ _jmpi(jit_state_t *_jit, jit_word_t i0)
 static void
 _prolog(jit_state_t *_jit, jit_node_t *node)
 {
-#if __WORDSIZE == 32
-    _jitc->function->stack = (((_jitc->function->self.alen -
-			       _jitc->function->self.aoff) + 15) & -16) + 12;
+    if (_jitc->function->define_frame || _jitc->function->assume_frame) {
+	jit_int32_t	frame = -_jitc->function->frame;
+	assert(_jitc->function->self.aoff >= frame);
+	if (_jitc->function->assume_frame)
+	    return;
+	_jitc->function->self.aoff = frame;
+    }
+#if __X64 && __CYGWIN__
+    _jitc->function->stack = (((/* first 32 bytes must be allocated */
+				(_jitc->function->self.alen > 32 ?
+				 _jitc->function->self.alen : 32) -
+				/* align stack at 16 bytes */
+				_jitc->function->self.aoff) + 15) & -16) +
+	stack_adjust;
 #else
     _jitc->function->stack = (((_jitc->function->self.alen -
-			       _jitc->function->self.aoff) + 15) & -16) + 8;
+			       _jitc->function->self.aoff) + 15) & -16) +
+	stack_adjust;
 #endif
-
+    subi(_RSP_REGNO, _RSP_REGNO, stack_framesize - REAL_WORDSIZE);
     /* callee save registers */
-    subi(_RSP_REGNO, _RSP_REGNO, stack_framesize - sizeof(jit_word_t));
-#if __WORDSIZE == 32
+#if __X32
     if (jit_regset_tstbit(&_jitc->function->regset, _RDI))
 	stxi(12, _RSP_REGNO, _RDI_REGNO);
     if (jit_regset_tstbit(&_jitc->function->regset, _RSI))
@@ -3328,6 +3421,42 @@ _prolog(jit_state_t *_jit, jit_node_t *node)
     if (jit_regset_tstbit(&_jitc->function->regset, _RBX))
 	stxi( 4, _RSP_REGNO, _RBX_REGNO);
 #else
+#  if __CYGWIN__
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM15))
+	sse_stxi_d(136, _RSP_REGNO, _XMM15_REGNO);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM14))
+	sse_stxi_d(128, _RSP_REGNO, _XMM14_REGNO);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM13))
+	sse_stxi_d(120, _RSP_REGNO, _XMM13_REGNO);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM12))
+	sse_stxi_d(112, _RSP_REGNO, _XMM12_REGNO);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM11))
+	sse_stxi_d(104, _RSP_REGNO, _XMM11_REGNO);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM10))
+	sse_stxi_d(96, _RSP_REGNO, _XMM10_REGNO);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM9))
+	sse_stxi_d(88, _RSP_REGNO, _XMM9_REGNO);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM8))
+	sse_stxi_d(80, _RSP_REGNO, _XMM8_REGNO);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM7))
+	sse_stxi_d(72, _RSP_REGNO, _XMM7_REGNO);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM6))
+	sse_stxi_d(64, _RSP_REGNO, _XMM6_REGNO);
+    if (jit_regset_tstbit(&_jitc->function->regset, _R15))
+	stxi(56, _RSP_REGNO, _R15_REGNO);
+    if (jit_regset_tstbit(&_jitc->function->regset, _R14))
+	stxi(48, _RSP_REGNO, _R14_REGNO);
+    if (jit_regset_tstbit(&_jitc->function->regset, _R13))
+	stxi(40, _RSP_REGNO, _R13_REGNO);
+    if (jit_regset_tstbit(&_jitc->function->regset, _R12))
+	stxi(32, _RSP_REGNO, _R12_REGNO);
+    if (jit_regset_tstbit(&_jitc->function->regset, _RSI))
+	stxi(24, _RSP_REGNO, _RSI_REGNO);
+    if (jit_regset_tstbit(&_jitc->function->regset, _RDI))
+	stxi(16, _RSP_REGNO, _RDI_REGNO);
+    if (jit_regset_tstbit(&_jitc->function->regset, _RBX))
+	stxi( 8, _RSP_REGNO, _RBX_REGNO);
+#  else
     if (jit_regset_tstbit(&_jitc->function->regset, _RBX))
 	stxi(40, _RSP_REGNO, _RBX_REGNO);
     if (jit_regset_tstbit(&_jitc->function->regset, _R12))
@@ -3338,6 +3467,7 @@ _prolog(jit_state_t *_jit, jit_node_t *node)
 	stxi(16, _RSP_REGNO, _R14_REGNO);
     if (jit_regset_tstbit(&_jitc->function->regset, _R15))
 	stxi( 8, _RSP_REGNO, _R15_REGNO);
+#  endif
 #endif
     stxi(0, _RSP_REGNO, _RBP_REGNO);
     movr(_RBP_REGNO, _RSP_REGNO);
@@ -3349,9 +3479,11 @@ _prolog(jit_state_t *_jit, jit_node_t *node)
 static void
 _epilog(jit_state_t *_jit, jit_node_t *node)
 {
+    if (_jitc->function->assume_frame)
+	return;
     /* callee save registers */
     movr(_RSP_REGNO, _RBP_REGNO);
-#if __WORDSIZE == 32
+#if __X32
     if (jit_regset_tstbit(&_jitc->function->regset, _RDI))
 	ldxi(_RDI_REGNO, _RSP_REGNO, 12);
     if (jit_regset_tstbit(&_jitc->function->regset, _RSI))
@@ -3359,6 +3491,42 @@ _epilog(jit_state_t *_jit, jit_node_t *node)
     if (jit_regset_tstbit(&_jitc->function->regset, _RBX))
 	ldxi(_RBX_REGNO, _RSP_REGNO,  4);
 #else
+#  if __CYGWIN__
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM15))
+	sse_ldxi_d(_XMM15_REGNO, _RSP_REGNO, 136);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM14))
+	sse_ldxi_d(_XMM14_REGNO, _RSP_REGNO, 128);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM13))
+	sse_ldxi_d(_XMM13_REGNO, _RSP_REGNO, 120);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM12))
+	sse_ldxi_d(_XMM12_REGNO, _RSP_REGNO, 112);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM11))
+	sse_ldxi_d(_XMM11_REGNO, _RSP_REGNO, 104);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM10))
+	sse_ldxi_d(_XMM10_REGNO, _RSP_REGNO, 96);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM9))
+	sse_ldxi_d(_XMM9_REGNO, _RSP_REGNO, 88);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM8))
+	sse_ldxi_d(_XMM8_REGNO, _RSP_REGNO, 80);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM7))
+	sse_ldxi_d(_XMM7_REGNO, _RSP_REGNO, 72);
+    if (jit_regset_tstbit(&_jitc->function->regset, _XMM6))
+	sse_ldxi_d(_XMM6_REGNO, _RSP_REGNO, 64);
+    if (jit_regset_tstbit(&_jitc->function->regset, _R15))
+	ldxi(_R15_REGNO, _RSP_REGNO, 56);
+    if (jit_regset_tstbit(&_jitc->function->regset, _R14))
+	ldxi(_R14_REGNO, _RSP_REGNO, 48);
+    if (jit_regset_tstbit(&_jitc->function->regset, _R13))
+	ldxi(_R13_REGNO, _RSP_REGNO, 40);
+    if (jit_regset_tstbit(&_jitc->function->regset, _R12))
+	ldxi(_R12_REGNO, _RSP_REGNO, 32);
+    if (jit_regset_tstbit(&_jitc->function->regset, _RSI))
+	ldxi(_RSI_REGNO, _RSP_REGNO, 24);
+    if (jit_regset_tstbit(&_jitc->function->regset, _RDI))
+	ldxi(_RDI_REGNO, _RSP_REGNO, 16);
+    if (jit_regset_tstbit(&_jitc->function->regset, _RBX))
+	ldxi(_RBX_REGNO, _RSP_REGNO,  8);
+#  else
     if (jit_regset_tstbit(&_jitc->function->regset, _RBX))
 	ldxi(_RBX_REGNO, _RSP_REGNO, 40);
     if (jit_regset_tstbit(&_jitc->function->regset, _R12))
@@ -3369,9 +3537,10 @@ _epilog(jit_state_t *_jit, jit_node_t *node)
 	ldxi(_R14_REGNO, _RSP_REGNO, 16);
     if (jit_regset_tstbit(&_jitc->function->regset, _R15))
 	ldxi(_R15_REGNO, _RSP_REGNO,  8);
+#  endif
 #endif
     ldxi(_RBP_REGNO, _RSP_REGNO, 0);
-    addi(_RSP_REGNO, _RSP_REGNO, stack_framesize - sizeof(jit_word_t));
+    addi(_RSP_REGNO, _RSP_REGNO, stack_framesize - REAL_WORDSIZE);
 
     ic(0xc3);
 }
@@ -3381,9 +3550,9 @@ _patch_at(jit_state_t *_jit, jit_node_t *node,
 	  jit_word_t instr, jit_word_t label)
 {
     switch (node->code) {
-#if __WORDSIZE == 64
+#  if __X64
 	case jit_code_calli:
-#endif
+#  endif
 	case jit_code_movi:
 	    patch_abs(instr, label);
 	    break;
@@ -3392,4 +3561,23 @@ _patch_at(jit_state_t *_jit, jit_node_t *node,
 	    break;
     }
 }
+
+#  if __X64 && !defined(HAVE_FFSL)
+static int
+ffsl(long i)
+{
+    int		bit;
+#    if __CYGWIN__
+    /* Bug workaround */
+    if ((int)i == (int)0x80000000)
+	bit = 32;
+    else
+#    endif
+    if ((bit = ffs((int)i)) == 0) {
+	if ((bit = ffs((int)((unsigned long)i >> 32))))
+	    bit += 32;
+    }
+    return (bit);
+}
+#  endif
 #endif
