@@ -22,15 +22,26 @@
 
 #define rc(value)			jit_class_##value
 #define rn(reg)				(jit_regno(_rvs[jit_regno(reg)].spec))
-
-#if __WORDSIZE == 32
-#  define stack_alignment		4
+#if __X32
+#  define jit_arg_reg_p(i)		0
+#  define jit_arg_f_reg_p(i)		0
 #  define stack_framesize		20
-#    define CVT_OFFSET			-12
+#  define stack_adjust			12
+#  define CVT_OFFSET			-12
+#  define REAL_WORDSIZE			4
 #else
-#  define stack_alignment		8
-#  define stack_framesize		56
-#    define CVT_OFFSET			-8
+#  if __CYGWIN__
+#    define jit_arg_reg_p(i)		((i) >= 0 && (i) < 4)
+#    define jit_arg_f_reg_p(i)		jit_arg_reg_p(i)
+#    define stack_framesize		152
+#  else
+#    define jit_arg_reg_p(i)		((i) >= 0 && (i) < 6)
+#    define jit_arg_f_reg_p(i)		((i) >= 0 && (i) < 8)
+#    define stack_framesize		56
+#  endif
+#  define stack_adjust			8
+#  define CVT_OFFSET			-8
+#  define REAL_WORDSIZE			8
 #endif
 
 /*
@@ -58,7 +69,7 @@ static void _x87_from_sse_d(jit_state_t*,jit_int32_t,jit_int32_t);
  */
 jit_cpu_t		jit_cpu;
 jit_register_t		_rvs[] = {
-#if __WORDSIZE == 32
+#if __X32
     { rc(gpr) | rc(rg8) | 0,		"%eax" },
     { rc(gpr) | rc(rg8) | 1,		"%ecx" },
     { rc(gpr) | rc(rg8) | 2,		"%edx" },
@@ -83,6 +94,40 @@ jit_register_t		_rvs[] = {
     { rc(fpr) | 5,			"st(5)" },
     { rc(fpr) | 6,			"st(6)" },
     { rc(fpr) | 7,			"st(7)" },
+#else
+#  if __CYGWIN__
+    { rc(gpr) | rc(rg8) | 0,		"%rax" },
+    { rc(gpr) | rc(rg8) | rc(rg8) | 10,	"%r10" },
+    { rc(gpr) | rc(rg8) | rc(rg8) | 11,	"%r11" },
+    { rc(sav) | rc(rg8) | rc(gpr) | 3,	"%rbx" },
+    { rc(sav) | rc(gpr) | 7,		"%rdi" },
+    { rc(sav) | rc(gpr) | 6,		"%rsi" },
+    { rc(sav) | rc(gpr) | 12,		"%r12" },
+    { rc(sav) | rc(gpr) | 13,		"%r13" },
+    { rc(sav) | rc(gpr) | 14,		"%r14" },
+    { rc(sav) | rc(gpr) | 15,		"%r15" },
+    { rc(arg) | rc(rg8) | rc(gpr) | 9,	"%r9" },
+    { rc(arg) | rc(rg8) | rc(gpr) | 8,	"%r8" },
+    { rc(arg) | rc(rg8) | rc(gpr) | 2,	"%rdx" },
+    { rc(arg) | rc(rg8) | rc(gpr) | 1,	"%rcx" },
+    { rc(sav) | 4,			"%rsp" },
+    { rc(sav) | 5,			"%rbp" },
+    { rc(xpr) | rc(fpr) | 4,		"%xmm4" },
+    { rc(xpr) | rc(fpr) | 5,		"%xmm5" },
+    { rc(sav) | rc(xpr) | rc(fpr) | 6,	"%xmm6" },
+    { rc(sav) | rc(xpr) | rc(fpr) | 7,	"%xmm7" },
+    { rc(sav) | rc(xpr) | rc(fpr) | 8,	"%xmm8" },
+    { rc(sav) | rc(xpr) | rc(fpr) | 9,	"%xmm9" },
+    { rc(sav) | rc(xpr) | rc(fpr) | 10,	"%xmm10" },
+    { rc(sav) | rc(xpr) | rc(fpr) | 11,	"%xmm11" },
+    { rc(sav) | rc(xpr) | rc(fpr) | 12,	"%xmm12" },
+    { rc(sav) | rc(xpr) | rc(fpr) | 13,	"%xmm13" },
+    { rc(sav) | rc(xpr) | rc(fpr) | 14,	"%xmm14" },
+    { rc(sav) | rc(xpr) | rc(fpr) | 15,	"%xmm15" },
+    { rc(xpr) | rc(arg) | rc(fpr) | 3,	"%xmm3" },
+    { rc(xpr) | rc(arg) | rc(fpr) | 2,	"%xmm2" },
+    { rc(xpr) | rc(arg) | rc(fpr) | 1,	"%xmm1" },
+    { rc(xpr) | rc(arg) | rc(fpr) | 0,	"%xmm0" },
 #else
     /* %rax is a pseudo flag argument for varargs functions */
     { rc(arg) | rc(gpr) | rc(rg8) | 0,	"%rax" },
@@ -117,6 +162,7 @@ jit_register_t		_rvs[] = {
     { rc(xpr) | rc(arg) | rc(fpr) | 2,	"%xmm2" },
     { rc(xpr) | rc(arg) | rc(fpr) | 1,	"%xmm1" },
     { rc(xpr) | rc(arg) | rc(fpr) | 0,	"%xmm0" },
+#  endif
     { rc(fpr) | 0,			"st(0)" },
     { rc(fpr) | 1,			"st(1)" },
     { rc(fpr) | 2,			"st(2)" },
@@ -209,12 +255,12 @@ jit_get_cpu(void)
 	} bits;
 	jit_uword_t	cpuid;
     } edx;
-#if __WORDSIZE == 32
+#if __X32
     int			ac, flags;
 #endif
     jit_uword_t		eax, ebx;
 
-#if __WORDSIZE == 32
+#if __X32
     /* adapted from glibc __sysconf */
     __asm__ volatile ("pushfl;\n\t"
 		      "popl %0;\n\t"
@@ -236,7 +282,7 @@ jit_get_cpu(void)
 #endif
 
     /* query %eax = 1 function */
-#if __WORDSIZE == 32
+#if __X32 || __X64_32
     __asm__ volatile ("xchgl %%ebx, %1; cpuid; xchgl %%ebx, %1"
 #else
     __asm__ volatile ("xchgq %%rbx, %1; cpuid; xchgq %%rbx, %1"
@@ -263,9 +309,13 @@ jit_get_cpu(void)
     jit_cpu.aes		= ecx.bits.aes;
     jit_cpu.avx		= ecx.bits.avx;
 
-#if __WORDSIZE == 64
     /* query %eax = 0x80000001 function */
+#if __X64
+#  if __X64_32
+    __asm__ volatile ("xchgl %%ebx, %1; cpuid; xchgl %%ebx, %1"
+#  else
     __asm__ volatile ("xchgq %%rbx, %1; cpuid; xchgq %%rbx, %1"
+#  endif
 		      : "=a" (eax), "=r" (ebx),
 		      "=c" (ecx.cpuid), "=d" (edx.cpuid)
 		      : "0" (0x80000001));
@@ -276,13 +326,13 @@ jit_get_cpu(void)
 void
 _jit_init(jit_state_t *_jit)
 {
-#if __WORDSIZE == 32
+#if __X32
     jit_int32_t		regno;
     static jit_bool_t	first = 1;
 #endif
 
     _jitc->reglen = jit_size(_rvs) - 1;
-#if __WORDSIZE == 32
+#if __X32
     if (first) {
 	if (!jit_cpu.sse2) {
 	    for (regno = _jitc->reglen; regno >= 0; regno--) {
@@ -422,75 +472,78 @@ _jit_epilog(jit_state_t *_jit)
     _jitc->function = NULL;
 }
 
+jit_bool_t
+_jit_arg_register_p(jit_state_t *_jit, jit_node_t *u)
+{
+    if (u->code == jit_code_arg)
+	return (jit_arg_reg_p(u->u.w));
+    assert(u->code == jit_code_arg_f || u->code == jit_code_arg_d);
+    return (jit_arg_f_reg_p(u->u.w));
+}
+
 jit_node_t *
 _jit_arg(jit_state_t *_jit)
 {
     jit_int32_t		offset;
 
     assert(_jitc->function);
-#if __WORDSIZE == 64
-    if (_jitc->function->self.argi < 6)
+#if __X64
+    if (jit_arg_reg_p(_jitc->function->self.argi)) {
 	offset = _jitc->function->self.argi++;
+#  if __CYGWIN__
+	_jitc->function->self.size += sizeof(jit_word_t);
+#  endif
+    }
     else
 #endif
     {
 	offset = _jitc->function->self.size;
-	_jitc->function->self.size += sizeof(jit_word_t);
+	_jitc->function->self.size += REAL_WORDSIZE;
     }
     return (jit_new_node_w(jit_code_arg, offset));
 }
 
-jit_bool_t
-_jit_arg_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-#if __WORDSIZE == 32
-    return (0);
-#else
-    return (offset >= 0 && offset < 6);
-#endif
-}
-
 jit_node_t *
 _jit_arg_f(jit_state_t *_jit)
 {
     jit_int32_t		offset;
 
     assert(_jitc->function);
-#if __WORDSIZE == 64
-    if (_jitc->function->self.argf < 8)
+#if __X64
+#  if __CYGWIN__
+    if (jit_arg_reg_p(_jitc->function->self.argi)) {
+	offset = _jitc->function->self.argi++;
+	_jitc->function->self.size += sizeof(jit_word_t);
+    }
+#  else
+    if (jit_arg_f_reg_p(_jitc->function->self.argf))
 	offset = _jitc->function->self.argf++;
+#  endif
     else
 #endif
     {
 	offset = _jitc->function->self.size;
-#if __WORDSIZE == 32
-	_jitc->function->self.size += sizeof(jit_float32_t);
-#else
-	_jitc->function->self.size += sizeof(jit_float64_t);
-#endif
+	_jitc->function->self.size += REAL_WORDSIZE;
     }
     return (jit_new_node_w(jit_code_arg_f, offset));
 }
 
-jit_bool_t
-_jit_arg_f_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-#if __WORDSIZE == 32
-    return (0);
-#else
-    return (offset >= 0 && offset < 8);
-#endif
-}
-
 jit_node_t *
 _jit_arg_d(jit_state_t *_jit)
 {
     jit_int32_t		offset;
 
     assert(_jitc->function);
-#if __WORDSIZE == 64
-    if (_jitc->function->self.argf < 8)
+#if __X64
+#  if __CYGWIN__
+    if (jit_arg_reg_p(_jitc->function->self.argi)) {
+	offset = _jitc->function->self.argi++;
+	_jitc->function->self.size += sizeof(jit_word_t);
+    }
+#  else
+    if (jit_arg_f_reg_p(_jitc->function->self.argf))
 	offset = _jitc->function->self.argf++;
+#  endif
     else
 #endif
     {
@@ -500,18 +553,13 @@ _jit_arg_d(jit_state_t *_jit)
     return (jit_new_node_w(jit_code_arg_d, offset));
 }
 
-jit_bool_t
-_jit_arg_d_reg_p(jit_state_t *_jit, jit_int32_t offset)
-{
-    return (jit_arg_f_reg_p(offset));
-}
-
 void
 _jit_getarg_c(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-#if __WORDSIZE == 64
-    if (v->u.w < 6)
-	jit_extr_c(u, _RDI - v->u.w);
+    assert(v->code == jit_code_arg);
+#if __X64
+    if (jit_arg_reg_p(v->u.w))
+	jit_extr_c(u, JIT_RA0 - v->u.w);
     else
 #endif
 	jit_ldxi_c(u, _RBP, v->u.w);
@@ -520,9 +568,10 @@ _jit_getarg_c(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 void
 _jit_getarg_uc(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-#if __WORDSIZE == 64
-    if (v->u.w < 6)
-	jit_extr_uc(u, _RDI - v->u.w);
+    assert(v->code == jit_code_arg);
+#if __X64
+    if (jit_arg_reg_p(v->u.w))
+	jit_extr_uc(u, JIT_RA0 - v->u.w);
     else
 #endif
 	jit_ldxi_uc(u, _RBP, v->u.w);
@@ -531,9 +580,10 @@ _jit_getarg_uc(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 void
 _jit_getarg_s(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-#if __WORDSIZE == 64
-    if (v->u.w < 6)
-	jit_extr_s(u, _RDI - v->u.w);
+    assert(v->code == jit_code_arg);
+#if __X64
+    if (jit_arg_reg_p(v->u.w))
+	jit_extr_s(u, JIT_RA0 - v->u.w);
     else
 #endif
 	jit_ldxi_s(u, _RBP, v->u.w);
@@ -542,9 +592,10 @@ _jit_getarg_s(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 void
 _jit_getarg_us(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-#if __WORDSIZE == 64
-    if (v->u.w < 6)
-	jit_extr_us(u, _RDI - v->u.w);
+    assert(v->code == jit_code_arg);
+#if __X64
+    if (jit_arg_reg_p(v->u.w))
+	jit_extr_us(u, JIT_RA0 - v->u.w);
     else
 #endif
 	jit_ldxi_us(u, _RBP, v->u.w);
@@ -553,20 +604,27 @@ _jit_getarg_us(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 void
 _jit_getarg_i(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-#if __WORDSIZE == 64
-    if (v->u.w < 6)
-	jit_extr_i(u, _RDI - v->u.w);
+    assert(v->code == jit_code_arg);
+#if __X64
+    if (jit_arg_reg_p(v->u.w)) {
+#  if __X64_32
+	jit_movr(u, JIT_RA0 - v->u.w);
+#  else
+	jit_extr_i(u, JIT_RA0 - v->u.w);
+#  endif
+     }
     else
 #endif
 	jit_ldxi_i(u, _RBP, v->u.w);
 }
 
-#if __WORDSIZE == 64
+#if __X64 && !__X64_32
 void
 _jit_getarg_ui(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 6)
-	jit_extr_ui(u, _RDI - v->u.w);
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
+	jit_extr_ui(u, JIT_RA0 - v->u.w);
     else
 	jit_ldxi_ui(u, _RBP, v->u.w);
 }
@@ -574,49 +632,145 @@ _jit_getarg_ui(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 void
 _jit_getarg_l(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-    if (v->u.w < 6)
-	jit_movr(u, _RDI - v->u.w);
+    assert(v->code == jit_code_arg);
+    if (jit_arg_reg_p(v->u.w))
+	jit_movr(u, JIT_RA0 - v->u.w);
     else
 	jit_ldxi_l(u, _RBP, v->u.w);
 }
 #endif
 
+void
+_jit_putargr(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg);
+#if __X64
+    if (jit_arg_reg_p(v->u.w))
+	jit_movr(JIT_RA0 - v->u.w, u);
+    else
+#endif
+	jit_stxi(v->u.w, _RBP, u);
+}
+
+void
+_jit_putargi(jit_state_t *_jit, jit_word_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg);
+#if __X64
+    if (jit_arg_reg_p(v->u.w))
+	jit_movi(JIT_RA0 - v->u.w, u);
+    else
+#endif
+    {
+	regno = jit_get_reg(jit_class_gpr);
+	jit_movi(regno, u);
+	jit_stxi(v->u.w, _RBP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_getarg_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-#if __WORDSIZE == 64
-    if (v->u.w < 8)
+    assert(v->code == jit_code_arg_f);
+#if __X64
+    if (jit_arg_f_reg_p(v->u.w))
 	jit_movr_f(u, _XMM0 - v->u.w);
     else
 #endif
 	jit_ldxi_f(u, _RBP, v->u.w);
 }
 
+void
+_jit_putargr_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg_f);
+#if __X64
+    if (jit_arg_reg_p(v->u.w))
+	jit_movr_f(_XMM0 - v->u.w, u);
+    else
+#endif
+	jit_stxi_f(v->u.w, _RBP, u);
+}
+
+void
+_jit_putargi_f(jit_state_t *_jit, jit_float32_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg_f);
+#if __X64
+    if (jit_arg_reg_p(v->u.w))
+	jit_movi_f(_XMM0 - v->u.w, u);
+    else
+#endif
+    {
+	regno = jit_get_reg(jit_class_gpr);
+	jit_movi_f(regno, u);
+	jit_stxi_f(v->u.w, _RBP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_getarg_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
-#if __WORDSIZE == 64
-    if (v->u.w < 8)
+    assert(v->code == jit_code_arg_d);
+#if __X64
+    if (jit_arg_f_reg_p(v->u.w))
 	jit_movr_d(u, _XMM0 - v->u.w);
     else
 #endif
 	jit_ldxi_d(u, _RBP, v->u.w);
 }
 
+void
+_jit_putargr_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
+{
+    assert(v->code == jit_code_arg_d);
+#if __X64
+    if (jit_arg_reg_p(v->u.w))
+	jit_movr_d(_XMM0 - v->u.w, u);
+    else
+#endif
+	jit_stxi_d(v->u.w, _RBP, u);
+}
+
+void
+_jit_putargi_d(jit_state_t *_jit, jit_float64_t u, jit_node_t *v)
+{
+    jit_int32_t		regno;
+    assert(v->code == jit_code_arg_d);
+#if __X64
+    if (jit_arg_reg_p(v->u.w))
+	jit_movi_d(_XMM0 - v->u.w, u);
+    else
+#endif
+    {
+	regno = jit_get_reg(jit_class_gpr);
+	jit_movi_d(regno, u);
+	jit_stxi_d(v->u.w, _RBP, regno);
+	jit_unget_reg(regno);
+    }
+}
+
 void
 _jit_pushargr(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
-#if __WORDSIZE == 64
-    if (_jitc->function->call.argi < 6) {
-	jit_movr(_RDI - _jitc->function->call.argi, u);
+#if __X64
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
+	jit_movr(JIT_RA0 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
+#  if __CYGWIN__
+	_jitc->function->call.size += sizeof(jit_word_t);
+#  endif
     }
     else
 #endif
     {
 	jit_stxi(_jitc->function->call.size, _RSP, u);
-	_jitc->function->call.size += sizeof(jit_word_t);
+	_jitc->function->call.size += REAL_WORDSIZE;
     }
 }
 
@@ -626,10 +780,13 @@ _jit_pushargi(jit_state_t *_jit, jit_word_t u)
     jit_int32_t		 regno;
 
     assert(_jitc->function);
-#if __WORDSIZE == 64
-    if (_jitc->function->call.argi < 6) {
-	jit_movi(_RDI - _jitc->function->call.argi, u);
+#if __X64
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
+	jit_movi(JIT_RA0 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
+#  if __CYGWIN__
+	_jitc->function->call.size += sizeof(jit_word_t);
+#  endif
     }
     else
 #endif
@@ -637,7 +794,7 @@ _jit_pushargi(jit_state_t *_jit, jit_word_t u)
 	regno = jit_get_reg(jit_class_gpr);
 	jit_movi(regno, u);
 	jit_stxi(_jitc->function->call.size, _RSP, regno);
-	_jitc->function->call.size += sizeof(jit_word_t);
+	_jitc->function->call.size += REAL_WORDSIZE;
 	jit_unget_reg(regno);
     }
 }
@@ -646,16 +803,30 @@ void
 _jit_pushargr_f(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
-#if __WORDSIZE == 64
-    if (_jitc->function->call.argf < 8) {
+#if __X64
+#  if __CYGWIN__
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
+	jit_movr_f(_XMM0 - _jitc->function->call.argi, u);
+	if (_jitc->function->call.call & jit_call_varargs) {
+	    jit_stxi_f(_jitc->function->call.size, _RSP,
+		       _XMM0 - _jitc->function->call.argi);
+	    jit_ldxi_i(JIT_RA0 - _jitc->function->call.argi, _RSP,
+		       _jitc->function->call.size);
+	}
+	++_jitc->function->call.argi;
+	_jitc->function->call.size += sizeof(jit_word_t);
+    }
+#  else
+    if (jit_arg_f_reg_p(_jitc->function->self.argf)) {
 	jit_movr_f(_XMM0 - _jitc->function->call.argf, u);
 	++_jitc->function->call.argf;
     }
+#  endif
     else
 #endif
     {
 	jit_stxi_f(_jitc->function->call.size, _RSP, u);
-	_jitc->function->call.size += sizeof(jit_word_t);
+	_jitc->function->call.size += REAL_WORDSIZE;
     }
 }
 
@@ -665,18 +836,32 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
     jit_int32_t		regno;
 
     assert(_jitc->function);
-#if __WORDSIZE == 64
-    if (_jitc->function->call.argf < 8) {
+#if __X64
+#  if __CYGWIN__
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
+	jit_movi_f(_XMM0 - _jitc->function->call.argi, u);
+	if (_jitc->function->call.call & jit_call_varargs) {
+	    jit_stxi_f(_jitc->function->call.size, _RSP,
+		       _XMM0 - _jitc->function->call.argi);
+	    jit_ldxi_i(JIT_RA0 - _jitc->function->call.argi, _RSP,
+		       _jitc->function->call.size);
+	}
+	++_jitc->function->call.argi;
+	_jitc->function->call.size += sizeof(jit_word_t);
+    }
+#  else
+    if (jit_arg_f_reg_p(_jitc->function->call.argf)) {
 	jit_movi_f(_XMM0 - _jitc->function->call.argf, u);
 	++_jitc->function->call.argf;
     }
+#  endif
     else
 #endif
     {
 	regno = jit_get_reg(jit_class_fpr);
 	jit_movi_f(regno, u);
 	jit_stxi_f(_jitc->function->call.size, _RSP, regno);
-	_jitc->function->call.size += sizeof(jit_word_t);
+	_jitc->function->call.size += REAL_WORDSIZE;
 	jit_unget_reg(regno);
     }
 }
@@ -685,11 +870,25 @@ void
 _jit_pushargr_d(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
-#if __WORDSIZE == 64
-    if (_jitc->function->call.argf < 8) {
+#if __X64
+#  if __CYGWIN__
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
+	jit_movr_d(_XMM0 - _jitc->function->call.argi, u);
+	if (_jitc->function->call.call & jit_call_varargs) {
+	    jit_stxi_d(_jitc->function->call.size, _RSP,
+		       _XMM0 - _jitc->function->call.argi);
+	    jit_ldxi_l(JIT_RA0 - _jitc->function->call.argi, _RSP,
+		       _jitc->function->call.size);
+	}
+	++_jitc->function->call.argi;
+	_jitc->function->call.size += sizeof(jit_word_t);
+    }
+#  else
+    if (jit_arg_f_reg_p(_jitc->function->call.argf)) {
 	jit_movr_d(_XMM0 - _jitc->function->call.argf, u);
 	++_jitc->function->call.argf;
     }
+#  endif
     else
 #endif
     {
@@ -704,11 +903,25 @@ _jit_pushargi_d(jit_state_t *_jit, jit_float64_t u)
     jit_int32_t		 regno;
 
     assert(_jitc->function);
-#if __WORDSIZE == 64
-    if (_jitc->function->call.argf < 8) {
+#if __X64
+#  if __CYGWIN__
+    if (jit_arg_reg_p(_jitc->function->call.argi)) {
+	jit_movi_d(_XMM0 - _jitc->function->call.argi, u);
+	if (_jitc->function->call.call & jit_call_varargs) {
+	    jit_stxi_d(_jitc->function->call.size, _RSP,
+		       _XMM0 - _jitc->function->call.argi);
+	    jit_ldxi_l(JIT_RA0 - _jitc->function->call.argi, _RSP,
+		       _jitc->function->call.size);
+	}
+	++_jitc->function->call.argi;
+	_jitc->function->call.size += sizeof(jit_word_t);
+    }
+#  else
+    if (jit_arg_f_reg_p(_jitc->function->call.argf)) {
 	jit_movi_d(_XMM0 - _jitc->function->call.argf, u);
 	++_jitc->function->call.argf;
     }
+#  endif
     else
 #endif
     {
@@ -723,13 +936,13 @@ _jit_pushargi_d(jit_state_t *_jit, jit_float64_t u)
 jit_bool_t
 _jit_regarg_p(jit_state_t *_jit, jit_node_t *node, jit_int32_t regno)
 {
-#if __WORDSIZE == 64
+#if __X64
     jit_int32_t		spec;
 
     spec = jit_class(_rvs[regno].spec);
     if (spec & jit_class_arg) {
 	if (spec & jit_class_gpr) {
-	    regno = _RDI - regno;
+	    regno = JIT_RA0 - regno;
 	    if (regno >= 0 && regno < node->v.w)
 		return (1);
 	}
@@ -753,7 +966,8 @@ _jit_finishr(jit_state_t *_jit, jit_int32_t r0)
     assert(_jitc->function);
     if (_jitc->function->self.alen < _jitc->function->call.size)
 	_jitc->function->self.alen = _jitc->function->call.size;
-#if __WORDSIZE == 64
+#if __X64
+#  if !__CYGWIN__
     if (_jitc->function->call.call & jit_call_varargs) {
 	if (jit_regno(reg) == _RAX) {
 	    reg = jit_get_reg(jit_class_gpr);
@@ -766,6 +980,7 @@ _jit_finishr(jit_state_t *_jit, jit_int32_t r0)
 	if (reg != r0)
 	    jit_unget_reg(reg);
     }
+#  endif
 #endif
     call = jit_callr(reg);
     call->v.w = _jitc->function->call.argi;
@@ -778,7 +993,7 @@ _jit_finishr(jit_state_t *_jit, jit_int32_t r0)
 jit_node_t *
 _jit_finishi(jit_state_t *_jit, jit_pointer_t i0)
 {
-#if __WORDSIZE == 64
+#if __X64
     jit_int32_t		reg;
 #endif
     jit_node_t		*node;
@@ -786,7 +1001,7 @@ _jit_finishi(jit_state_t *_jit, jit_pointer_t i0)
     assert(_jitc->function);
     if (_jitc->function->self.alen < _jitc->function->call.size)
 	_jitc->function->self.alen = _jitc->function->call.size;
-#if __WORDSIZE == 64
+#if __X64
     /* FIXME preventing %rax allocation is good enough, but for consistency
      * it should automatically detect %rax is dead, in case it has run out
      * registers, and not save/restore it, what would be wrong if using the
@@ -837,7 +1052,7 @@ _jit_retval_us(jit_state_t *_jit, jit_int32_t r0)
 void
 _jit_retval_i(jit_state_t *_jit, jit_int32_t r0)
 {
-#if __WORDSIZE == 32
+#if __X32 || __X64_32
     if (r0 != JIT_RET)
 	jit_movr(r0, JIT_RET);
 #else
@@ -845,7 +1060,7 @@ _jit_retval_i(jit_state_t *_jit, jit_int32_t r0)
 #endif
 }
 
-#if __WORDSIZE == 64
+#if __X64 && !__X64_32
 void
 _jit_retval_ui(jit_state_t *_jit, jit_int32_t r0)
 {
@@ -863,7 +1078,7 @@ _jit_retval_l(jit_state_t *_jit, jit_int32_t r0)
 void
 _jit_retval_f(jit_state_t *_jit, jit_int32_t r0)
 {
-#  if __WORDSIZE == 32
+#  if __X32
     jit_new_node_w(jit_code_x86_retval_f, r0);
 #  else
     if (r0 != JIT_FRET)
@@ -874,7 +1089,7 @@ _jit_retval_f(jit_state_t *_jit, jit_int32_t r0)
 void
 _jit_retval_d(jit_state_t *_jit, jit_int32_t r0)
 {
-#  if __WORDSIZE == 32
+#  if __X32
     jit_new_node_w(jit_code_x86_retval_d, r0);
 #  else
     if (r0 != JIT_FRET)
@@ -1136,14 +1351,23 @@ _emit_code(jit_state_t *_jit)
 	if (_jit->pc.uc >= _jitc->code.end)
 	    return (NULL);
 
+#if DEVEL_DISASSEMBLER
+	node->offset = _jit->pc.w;
+#endif
 	value = jit_classify(node->code);
 	jit_regarg_set(node, value);
 	switch (node->code) {
+	    case jit_code_align:
+		assert(!(node->u.w & (node->u.w - 1)) &&
+		       node->u.w <= sizeof(jit_word_t));
+		if ((word = _jit->pc.w & (node->u.w - 1)))
+		    nop(node->u.w - word);
+		break;
 	    case jit_code_note:		case jit_code_name:
 		node->u.w = _jit->pc.w;
 		break;
 	    case jit_code_label:
-		if (node->link &&
+		if ((node->link || (node->flag & jit_flag_use)) &&
 		    (word = _jit->pc.w & (sizeof(jit_word_t) - 1)))
 		    nop(sizeof(jit_word_t) - word);
 		/* remember label is defined */
@@ -1162,6 +1386,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrw(subx,);
 		case_rrr(subc,);
 		case_rrw(subc,);
+		case_rrw(rsb,);
 		case_rrr(mul,);
 		case_rrw(mul,);
 		case_rrrr(qmul,);
@@ -1232,18 +1457,22 @@ _emit_code(jit_state_t *_jit)
 		else
 		    movi(rn(node->u.w), node->v.w);
 		break;
-		case_rr(hton,);
+		case_rr(hton, _us);
+		case_rr(hton, _ui);
+#if __X64 && !__X64_32
+		case_rr(hton, _ul);
+#endif
 		case_rr(ext, _c);
 		case_rr(ext, _uc);
 		case_rr(ext, _s);
 		case_rr(ext, _us);
-#if __WORDSIZE == 64
+#if __X64 && !__X64_32
 		case_rr(ext, _i);
 		case_rr(ext, _ui);
 #endif
 		case_rf(trunc, _f_i);
 		case_rf(trunc, _d_i);
-#if __WORDSIZE == 64
+#if __X64
 		case_rf(trunc, _f_l);
 		case_rf(trunc, _d_l);
 #endif
@@ -1257,7 +1486,7 @@ _emit_code(jit_state_t *_jit)
 		case_rw(ld, _us);
 		case_rr(ld, _i);
 		case_rw(ld, _i);
-#if __WORDSIZE == 64
+#if __X64 && !__X64_32
 		case_rr(ld, _ui);
 		case_rw(ld, _ui);
 		case_rr(ld, _l);
@@ -1273,7 +1502,7 @@ _emit_code(jit_state_t *_jit)
 		case_rrw(ldx, _us);
 		case_rrr(ldx, _i);
 		case_rrw(ldx, _i);
-#if __WORDSIZE == 64
+#if __X64 && !__X64_32
 		case_rrr(ldx, _ui);
 		case_rrw(ldx, _ui);
 		case_rrr(ldx, _l);
@@ -1285,7 +1514,7 @@ _emit_code(jit_state_t *_jit)
 		case_wr(st, _s);
 		case_rr(st, _i);
 		case_wr(st, _i);
-#if __WORDSIZE == 64
+#if __X64 && !__X64_32
 		case_rr(st, _l);
 		case_wr(st, _l);
 #endif
@@ -1295,7 +1524,7 @@ _emit_code(jit_state_t *_jit)
 		case_wrr(stx, _s);
 		case_rrr(stx, _i);
 		case_wrr(stx, _i);
-#if __WORDSIZE == 64
+#if __X64 && !__X64_32
 		case_rrr(stx, _l);
 		case_wrr(stx, _l);
 #endif
@@ -1343,6 +1572,7 @@ _emit_code(jit_state_t *_jit)
 		case_ffw(add, _f, 32);
 		case_fff(sub, _f);
 		case_ffw(sub, _f, 32);
+		case_ffw(rsb, _f, 32);
 		case_fff(mul, _f);
 		case_ffw(mul, _f, 32);
 		case_fff(div, _f);
@@ -1441,6 +1671,7 @@ _emit_code(jit_state_t *_jit)
 		case_ffw(add, _d, 64);
 		case_fff(sub, _d);
 		case_ffw(sub, _d, 64);
+		case_ffw(rsb, _d, 64);
 		case_fff(mul, _d);
 		case_ffw(mul, _d, 64);
 		case_fff(div, _d);
@@ -1539,15 +1770,19 @@ _emit_code(jit_state_t *_jit)
 		jmpr(rn(node->u.w));
 		break;
 	    case jit_code_jmpi:
-		temp = node->u.n;
-		assert(temp->code == jit_code_label ||
-		       temp->code == jit_code_epilog);
-		if (temp->flag & jit_flag_patch)
-		    jmpi(temp->u.w);
-		else {
-		    word = jmpi(_jit->pc.w);
-		    patch(word, node);
+		if (node->flag & jit_flag_node) {
+		    temp = node->u.n;
+		    assert(temp->code == jit_code_label ||
+			   temp->code == jit_code_epilog);
+		    if (temp->flag & jit_flag_patch)
+			jmpi(temp->u.w);
+		    else {
+			word = jmpi(_jit->pc.w);
+			patch(word, node);
+		    }
 		}
+		else
+		    jmpi(node->u.w);
 		break;
 	    case jit_code_callr:
 		callr(rn(node->u.w));
@@ -1597,7 +1832,7 @@ _emit_code(jit_state_t *_jit)
 		epilog(node);
 		_jitc->function = NULL;
 		break;
-#if __WORDSIZE == 32
+#if __X32
 	    case jit_code_x86_retval_f:
 		if (jit_sse_reg_p(node->u.w)) {
 		    fstpr(_ST1_REGNO);
@@ -1623,6 +1858,7 @@ _emit_code(jit_state_t *_jit)
 		abort();
 	}
 	jit_regarg_clr(node, value);
+	assert(_jitc->regarg == 0);
 	/* update register live state */
 	jit_reglive(node);
     }
@@ -1651,6 +1887,8 @@ _emit_code(jit_state_t *_jit)
 	patch_at(node, _jitc->patches.ptr[offset].inst, word);
     }
 
+    jit_flush(_jit->code.ptr, _jit->pc.uc);
+
     return (_jit->code.ptr);
 }
 
@@ -1660,6 +1898,11 @@ _emit_code(jit_state_t *_jit)
 #  include "jit_x86-x87.c"
 #undef CODE
 
+void
+jit_flush(void *fptr, void *tptr)
+{
+}
+
 void
 _emit_ldxi(jit_state_t *_jit, jit_gpr_t r0, jit_gpr_t r1, jit_word_t i0)
 {
