@@ -44,7 +44,7 @@
 /*
  * Prototypes
  */
-static jit_word_t hash_data(jit_pointer_t, jit_word_t);
+static jit_word_t hash_data(const void*, jit_word_t);
 
 #define new_pool()			_new_pool(_jit)
 static void _new_pool(jit_state_t*);
@@ -178,7 +178,7 @@ _patch_register(jit_state_t *jit, jit_node_t *node, jit_node_t *link,
  * Implementation
  */
 void
-init_jit(char *progname)
+init_jit(const char *progname)
 {
     jit_get_cpu();
     jit_init_debug(progname);
@@ -536,9 +536,9 @@ _jit_load(jit_state_t *_jit, jit_int32_t reg)
 }
 
 static jit_word_t
-hash_data(jit_pointer_t data, jit_word_t length)
+hash_data(const void *data, jit_word_t length)
 {
-    jit_uint8_t		*ptr;
+    const jit_uint8_t		*ptr;
     jit_word_t		 i, key;
     for (i = key = 0, ptr = data; i < length; i++)
 	key = (key << (key & 1)) ^ ptr[i];
@@ -558,7 +558,7 @@ _jit_address(jit_state_t *_jit, jit_node_t *node)
 }
 
 jit_node_t *
-_jit_data(jit_state_t *_jit, jit_pointer_t data,
+_jit_data(jit_state_t *_jit, const void *data,
 	  jit_word_t length, jit_int32_t align)
 {
     jit_word_t		 key;
@@ -1127,6 +1127,24 @@ _jit_link(jit_state_t *_jit, jit_node_t *node)
     ++_jitc->blocks.offset;
 }
 
+jit_bool_t
+_jit_forward_p(jit_state_t *_jit, jit_node_t *node)
+{
+    return (node->code == jit_code_label && !node->next && node != _jitc->tail);
+}
+
+jit_bool_t
+_jit_indirect_p(jit_state_t *_jit, jit_node_t *node)
+{
+    return (node->code == jit_code_label && !!(node->flag & jit_flag_use));
+}
+
+jit_bool_t
+_jit_target_p(jit_state_t *_jit, jit_node_t *node)
+{
+    return (node->code == jit_code_label && !!node->link);
+}
+
 void
 _jit_prepare(jit_state_t *_jit)
 {
@@ -1177,9 +1195,9 @@ _jit_classify(jit_state_t *_jit, jit_code_t code)
     jit_int32_t		mask;
 
     switch (code) {
-	case jit_code_data:	case jit_code_save:	case jit_code_load:
-	case jit_code_name:	case jit_code_label:	case jit_code_note:
-	case jit_code_prolog:	case jit_code_epilog:
+	case jit_code_data:	case jit_code_align:	case jit_code_save:
+	case jit_code_load:	case jit_code_name:	case jit_code_label:
+	case jit_code_note:	case jit_code_prolog:	case jit_code_epilog:
 	    mask = 0;
 	    break;
 	case jit_code_live:
@@ -1219,7 +1237,8 @@ _jit_classify(jit_state_t *_jit, jit_code_t code)
 	case jit_code_extr_us:	case jit_code_extr_i:	case jit_code_extr_ui:
 	case jit_code_truncr_f_i:			case jit_code_truncr_f_l:
 	case jit_code_truncr_d_i:			case jit_code_truncr_d_l:
-	case jit_code_htonr:	case jit_code_ldr_c:	case jit_code_ldr_uc:
+	case jit_code_htonr_us:	case jit_code_htonr_ui:	case jit_code_htonr_ul:
+	case jit_code_ldr_c:	case jit_code_ldr_uc:
 	case jit_code_ldr_s:	case jit_code_ldr_us:	case jit_code_ldr_i:
 	case jit_code_ldr_ui:	case jit_code_ldr_l:	case jit_code_negr_f:
 	case jit_code_absr_f:	case jit_code_sqrtr_f:	case jit_code_movr_f:
@@ -1237,6 +1256,7 @@ _jit_classify(jit_state_t *_jit, jit_code_t code)
 	    break;
 	case jit_code_addi:	case jit_code_addxi:	case jit_code_addci:
 	case jit_code_subi:	case jit_code_subxi:	case jit_code_subci:
+	case jit_code_rsbi:
 	case jit_code_muli:	case jit_code_divi:	case jit_code_divi_u:
 	case jit_code_remi:	case jit_code_remi_u:	case jit_code_andi:
 	case jit_code_ori:	case jit_code_xori:	case jit_code_lshi:
@@ -1254,20 +1274,22 @@ _jit_classify(jit_state_t *_jit, jit_code_t code)
 	    mask = jit_cc_a0_reg|jit_cc_a0_rlh|jit_cc_a0_chg|
 		   jit_cc_a1_reg|jit_cc_a2_int;
 	    break;
-	case jit_code_addi_f:	case jit_code_subi_f:	case jit_code_muli_f:
-	case jit_code_divi_f:	case jit_code_lti_f:	case jit_code_lei_f:
-	case jit_code_eqi_f:	case jit_code_gei_f:	case jit_code_gti_f:
-	case jit_code_nei_f:	case jit_code_unlti_f:	case jit_code_unlei_f:
-	case jit_code_uneqi_f:	case jit_code_ungei_f:	case jit_code_ungti_f:
-	case jit_code_ltgti_f:	case jit_code_ordi_f:	case jit_code_unordi_f:
+	case jit_code_addi_f:	case jit_code_subi_f:	case jit_code_rsbi_f:
+	case jit_code_muli_f:	case jit_code_divi_f:	case jit_code_lti_f:
+	case jit_code_lei_f:	case jit_code_eqi_f:	case jit_code_gei_f:
+	case jit_code_gti_f:	case jit_code_nei_f:	case jit_code_unlti_f:
+	case jit_code_unlei_f:	case jit_code_uneqi_f:	case jit_code_ungei_f:
+	case jit_code_ungti_f:	case jit_code_ltgti_f:	case jit_code_ordi_f:
+	case jit_code_unordi_f:
 	    mask = jit_cc_a0_reg|jit_cc_a0_chg|jit_cc_a1_reg|jit_cc_a2_flt;
 	    break;
-	case jit_code_addi_d:	case jit_code_subi_d:	case jit_code_muli_d:
-	case jit_code_divi_d:	case jit_code_lti_d:	case jit_code_lei_d:
-	case jit_code_eqi_d:	case jit_code_gei_d:	case jit_code_gti_d:
-	case jit_code_nei_d:	case jit_code_unlti_d:	case jit_code_unlei_d:
-	case jit_code_uneqi_d:	case jit_code_ungei_d:	case jit_code_ungti_d:
-	case jit_code_ltgti_d:	case jit_code_ordi_d:	case jit_code_unordi_d:
+	case jit_code_addi_d:	case jit_code_subi_d:	case jit_code_rsbi_d:
+	case jit_code_muli_d:	case jit_code_divi_d:	case jit_code_lti_d:
+	case jit_code_lei_d:	case jit_code_eqi_d:	case jit_code_gei_d:
+	case jit_code_gti_d:	case jit_code_nei_d:	case jit_code_unlti_d:
+	case jit_code_unlei_d:	case jit_code_uneqi_d:	case jit_code_ungei_d:
+	case jit_code_ungti_d:	case jit_code_ltgti_d:	case jit_code_ordi_d:
+	case jit_code_unordi_d:
 	    mask = jit_cc_a0_reg|jit_cc_a0_chg|jit_cc_a1_reg|jit_cc_a2_dbl;
 	    break;
 	case jit_code_addr:	case jit_code_addxr:	case jit_code_addcr:
@@ -1920,6 +1942,45 @@ fail:
     return (NULL);
 }
 
+void
+_jit_frame(jit_state_t *_jit, jit_int32_t frame)
+{
+    jit_trampoline(frame, 1);
+}
+
+void
+_jit_tramp(jit_state_t *_jit, jit_int32_t frame)
+{
+    jit_trampoline(frame, 0);
+}
+
+void
+_jit_trampoline(jit_state_t *_jit, jit_int32_t frame, jit_bool_t prolog)
+{
+    jit_int32_t		regno;
+
+    /* Must be called after prolog, actually, just to simplify
+     * tests and know there is a current function and that
+     * _jitc->function->self.aoff is at the before any alloca value */
+    assert(_jitc->tail && _jitc->tail->code == jit_code_prolog);
+
+    /* + 24 for 3 possible spilled temporaries (that could be a double) */
+    frame += 24;
+#if defined(__hppa__)
+    frame += _jitc->function->self.aoff;
+#else
+    frame -= _jitc->function->self.aoff;
+#endif
+    _jitc->function->frame = frame;
+    if (prolog)
+	_jitc->function->define_frame = 1;
+    else
+	_jitc->function->assume_frame = 1;
+    for (regno = 0; regno < _jitc->reglen; regno++)
+	if (jit_class(_rvs[regno].spec) & jit_class_sav)
+	    jit_regset_setbit(&_jitc->function->regset, regno);
+}
+
 /*   Compute initial reglive and regmask set values of a basic block.
  * reglive is the set of known live registers
  * regmask is the set of registers not referenced in the block
@@ -2006,6 +2067,7 @@ _jit_update(jit_state_t *_jit, jit_node_t *node,
 	    jit_regset_t *live, jit_regset_t *mask)
 {
     jit_int32_t		 spec;
+    jit_int32_t		 regno;
     jit_regset_t	 ztmp;
     jit_regset_t	 zmask;
     unsigned long	 value;
@@ -2165,6 +2227,22 @@ _jit_update(jit_state_t *_jit, jit_node_t *node,
 			}
 			continue;
 		    }
+		    /* Should not really mark as live all registers in unknown
+		     * state if using jit_jmpr(), or jit_jmpi(absolute_address)
+		     * because that would leave the register allocator with
+		     * no options for "nospill" temporaries (other temporaries
+		     * also benefit from not needing to spill/reload), so, the
+		     * user must ensure to either spill/reload, or only leave
+		     * live values on registers that are advertised as
+		     * callee save (as per jit_callee_save_p); on most targets
+		     * these are the JIT_Vn registers. */
+		    for (regno = 0; regno < _jitc->reglen; regno++) {
+			spec = jit_class(_rvs[regno].spec);
+			if (jit_regset_tstbit(mask, regno) &&
+			    (spec & (jit_class_gpr|jit_class_fpr)) &&
+			    !(spec & jit_class_sav))
+			    jit_regset_clrbit(mask, regno);
+		    }
 		    /* assume value is live due to jump to unknown location */
 		    jit_regset_ior(live, live, mask);
 		    jit_regset_set_ui(mask, 0);
@@ -2278,6 +2356,8 @@ _shortcut_jump(jit_state_t *_jit, jit_node_t *prev, jit_node_t *node)
     for (next = jump->next; next; next = next->next) {
 	switch (next->code) {
 	    case jit_code_jmpi:
+		if (!(next->flag & jit_flag_node))
+		    return (0);
 		if (jump->link == node)
 		    jump->link = node->link;
 		else {
@@ -2337,7 +2417,8 @@ _redundant_jump(jit_state_t *_jit, jit_node_t *prev, jit_node_t *node)
 		    return (1);
 		}
 		break;
-	    case jit_code_note:
+	    case jit_code_name:		case jit_code_note:
+	    case jit_code_align:
 		break;
 	    default:
 		return (0);
@@ -2483,7 +2564,8 @@ _reverse_jump(jit_state_t *_jit, jit_node_t *prev, jit_node_t *node)
 	return (0);
     /* =><cond_jump L0> <jump L1> <label L0> */
     local_next = node->next;
-    if (local_next->code != jit_code_jmpi)
+    if (local_next->code != jit_code_jmpi ||
+	!(local_next->flag & jit_flag_node))
 	return (0);
     /* <cond_jump L0> =><jump L1> <label L0> */
 
@@ -2677,7 +2759,8 @@ _simplify_ldxi(jit_state_t *_jit, jit_node_t *prev, jit_node_t *node)
     regno = jit_regno(node->u.w);
     right = jit_regno(node->v.w);
     value = _jitc->values + regno;
-    if (value->kind == jit_kind_code && value->code == node->code &&
+    if (regno != right &&
+	value->kind == jit_kind_code && value->code == node->code &&
 	value->base.q.l == right && value->base.q.h == _jitc->gen[right] &&
 	node->w.w == value->disp.w) {
 	del_node(prev, node);
@@ -2706,9 +2789,10 @@ _simplify_stxi(jit_state_t *_jit, jit_node_t *prev, jit_node_t *node)
     value = _jitc->values + regno;
 
     /* check for redundant store after load */
-    if (value->kind == jit_kind_code && value->code == node->code &&
+    if (regno != right &&
+	value->kind == jit_kind_code && value->code == node->code &&
 	value->base.q.l == right && value->base.q.h == _jitc->gen[right] &&
-	node->w.w == value->disp.w) {
+	node->u.w == value->disp.w) {
 	del_node(prev, node);
 	return (1);
     }
@@ -2736,7 +2820,6 @@ _simplify_stxi(jit_state_t *_jit, jit_node_t *prev, jit_node_t *node)
 	    default:		 	abort();
 	}
 	value->kind = jit_kind_code;
-	value->code = node->code;
 	value->base.q.l = right;
 	value->base.q.h = _jitc->gen[right];
 	value->disp.w = node->u.w;
@@ -2774,12 +2857,12 @@ _simplify(jit_state_t *_jit)
     jit_node_t		*next;
     jit_int32_t		 info;
     jit_int32_t		 regno;
-    jit_int32_t		 offset;
 
     for (prev = NULL, node = _jitc->head; node; prev = node, node = next) {
 	next = node->next;
 	switch (node->code) {
 	    case jit_code_label:	case jit_code_prolog:
+	    case jit_code_callr:	case jit_code_calli:
 	    reset:
 		memset(_jitc->gen, 0, sizeof(jit_int32_t) * _jitc->reglen);
 		memset(_jitc->values, 0, sizeof(jit_value_t) * _jitc->reglen);
@@ -2855,14 +2938,6 @@ _simplify(jit_state_t *_jit)
 		if (simplify_stxi(prev, node))
 		    simplify_spill(node = prev, regno);
 		break;
-	    case jit_code_callr:	case jit_code_calli:
-		for (offset = 0; offset < _jitc->reglen; offset++) {
-		    if (!(jit_class(_rvs[offset].spec) & jit_class_sav)) {
-			_jitc->values[offset].kind = 0;
-			++_jitc->gen[offset];
-		    }
-		}
-		break;
 	    default:
 		info = jit_classify(node->code);
 		if (info & jit_cc_a0_jmp)
@@ -3111,8 +3186,8 @@ _patch_register(jit_state_t *_jit, jit_node_t *node, jit_node_t *link,
 #  include "jit_hppa.c"
 #elif defined(__aarch64__)
 #  include "jit_aarch64.c"
-#elif defined(__s390x__)
-#  include "jit_s390x.c"
+#elif defined(__s390__) || defined(__s390x__)
+#  include "jit_s390.c"
 #elif defined(__alpha__)
 #  include "jit_alpha.c"
 #endif
