@@ -20,8 +20,9 @@
 #include <lightning.h>
 #include <lightning/jit_private.h>
 #include <stdio.h>
+#include "lib/jit_names.c"
 
-jit_int16_t	_szs[jit_code_x86_retval_d + 1];
+jit_int16_t	_szs[jit_code_last_code + 1];
 
 int
 main(int argc, char *argv[])
@@ -39,7 +40,7 @@ main(int argc, char *argv[])
     fclose(fp);
 
     max = 0;
-    for (offset = 0; offset <= jit_code_x86_retval_d; offset++)
+    for (offset = 0; offset <= jit_code_last_code; offset++)
 	if (max < _szs[offset])
 	    max = _szs[offset];
 
@@ -47,7 +48,20 @@ main(int argc, char *argv[])
 	exit(-1);
 
 
+#if __X64 || __X32
+#  if __X64
+    fprintf(fp, "#if __X64\n");
+#    if __X64_32
+    fprintf(fp, "#  if __X64_32\n");
+#    else
+    fprintf(fp, "#  if !__X64_32\n");
+#    endif
+#  else
+    fprintf(fp, "#if __X32\n");
+#  endif
+#else
     fprintf(fp, "#if __WORDSIZE == %d\n", __WORDSIZE);
+#endif
 #if defined(__arm__)
 #  if defined(__ARM_PCS_VFP)
     fprintf(fp, "#if defined(__ARM_PCS_VFP)\n");
@@ -68,8 +82,8 @@ main(int argc, char *argv[])
     fprintf(fp, "#if defined(__powerpc__)\n");
 #endif
     fprintf(fp, "#define JIT_INSTR_MAX %d\n", max);
-    for (offset = 0; offset <= jit_code_x86_retval_d; offset++)
-	fprintf(fp, "    %d,\n", _szs[offset]);
+    for (offset = 0; offset <= jit_code_last_code; offset++)
+	fprintf(fp, "    %d,	/* %s */\n", _szs[offset], code_name[offset]);
 #if defined(__arm__)
     fprintf(fp, "#endif /* __ARM_PCS_VFP */\n");
 #elif defined(__mips__)
@@ -81,7 +95,17 @@ main(int argc, char *argv[])
 #elif defined(__powerpc__)
     fprintf(fp, "#endif /* __powerpc__ */\n");
 #endif
+#if __X64 || __X32
+#  if __X64
+    fprintf(fp, "#  endif /* __X64_32 */\n");
+    fprintf(fp, "#endif /* __X64 */\n");
+#  else
+    fprintf(fp, "#if __X32\n");
+#  endif
+#else
     fprintf(fp, "#endif /* __WORDSIZE */\n");
+#endif
+
     fclose(fp);
 
     return (0);
